/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package debug;

import java.io.*;
import java.net.*;

/**
 * Debug class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
public class Debug
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Debug activation / deactivation
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * True if the debug function is active.
   * Putting this variable to false and using an optimizing compiler such as Jikes
   * guarantee that all debug code will be removed from generated bytecode; this to
   * avoid overhead in the execution when debug is not needed.
   */
  public static final boolean DEBUG = true;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Debugging level (console) */
  private static int clevel;

  /** Debugging level (file) */
  private static int flevel;

  /** Log writer */
  private static PrintWriter out;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initializer
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Reads and initializes the 
   */
  static {
    String cString = System.getProperty("jgroup.debug.outLevel");
    if (cString != null) {
      try {
        clevel = Integer.parseInt(cString);
      } catch (NumberFormatException e) {
      }
    }
    String fString = System.getProperty("jgroup.debug.fileLevel");
    if (fString != null) {
      try {
        flevel = Integer.parseInt(fString);
      } catch (NumberFormatException e) {
      }
    }
    String dirString = System.getProperty("jgroup.debug.dir");
    if (dirString != null) {
      try {
        File dir = new File(dirString);
        String name = InetAddress.getLocalHost().toString();
        int t1 = name.indexOf('.');
        int t2 = name.indexOf('/');
        t1 = (t1 < t2 ? t1 : t2);
        name = name.substring(0, t1)+".log";
        out = new PrintWriter(new FileOutputStream(new File(dir, name)));
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initializer
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static void printError()
  {
    try {
      throw new Exception("Debug method called when Debug is not active");
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Set the file debug level to <t>level</t>.
   * @param level the new debugging level.
   */
  public static void setFileLevel(int level)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null)
      flevel = level;
  }

  /**
   * Set the console debug level to <t>level</t>.
   * @param level the new debugging level.
   */
  public static void setOutLevel(int level)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null)
      clevel = level;
  }
  
  
  /**
   * Prints the specified information on the standard output. If the debug is
   * active, prints also the information on the log file.
   *
   * @param string the string to be printed.
   */
  public static void print(String string)
  {
    System.out.print(string);
    if (DEBUG) {
      if (out != null)
        out.print(string);
    }
  }

  /**
   * Prints the specified information on the standard output. If the debug is
   * active, prints also the information on the log file.
   *
   * @param obj the object to be printed.
   */
  public static void print(Object obj)
  {
    System.out.print(obj);
    if (DEBUG) {
      if (out != null)
        out.print(obj);
    }
  }

  /**
   * Prints the specified information on the standard output and terminates
   * the line. If the debug is active, prints also the information on the
   * log file and terminates the line.
   *
   * @param string the string to be printed.
   */
  public static void println(String string)
  {
    System.out.println(string);
    if (DEBUG) {
      if (out != null)
        out.println(string);
    }
  }

  /**
   * Prints the specified information on the standard output and terminates
   * the line. If the debug is active, prints also the information on the
   * log file and terminates the line.
   *
   * @param obj the object to be printed.
   */
  public static void println(Object obj)
  {
    System.out.println(obj);
    if (DEBUG) {
      if (out != null)
        out.println(obj);
    }
  }

  /**
   * If the debug is not active, print an exception on the standard output.
   * If the debug is active, prints the specified information on the log
   * file; if <t>level</t> is greater or equal than the current debug level,
   * prints the specified information on the standard output as well.
   *
   * @param level the debug level of this information
   * @param string the string to be printed.
   */
  public static void print(int level, String string)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null) {
      if (level >= clevel)
        System.out.print(string);
      if (level >= flevel)
        out.print(string);
    }
  }

  /**
   * If the debug is not active, print an exception on the standard output.
   * If the debug is active, prints the specified information on the log
   * file; if <t>level</t> is greater or equal than the current debug level,
   * prints the specified information on the standard output as well.
   *
   * @param level the debug level of this information
   * @param obj the object to be printed.
   */
  public static void print(int level, Object obj)
  {
    if (!DEBUG) {
      printError();
      return;
    }

		System.out.println("level " + level + " clevel " + clevel + " flevel " + flevel); 
    if (out != null) {
      if (level >= clevel)
        System.out.print(obj);
      if (level >= flevel)
        out.print(obj);
    }
  }

  /**
   * If the debug is not active, print an exception on the standard output.
   * If the debug is active, prints the specified information on the log
   * file and terminates the line; if <t>level</t> is greater or equal than
   * the current debug level, prints the specified information on the standard
   * output as well and terminates the line.
   *
   * @param level the debug level of this information
   * @param string the string to be printed.
   */
  public static void println(int level, String string)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null) {
      if (level >= clevel)
        System.out.println(string);
      if (level >= flevel)
        out.println(string);
      out.flush();
    }
  }

  /**
   * If the debug is not active, print an exception on the standard output.
   * If the debug is active, prints the specified information on the log
   * file and terminates the line; if <t>level</t> is greater or equal than
   * the current debug level, prints the specified information on the standard
   * output as well and terminates the line.
   *
   * @param level the debug level of this information
   * @param obj the object to be printed.
   */
  public static void println(int level, Object obj)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null) {
      if (level >= clevel)
        System.out.println(obj);
      if (level >= flevel)
        out.println(obj);
      out.flush();
    }
  }

  /**
   * Print the stack trace of the specified exception on the standard output;
   * if the debug function is active, print the stack trace on the log file
   * as well.
   *
   * @param e the exception to be printed.
   */
  public static void printStackTrace(Exception e)
  {
    e.printStackTrace();
    if (DEBUG) {
      if (out != null)
        e.printStackTrace(out);
    }
  }

  /**
   * If the debug is not active, prints an exception on the standard output.
   * If the debug is active, prints the stack trace of the specified exception
   * on the log file; if <t>level</t> is greater or equal than the current
   * debug level, prints the specified the stack trace on the standard output
   * as well.
   *
   * @param e the exception to be printed.
   */
  public static void printStackTrace(int level, Exception e)
  {
    if (!DEBUG) {
      printError();
      return;
    }
    if (out != null) {
      if (level >= clevel)
        e.printStackTrace();
      if (level >= flevel)
        e.printStackTrace(out);
    }
  }

  /**
   * Print information about the current threads.
   */
  public static void printThreads()
  {
    Thread.dumpStack();
    Thread thread = Thread.currentThread();
    ThreadGroup tgroup = thread.getThreadGroup();
    while (tgroup.getParent() != null)
      tgroup = tgroup.getParent();
    Thread[] threads = new Thread[1000];
    tgroup.enumerate(threads, true);
    for (int i=0; threads[i] != null && i < threads.length; i++)
      System.out.println(threads[i]);
  }

} // END Debug

