/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * Revision history:
 * 24 Nov. 2000:
 *   Added support for dynamic generation of server proxies.   
 */
 

package jgroup;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import java.rmi.*;

import jgroup.*;
import jgroup.multicast.*;
import jgroup.reference.*;
import relacs.config.*;
import relacs.gm.*;
import relacs.rmi.*;
import relacs.rmi.proxy.*;

/**
 * The <code>GroupManager</code> class is used to configure the Jgroup system. In order to
 * make use of the Jgroup facilities, a group manager must be obtained through the invocation
 * of static method <tt>getGroupManager</tt>. As a parameter, this method takes a set of
 * objects implementing one or more listener interfaces. A group manager is constructed
 * by collecting and configuring the services needed by these objects. In order to obtain
 * an handler to one of these services, method <tt>getService</tt> could be invoked on
 * the group manager, by specifing the interface class of the desired service.

n object must request a group manager using the static
 * method of this class.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.8
 */
public class    GroupManager
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static variables
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** True if the class library contains dynamic proxy classes (1.3 or greater) */
  private static boolean hasDynamicProxy = false;

  /** Method object for class java.lang.reflect.Proxy */
  private static Method proxyMethod;

  /** Method object for class java.lang.reflect.Proxy */
  private static Method getLayerMethod;
  
  static {
    try {
      /* 
       * We failed to obtain a static proxy; so, we try to create
       * a dynamic one 
       */
      Class proxyClass = Class.forName("java.lang.reflect.Proxy");
      Class cl = Class.forName("java.lang.reflect.InvocationHandler");
      Class[] types = new Class[] { ClassLoader.class, Class[].class, cl };
      proxyMethod = proxyClass.getMethod("newProxyInstance", types);
      Class handlerClass = Class.forName("relacs.rmi.IntGroupHandler3");
      types = new Class[] { Object[].class, MulticastService.class, DispatcherService.class };
      getLayerMethod = handlerClass.getMethod("getLayer", types);
      hasDynamicProxy = true;
    } catch (Exception e) { /* Do nothing; hasDynamicProxy is false */}
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  HashMap map;    // HashMap mapping classes to their correspondant objects

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  private GroupManager()
  {
  }

  private GroupManager(HashMap map)
  {
    this.map = map;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns an handler to a Jgroup service (e.g., the membership service) specified by the
   * interface class <tt>cl</tt>. In order to such service to be returned by a group manager,
   * the corresponding listener interface must be implemented by one of the listener objects
   * that have been passed as an argument of method <tt>getGroupManager</tt> when creating
   * the group manager.
   *
   * @param     cl
   *            Class specifying the interface of the required service
   *
   * @exception ClassCastException
   *            If the generated group manager is unable to provide the requested service
   *            class.
   */
  public Object getService(Class cl)
  {
    Object service = map.get(cl);
    if (service == null)
      throw new ClassCastException("Class " + cl.getName() + " not found");
    return service;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a group manager object containing handlers for each of the services associated
   * to the listener interfaces implemented by objects contained in the <tt>listeners</tt>
   * array.
   *
   * @param configuration objects needed to configure the requested services
   * @param listeners objects implementing the listeners interfaces associated to the
   *  the requested services (currently, this object may contain at most one listener)
   * @exception JgroupException raised if some configuration error occurs.
   */
  public static GroupManager getGroupManager(Object[] configuration, Object[] listeners)
    throws JgroupException, RemoteException
  {
    DispatcherLayer         dispatcherLayer = null;
    BaseLayer               baseLayer = null;
    MergingLayer            mergingLayer = null;
    ExtGMILayer             extGMILayer = null;

    HashMap map = new HashMap();

    /*
     *  For the moment, this version of method getGroupManager is 
     *  restricted to return a group manager for at most one listener.
     *  Future version (see getGroupManager2) will enable multiple
     *  listeners to share the same group manager.
     */
    if (listeners == null || listeners.length != 1)
      throw new JgroupException("At most one listener admitted in the listeners array");
    Object server = listeners[0];

    dispatcherLayer = DispatcherLayer.getLayer(configuration);
    baseLayer = BaseLayer.getLayer(configuration, dispatcherLayer);
    dispatcherLayer.addListener(baseLayer);
    if (server instanceof MembershipListener) {
      if (!(server instanceof MergingListener)) {
        baseLayer.addListener((MembershipListener) server);
      }
      map.put(MembershipService.class, baseLayer);
    }

    if (server instanceof MergingListener) {
      mergingLayer = MergingLayer.getLayer(configuration, baseLayer, baseLayer);
      baseLayer.addListener((MulticastListener) mergingLayer);
      baseLayer.addListener((MembershipListener) mergingLayer);
      mergingLayer.addListener((MembershipListener) server);
      mergingLayer.addListener((MergingListener) server);
    }

    if (server instanceof ExternalGMIListener) {
      extGMILayer = ExtGMILayer.getLayer(configuration, baseLayer, baseLayer);
      baseLayer.addListener((MulticastListener) extGMILayer);
      extGMILayer.addListener((ExternalGMIListener) server);
      map.put(ExternalGMIService.class, extGMILayer);
    }

    if (server instanceof InternalGMIListener) {
      
      /* 
       *  Compute the list of internal interfaces implemented by this
       *  service 
       */
      ArrayList list = new ArrayList();
      Class[] interfaces = server.getClass().getInterfaces();
      for (int i=0; i < interfaces.length; i++) {
        if (InternalGMIListener.class.isAssignableFrom(interfaces[i])) {
          list.add(interfaces[i]);
        }
      }
      
      /*
       *  Obtain a proxy for the server
       */
      Object proxy = getProxy(server, list, configuration, baseLayer, dispatcherLayer);

      /*
       *  Register the handler in the group manager graph.  The handler 
       *  is obtained by invoking method getHandler on the object; note 
       *  that we cannot invoke this method staticly, because it is
       *  defined in two different interfaces (java.lang.reflect.Proxy
       *  and the _ServerProxy class).
       */
      try {
        IntGroupHandler2 handler = (IntGroupHandler2) 
          ((HandlerGetter) proxy).getInvocationHandler();
        baseLayer.addListener((MulticastListener) handler);
        handler.addListener((InternalGMIListener) server);
      } catch (ClassCastException e) { 
        throw new JgroupException("Class " + proxy.getClass().getName() +
          "does not implement interface relacs.rmi.proxy.HandlerGetter()");
      }
      
      /*
       *  Associate the proxy to every internal interface implemented by
       *  the server 
       */
      for (int i=0; i < list.size(); i++) {
        map.put(list.get(i), proxy);
      }
      
      /*
       *  Associate the proxy to the InternalGMIService interface, also
       *  implemented by the proxy.
       */
      map.put(InternalGMIService.class, proxy);
    }

    if (server instanceof MulticastListener) {
      baseLayer.addListener((MulticastListener) server);
      map.put(MulticastService.class, baseLayer);
    }

    return new GroupManager(map);
  }


  /**
   *  Returns a proxy object for the server indicated in the
   *
   */
  static Object getProxy(Object server, ArrayList list, Object[] conf,
    MulticastService multicastService, DispatcherService dispatcherService)
    throws JgroupException
  {
    String name = server.getClass().getName()+"_ServerProxy";
    
    /* 
     * First of all, we attempt to find a static stub for this class;
     * static stubs do not require dynamic generation and thus are
     * more efficient 
     */

    try { 

      Class cl = Class.forName(name); 
      Constructor constructor = cl.getConstructor(new Class[] { JgroupHandler.class });

      /* IntGroupHandler2 is an invocation handler not JDK1.3-aware */
      IntGroupHandler2 handler = IntGroupHandler2.getLayer(conf, multicastService, dispatcherService);

      /* Now, we instantiate a proxy object using the handler create above and
       * we add the appropriate listeners */
      return constructor.newInstance(new Object[] { handler });
    } catch (ClassNotFoundException e) {
      if (!hasDynamicProxy) {
        throw new JgroupException("Server proxy not found: " + name);
      }
    } catch (NoSuchMethodException e) {
      if (!hasDynamicProxy) {
        throw new JgroupException("Constructor not found in " + name);
      }
    } catch (Exception e) {
      if (!hasDynamicProxy) {
        throw new JgroupException(e.getMessage());
      }
    }
    
    /**
     *  If we get here, this means that (1) no static proxy has been found, 
     *  and (2) we are running on Jdk 1.3, since dynamic proxy classes are 
     *  available.
     */

    try {
      
      /* First, the class loader */
      ClassLoader classLoader = server.getClass().getClassLoader();
      
      /* Then, the interfaces to be implemented (those listed in list
         plus InternalGMIService */
      Class[] interfaces = new Class[list.size()+2];
      list.toArray(interfaces);
      interfaces[list.size()] = InternalGMIService.class;
      interfaces[list.size()+1] = HandlerGetter.class;
      
      /* Finally, the invocation handler */
      Class[] types = new Class[] { Object[].class,  MulticastService.class, DispatcherService.class };
      IntGroupHandler2 handler = (IntGroupHandler2)
        getLayerMethod.invoke( null, new Object[] { conf, multicastService, dispatcherService });
      
      /* Now we can obtain the proxy object */
      return proxyMethod.invoke( null, new Object[] { classLoader, interfaces, handler } );

    } catch (Exception e) { 
      e.printStackTrace();
      /** Something went wrong; FIXME what? */
      throw new JgroupException(e.getMessage());
    }
  }


  /**
   *  This method is a first schema of a completely dynamic solution
   *  to the problem of generating a group manager for a set of
   *  interfaces implemented by a server.
   */
  private GroupManager getGroupManager2(Object[] configuration, Object[] listeners)
    throws JgroupException, RemoteException
  {

    /*
     *  Creates a vector to contain the list of components composing
     *  the vector. The first component corresponds to the application
     *  server.
     */
    ArrayList components = new ArrayList();
    components.add(new Component(listeners[0]));

    /*
     *  Compute the list of providers needed to satisfy the services
     *  requested by the application server.
     */
    Component comp;
    for (int i=0; i < components.size(); i++) {
      comp = (Component) components.get(i);
      comp.getProviders(components, i-1);
    }


    /*
     *  Instantiates and link the components
     */

    return null;
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static field
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static Class listenerClass;
  private static Class ExternalGMIListenerClass;
  private static Class InternalGMIListenerClass;

  static {
    try {
      listenerClass = Class.forName("jgroup.reference.Listener");
      ExternalGMIListenerClass = Class.forName("jgroup.ExternalGMIListener");
      InternalGMIListenerClass = Class.forName("jgroup.InternalGMIListener");
    } catch (Exception e) {
      // PANIC!!!
      e.printStackTrace();
      System.exit(0);
    }
  }


/**
 *
 * @author      Alberto Montresor
 * @since       Jgroup 0.9
 */
final class     Component
{


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The class implemented by this layer
   */
  Class         layerClass;

  /**
   *  Hash table which maps the interfaces of the services provided by
   *  this layer into the corresponding component
   */
  HashMap       consumers;

  /**
   *  Hash table which maps the interfaces of the services requested by
   *  this layer into the corresponding component
   */
  HashMap       providers;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Construct a new <tt>Component</tt> object for a layer
   *  object implementing class <tt>cl</tt>.
   */
  Component(Class cl)
    throws JgroupException
  {
    Method[]      methods;
    Constructor[] constructors;
    Class[]       parameters;

    // Copy the layer class
    layerClass = cl;

    // Identify the services consumers by this layer
    consumers = new HashMap();
    methods = cl.getMethods();
    for (int i=0; i < methods.length; i++) {
      if (methods[i].getName().equals("addListener")) {
        parameters = methods[i].getParameterTypes();
        if (parameters.length != 1)
          throw new MalformedLayerException("addListener methods must have exactly one parameter");
        if (!Listener.class.isAssignableFrom(parameters[0]))
          throw new MalformedLayerException("addListener arguments must extend jgroup.Listener");
        consumers.put(parameters[0], new ArrayList());
      }
    }

    // Identify the services requested by this layer
    providers = new HashMap();
    constructors = cl.getConstructors();
    if (methods.length != 1)
      throw new MalformedLayerException("GM layer must have exactly one constructor");
    parameters = methods[0].getParameterTypes();
    for (int i=0; i < parameters.length; i++)
      if (Service.class.isAssignableFrom(parameters[0]))
        providers.put(parameters[0], null);
  }

  /*
   *  Construct a new <tt>Component</tt> object for an application
   *  server.
   */
  Component(Object server)
    throws JgroupException
  {
    providers = new HashMap();
    Class[] interfaces = server.getClass().getInterfaces();
    for (int i=0; i < interfaces.length; i++) {
      if (listenerClass.isAssignableFrom(interfaces[i])) {
        providers.put(interfaces[i],null);
      } else if (ExternalGMIListenerClass.isAssignableFrom(interfaces[i])) {
        providers.put(interfaces[i],null);
      } else if (InternalGMIListenerClass.isAssignableFrom(interfaces[i])) {
        providers.put(interfaces[i],null);
      }
    }
    consumers = new HashMap();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  void getProviders(ArrayList components, int last)
    throws JgroupException
  {
    Component comp;
    Class key;
    ArrayList myset;
    ArrayList set;

    // Iterate of the interfaces consumed from the providers
    Iterator i = providers.keySet().iterator();
    while (i.hasNext()) {
      key = (Class) i.next();

      // Identify the component providing the requested service
      comp = getProvider(key, components, last);

      // Add it to the providers hash table
      providers.put(key, comp);

      // Identify the service record associated to a provided service
      // for both this component and the component providing the service
      myset = (ArrayList) consumers.get(key);
      set = (ArrayList) comp.consumers.get(key);
      if (myset != null) {
        // If this component both offers and consumes the service represented
        // by the interface <key>, the existing component providing the service
        // is substituted with this component.

        // First we add all the consumers of the service to this component
        consumers.put(key, set);

        // Then we change the mapping in variable <providers> of the
        // components using this component
        for (int j=set.size()-1; j >=0; j--)
          ((Component) set.get(j)).providers.put(key, this);

        // Then we change the mapping in variable <consumers>
        set = new ArrayList();
        comp.consumers.put(key, set);
      }
      // Just add this component to the set of consumers of the
      // component providing the service
      set.add(this);
    }
  }

  private Component getProvider(Class listener, ArrayList components, int last)
    throws JgroupException
  {
    Component comp = null;
    Class cl = null;
    String classname = Service.getResource(listener.getName());
    if (classname == null)
      throw new MalformedLayerException("Reosurce " + listener.getName() + " not found");
    try {
      cl = Class.forName(classname);
    } catch (ClassNotFoundException e) {
      throw new MalformedLayerException("Layer " + classname +
        " associated to the listener interface " + listener.getName() + " not found");
    }
    boolean found = false;
    for (int i=last; i >= 0 && !found; i++) {
      comp = (Component) components.get(i);
      if (comp.layerClass.equals(cl))
        found = true;
    }
    if (!found) {
      comp = new Component(cl);
      components.add(comp);
    }
    return comp;
  }

  /**
   * Compares two <code>Component</code>s for content equality.
   *
   * @param obj the Object to compare with
   * @return  true if these Objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if ((obj != null) && (obj instanceof Component)) {
      Component comp = (Component)obj;
      return (layerClass == comp.layerClass);
    } else {
      return false;
    }
  }

} // END Component

} // END GroupManager






