/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.exec;

import java.io.*;
import debug.Debug;
/**
 * ClassData specifies the class and arguments of a replica. These are also
 * used to compute the group identifier.
 *
 * @author      Hein Meling
 * @since       Jgroup 1.1
 */
public class ClassData implements Serializable {

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private String    className;
  private String[]  classArgs;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  public ClassData(String className, String[] classArgs)
    throws ClassNotFoundException
  {
    Class cl = Class.forName(className);
    this.className = className;
    this.classArgs = classArgs;
  }

  public ClassData(Class clasz, String[] classArgs)
  {
    this.className = clasz.getName();
    this.classArgs = classArgs;
  }

  /////////////////////////////////////////////////////////////////////////////
  // Static class methods
  /////////////////////////////////////////////////////////////////////////////

  public static int computeGid(Object obj, String[] argv)
  {
    ClassData cdata = new ClassData(obj.getClass(), argv);
    return cdata.getGid();
  }

  /////////////////////////////////////////////////////////////////////////////
  // Access methods
  /////////////////////////////////////////////////////////////////////////////

  public ClassData getClassData()
  {
    return this;
  }

  public Class getClassObject()
  {
    Class cl = null;
    try {
      cl = Class.forName(className);
    } catch (ClassNotFoundException cnf) {
      /* Should never happen since we throw an exception in the constructor
       * if the class was not found, and thus we cannot get this far if the
       * class was not available.
       */
    }
    return cl;
  }

  public String getClassName()
  {
    return className;
  }

  public String[] getClassArgs()
  {
    return classArgs;
  }

  public int getGid()
  {
    return Math.abs(this.hashCode());
    // Group IDs are currently limited to two bytes.
//     return (this.hashCode() & 0xFFFF);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Object methods; overriding from the Object class
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    return "[" + className + " " + classArgs + "]";
  }

  /**
   *  Returns a hashcode for the <code>ClassData</code> class.
   *  Two <code>ClassData</code> objects will return the same hashcode
   *  if their class name and arguments are equal.
   */
  public int hashCode()
  {
//    return className.hashCode() ^ classArgs.hashCode();
    int hash = className.hashCode();
    for (int i = 0; i < classArgs.length; i++) {
      hash ^= classArgs[i].hashCode();
    }
    return hash;
  }

  /**
   *  Compares two <code>ClassData</code> objects for content equality.
   *
   *  @param  obj   The object to compare with.
   *  @return true  If these objects are equal; false otherwise.
   */
  public boolean equals(Object obj)
  {
    if (!(obj instanceof ClassData))
      return false;
    ClassData cd = (ClassData) obj;
    return (className.equals(cd.className) && classArgs.equals(cd.classArgs));
  }

} // END of ClassData
