/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.exec;

import java.io.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.*;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

import jgroup.*;
import jgroup.registry.*;
import relacs.config.RelacsDS;
import debug.Debug;

/**
 * ExecDaemon implements the ExecService allowing a replica manager
 * or other entity to start replicas on hosts running the ExecDaemon.
 *
 * Note an important property of ExecDaemon is that it does not have any state.
 *
 * @author      Hein Meling
 * @since       Jgroup 1.1
 */
public class ExecDaemon implements ExecService {

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  /** Map of replica classes running on this host */
  private HashMap   replicaMap;
  /** Count of replicas running on this host */
  private int       replicaCount = 0;

  /////////////////////////////////////////////////////////////////////////////
  // Static fields
  /////////////////////////////////////////////////////////////////////////////

  /** JVM executable and classpath */
  private static String    jvmCmd   = null;
  /** The Runtime environment for this VM */
  private static Runtime   rt       = Runtime.getRuntime();
  /** Determine the local host name */
  private static String    thisHost = RelacsDS.getLocalHostName();

  static {
    String s = File.separator;
    jvmCmd = System.getProperty("java.home")+ s +"bin"+ s +"java";
    if (Debug.DEBUG) {
      String logdir = System.getProperty("user.home")+ s +"Jgroup"+ s +"log";
      jvmCmd += " -Djgroup.debug.dir="+ logdir +" -Djgroup.debug.outLevel=60";
    }
    String classpath = System.getProperty("java.class.path");
    if (classpath != null) {
      jvmCmd += " -classpath "+ classpath +" ";
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Constructor
  /////////////////////////////////////////////////////////////////////////////

  public ExecDaemon(String configURL, int eport, int rport)
//    throws Exception
  {
    replicaMap = new HashMap();

    try {
      UnicastRemoteObject.exportObject(this);
      Registry registry = LocateRegistry.createRegistry(eport);
      registry.bind(EXEC_DAEMON_NAME, this);
    } catch (RemoteException re) {
      // throw new Abort("ExecDaemon; RemoteException", re); // problem in compiling ; Alberto 3/5/01
    } catch (AlreadyBoundException abe) {
      // throw new Abort("ExecDaemon already running on " + thisHost); // problem in compiling ; Alberto 3/5/01
    }

    try {
      /* If the ExecDaemon is started after the ReplicaManager, then lookup
       * the ReplicaManager in the dependable registry and notify our presence. */
      DistributedSystem ds = new DistributedSystem(configURL);
      /* Obtain a proxy for the dependable registry running in the distributed system. */
      DependableRegistry reg = RegistryFactory.getRegistry(ds, rport);
      ReplicationManager rm = (ReplicationManager)
        reg.lookup(ReplicationManager.REPLICATION_MANAGER_NAME);
      rm.notifyPresence(thisHost, this);
    } catch(RemoteException e) {
      System.out.println(ReplicationManager.REPLICATION_MANAGER_NAME + " not yet available.");
      if (Debug.DEBUG) {
        e.printStackTrace();
      }
    } catch(NotBoundException nbe) {
      System.out.println(ReplicationManager.REPLICATION_MANAGER_NAME + " not yet available.");
      if (Debug.DEBUG) {
        nbe.printStackTrace();
      }
    }
    System.out.println("ExecDaemon ready...");
  }

  /////////////////////////////////////////////////////////////////////////////
  // Main method
  /////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv) {
    // Default values
    String configURL = null;
    int eport = EXEC_BOOTSTRAP_PORT;
    int rport = 0;

    // Read arguments
    for (int i = 0 ; i < argv.length ; i++) {
      if  (argv[i].equals("-eport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -eport must be followed by a port number");
          usage();
          return;
        }
        try {
          eport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].equals("-rport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -rport must be followed by a port number");
          usage();
          return;
        }
        try {
          rport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (configURL != null) {
          System.err.println("Configuration URL already specified");
          usage();
          return;
        }
        configURL = argv[i];
      }
    }
    if (configURL == null) {
      System.err.println("You must specify a configuration URL");
      usage();
      return;
    }

    // Start the ExecDaemon
    ExecDaemon exec = new ExecDaemon(configURL, eport, rport);
    while (true) {
      try { Thread.sleep(0x7fffffffffffffffL);
      } catch (InterruptedException ex) {}
    }
  }

  private static void usage() {
    System.out.println("Usage: daemon <configURL> [-eport <port number>] [-rport <port number>]");
    System.exit(0);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Methods from ExecService interface
  /////////////////////////////////////////////////////////////////////////////

  public void start(Class c, String[] args)
    throws RemoteException, ExecException
  {
    try {
      // Get main(String[] args) method
      Method main = c.getDeclaredMethod("main", new Class[] {String[].class});
      int m = main.getModifiers();
      if (main == null || !Modifier.isPublic(m) || !Modifier.isStatic(m)) {
        throw new ExecException("Class " + c.getName() + " does not implement"
                                + " a 'public static void main()' method.");
      } else {
        Debug.println("Starting class " + c.getName());
        try {
          // Invoke the main method
          main.invoke(null, new Object[] {args});
          System.out.println("Method main invoked...");
        } catch (IllegalAccessException e) {
          throw new ExecException("The main method of class " + c.getName()
                                  + " is not accessible.", e);
        } catch (IllegalArgumentException e) {
          throw new ExecException("The main method of class " + c.getName()
                                  + " contains illegal arguments.", e);
        } catch (InvocationTargetException e) {
          throw new ExecException("The main method of class " + c.getName()
                                  + " has thrown an exception.", e.getTargetException());
        }
      }
    } catch (SecurityException e) {
      throw new ExecException("Security voilation detected", e);
    } catch (NoSuchMethodException e) {
      throw new ExecException("Class " + c.getName()
                              + " does not implement a main() method.", e);
    }
  }

  /**
   *  Removes the local replica belonging to the group specified in groupid.
   *
   *  @param groupid  Group identifier
   *
   */
  public void removeReplica(int groupid)
    throws RemoteException, ExecException
  {
    Integer gid = new Integer(groupid);
    if (replicaMap.containsKey(gid)) {
      /* Get the local replica thread for the specified group. */
      ReplicaThread replicaThrd = (ReplicaThread) replicaMap.get(gid);
      /* Remove the local replica and its thread. */
      replicaThrd.remove();
      replicaCount--;
    } else {
      Debug.println("Replica not running on host " + thisHost);
      throw new ExecException("Replica not running on host " + thisHost);
    }
  }


  /**
   *  Creates a new replica in a seperate process. Only one replica matching
   *  gid can be started on each host. This makes the operation idempotent,
   *  allowing several replication managers to invoke the same create operation
   *  on the same host as another replication manager already may have done.
   *  This prevents creating multiple instances of the same class with identical
   *  arguments. Thus, different arguments are not considered the same instance.
   *
   */
  public void createReplica(ClassData classData)
    throws RemoteException, ExecException
  {
    Integer gid = new Integer(classData.getGid());
    if (replicaMap.containsKey(gid)) {
      Debug.println("Replica already running on host " + thisHost);
      throw new ExecException("Replica already running on host " + thisHost);
    } else {
      synchronized(this) {
        ReplicaThread replicaThrd = new ReplicaThread(classData);
        Thread t = new Thread(replicaThrd);
        t.start();
        replicaCount++;
        /* Keep the replica thread object in case we want to remove it */
        replicaMap.put(gid, replicaThrd);
      }
      Debug.println("Returning... createReplica: " + classData.getClassName());
    }
  }

  /**
   * Private inner class used to start replicas, and keep waiting
   * for the replica to end.
   *
   */
  private class ReplicaThread implements Runnable {
    private Process  replica;
    private String   classname;
    private String   shortClass;
    private Integer  gid;

    public ReplicaThread(ClassData classData)
      throws ExecException
    {
      gid = new Integer(classData.getGid());
      classname = classData.getClassName();
      String[] args = classData.getClassArgs();
      shortClass = classname.substring(classname.lastIndexOf('.') + 1);
      String cmd = jvmCmd + classname;
      for (int i = 0; i < args.length; i++) {
        cmd += " " + args[i];
      }
      try {
        replica = rt.exec(cmd);
        Debug.println("Starting new replica: " + cmd);
        Debug.println("Starting new replica: " + classname);
      } catch (java.io.IOException ioe) {
        Debug.println("IOException: " + ioe);
        throw new ExecException("Replica could not be started.", ioe);
      }
    }

    public void remove() {
      /*
       * This will cause the run() method below to pass through the waitFor()
       * method and thus it will clean up its tables and exit this thread.
       */
      replica.destroy();
    }

    public void run() {
      try {
        InputStream drOut = replica.getInputStream();
        BufferedReader out = new BufferedReader(new InputStreamReader(drOut));
        InputStream drErr = replica.getErrorStream();
        BufferedReader err = new BufferedReader(new InputStreamReader(drErr));
        String line;
        while ((line = out.readLine()) != null) {
          synchronized(this) {
            System.out.println(shortClass + "-out: " + line);
          }
        }
        while ((line = err.readLine()) != null) {
          synchronized(this) {
            System.err.println(shortClass + "-err: " + line);
          }
        }
        if (replica.waitFor() != 0) {
          System.err.println("Got non-zero return value; replica failed or removed.");
        }
      } catch (java.io.IOException ioe) {
        Debug.println("IOException: " + ioe);
      } catch (InterruptedException ie) {
        Debug.println("InterruptedException: " + ie);
      } finally {
        // Replica exit; always clean up.
        Debug.println("Replica exiting: " + classname);
        synchronized(this) {
          replicaMap.remove(gid);
          replicaCount--;
        }
      }
    }

  } // END of pirvate inner class ReplicaThread

} // END of public class ExecDaemon
