/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.exec;

import java.util.*;
// import java.rmi.*;
// import java.io.*;

import jgroup.*;
// import relacs.general.*;
// import relacs.rmi.*;
import debug.Debug;

/**
 *
 *
 * @author  Hein Meling
 * @since   Jgroup 1.1
 */
public final class GroupTable
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /* This class prints debug information for this level */
  private static final int debugLevel = 80;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Table of all groups in the system */
  private HashMap     grouptable;
  /** Table of replicas in the system; not their location though. */
  private HashMap     replicatable;

  private View        view;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public GroupTable()
  {
    grouptable = new HashMap();
    replicatable = new HashMap();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  synchronized public void add(ReplicaData rd)
    throws GroupExistsException
  {
    Integer gid = new Integer(rd.getGid());
    if (replicatable.containsKey(gid)) {
      throw new GroupExistsException();
    }
    /*
     * Keep the replica data for lookup and moving between RM replicas.
     */
    replicatable.put(gid, rd);
    /*
     * Create new membertable for this group and add it to the grouptable.
     * The MemberTable viewChange() method will take care of the addition
     * and removal of members from the membertable. The viewChange() will
     * occur once the replicas have been created and start to join the group.
     * Therefore, we need not add anything to the membertable here.
     */
    MemberTable membertable = new MemberTable();
    grouptable.put(gid, membertable);
  }

  /**
   *
   */
  public void remove(ReplicaData rd)
  {
    remove(rd.getGid());
  }

  /**
   *
   */
  public void remove(ClassData rd)
  {
    remove(rd.getGid());
  }

  /**
   *
   */
  synchronized public void remove(int groupid)
  {
    Integer gid = new Integer(groupid);
    replicatable.remove(gid);
    grouptable.remove(gid);
  }


  /**
   *
   */
  synchronized public void viewChange(View view)
  {
    Integer gid = new Integer(view.getGid()); // Obtain the group identifier
    MemberTable table = null;                 // Member table for this group

    /* Check if the specified replica group already exists. */
    if (grouptable.containsKey(gid)) {
      table = (MemberTable) grouptable.get(gid);
    } else {
      /*
       * Create new membertable for this group and add it to the grouptable.
       * The MemberTable viewChange() method will take care of the addition
       * and removal of members from the membertable. This is done below.
       */
      table = new MemberTable();
      grouptable.put(gid, table);
    }
    // Compute delta views

// We need to determine if we have majority also...
// if we have, we can create new replicas to compensate for
// partitioned members.

// We will likely get many invocations of viewChange() if
// a host crashes, we need to correlate this information.

    // Update the table according to the current view
    table.viewChange(view);
    MemberId[] members = table.members();
    for (int i=0; i < members.length; i++) {
      switch (table.getState(members[i])) {
          case MemberTable.NEWMEMBER:
            /* This is a new member */
            if (Debug.DEBUG)
              Debug.println(debugLevel, "New member " + members[i]);
//              table.put(members[i], new IntList());
            break;
          case MemberTable.RECOVERING:
            /* This member has crashed and then recovered */
            if (Debug.DEBUG)
              Debug.println(debugLevel, "New incarnation " + members[i]);
//              table.put(members[i], new IntList());
            break;
          case MemberTable.SURVIVED:
            /* This member is survived from the previous view */
            if (Debug.DEBUG)
              Debug.println(debugLevel, "Surviving member " + members[i]);
            break;
          case MemberTable.PARTITIONED:
            /* This member has partitioned since the previous view */
            if (Debug.DEBUG)
              Debug.println(debugLevel, "Partitioned member " + members[i]);
            break;
          case MemberTable.MERGING:
            /* This member has merged since the previous view */
            if (Debug.DEBUG)
              Debug.println(debugLevel, "Merging member " + members[i]);
            break;
      }
    }
  }


  /**
   *
   */
/*
  synchronized Status getState(MemberId[] dests, MemberId me)
  {
    ServerData  server;
    Status      status = new Status();
    int         size = servertable.size();

    status.vmids = new VMIDImpl[size];
    status.removed = new IntList[size];
    status.bindings = new BindingList[size][];

    int i=0;
    for (Enumeration e = servertable.elements() ; e.hasMoreElements() ; i++) {
      server = (ServerData) e.nextElement();
      status.vmids[i] = server.vmid;
      status.removed[i] = server.removed;
      status.bindings[i] = server.getState(dests, me);
    }
    return status;
  }

  synchronized void putState(Status status, MemberId[] dests)
    throws RemoteException, ClassNotFoundException
  {
    ServerData      server;
    for (int i=0; i < status.vmids.length; i++) {
      if (Debug.DEBUG)
        Debug.println(60, "putState: status.vmids["+i+"]: " + status.vmids[i]);
      server = (ServerData) get(status.vmids[i]);
      server.putState(servicetable, status.removed[i], status.bindings[i], dests);
    }
  }
*/

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////
/*
  private ServerData get(VMIDImpl vmid)
  {
    ServerData          server;

    server = (ServerData) servertable.get(vmid);
    if (server == null) {
      server = new ServerData(vmid);
      server.viewChange(view);
      servertable.put(vmid, server);
    }
    return server;
  }
*/
}
