/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.exec;

import java.util.*;
import java.rmi.*;
import java.rmi.registry.*;

import jgroup.*;
import relacs.config.*;
import debug.Debug;

/**
 * The LoadCountScheme is an implementation of a DistributionScheme based on
 * counting the number of times a hostname is returned to the ReplicaManager.
 * This is a simple load balancing scheme, designed to have a natural
 * distribution on all hosts in the system. That is replicas will be started
 * on the host with the least number of running replicas. Assuming all hosts
 * have the same computational power, and no other services are running on
 * any of the hosts, this scheme should work fairly well.
 *
 * BUGS: See the BUG FIXME entry within the getHosts() method.
 *
 * @author Hein Meling
 * @since Jgroup 1.1
 */
public class LoadCountScheme implements DistributionScheme {

  /////////////////////////////////////////////////////////////////////////////
  // Class variables
  /////////////////////////////////////////////////////////////////////////////

  private static final long HOST_NOT_AVAILABLE = Long.MAX_VALUE;

  private static ArrayList  domains;

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private Random     rnd;

  /** The ExecDaemon port */
  private int        eport;

  /** Number of domains specified for the distributed system */
  private int        domainCount;

  /** (domain index, host index) -> no. of replicas currently running on host */
  private long[][]   replicaCount;

  /** (domain index) -> no. of hosts in domain (rank) */
  private int[]      domainRank;

  /** (domain index) -> corresponding Domain object */
  private HashMap    domainMap;

  /** (hostname)     -> ExecService remote reference */
  private HashMap    execDaemonMap;

  /** (hostname)     -> Index used to access the replicaCount array */
  private HashMap    hostIndexMap;

  /////////////////////////////////////////////////////////////////////////////
  // Constructor
  /////////////////////////////////////////////////////////////////////////////

  public LoadCountScheme(DistributedSystem ds)
    throws JgroupException
  {
    this(ds, ExecService.EXEC_BOOTSTRAP_PORT);
  }

  public LoadCountScheme(DistributedSystem ds, int eport)
    throws JgroupException
  {
    this.eport = eport;
    rnd = new Random();
    RelacsDS rds;
    if (ds instanceof RelacsDS) {
      rds = (RelacsDS) ds;
    } else {
      rds = new RelacsDS(ds);
    }
    domains = rds.getDomains();

    // Build replicaCount and domainRank tables
    domainCount = domains.size();
    hostIndexMap = new HashMap();
    domainMap = new HashMap();
    execDaemonMap = new HashMap();
    domainRank = new int[domainCount];
    replicaCount = new long[domainCount][];
    for (int i = 0; i < domainCount; i++) {
      // i = domain index
      RelacsDomain d = (RelacsDomain) domains.get(i);
      String[] hosts = d.getNamesWithDomain();
      int hostsInDomain = d.size();
      // The domain rank corrsponds to the size of each domain.
      domainRank[i] = hostsInDomain;
      replicaCount[i] = new long[hostsInDomain];
      // Add this domain to the map, with i as key.
      domainMap.put(new Integer(i), d);
      for (int j = 0; j < hostsInDomain; j++) {
        // j = host index
        try {
          // Add the index value for this host
          hostIndexMap.put(hosts[j], new Index(i, j));
          // Get the ExecDaemon reference
          ExecService edaemon = getExecDaemon(hosts[j], eport);
          // Keep the remote reference
          execDaemonMap.put(hosts[j], edaemon);
          // If the ExecDaemon was found on this host, initialize its element
          // in the replicaCount table; zero replicas running.
          replicaCount[i][j] = 0;
        } catch (Exception e) {
          // RemoteException or NotBoundException
          Debug.println("ExecDaemon not found on: " + hosts[j]);
          // Mark this host as not available.
          replicaCount[i][j] = HOST_NOT_AVAILABLE;
        }
      }
    }
  }


  /**
   *  Check the ExecDaemon map for the specified host.
   */
  private boolean checkExecDaemon(String host)
  {
    try {
      /* Get the current ExecDaemon */
      ExecService es = getExecDaemon(host, eport);
    } catch (Exception e) {
      return false;
    }
    return true;
  }


  /**
   *  Update the ExecDaemon map for the specified host.
   */
  private boolean updateExecDaemonMap(String host)
  {
    ExecService prev_exec = (ExecService) execDaemonMap.get(host);
    try {
      /* Get the current ExecDaemon */
      ExecService curr_exec = getExecDaemon(host, eport);
      if (!prev_exec.equals(curr_exec)) {
        /* Update the map with the current ExecDaemon proxy. */
        execDaemonMap.put(host, curr_exec);
        // Should we update the replicaCount variable??
      }
      return true;
    } catch (Exception e) {
      if (execDaemonMap.containsKey(host)) {
        execDaemonMap.remove(host);
        Index index = (Index) hostIndexMap.get(host);
        replicaCount[index.i][index.j] = HOST_NOT_AVAILABLE;
      }
      return false;
    }
  }


  /**
   *  Get the ExecDaemon for the specified host and port.
   */
  private static ExecService getExecDaemon(String hostname, int port)
    throws RemoteException, NotBoundException
  {
    Registry reg = LocateRegistry.getRegistry(hostname, port);
    return (ExecService) reg.lookup(ExecService.EXEC_DAEMON_NAME);
  }

  /////////////////////////////////////////////////////////////////////////////
  // DistributionScheme interface methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Updates the internal tables based on a list of suspected hosts.
   *
   *  @param suspect_hosts  List of suspected hosts to check for availability.
   *
   */
  public void suspect(ArrayList suspect_hosts)
  {
    for (int i = 0; i < suspect_hosts.size(); i++) {
      updateExecDaemonMap((String) suspect_hosts.get(i));
    }
  }

  /**
   *
   *
   */
  public ExecService getExecRef(String host)
    throws HostException
  {
    if (execDaemonMap.containsKey(host)) {
      return (ExecService) execDaemonMap.get(host);
    } else {
      throw new HostException("Specified host not found in table:" + host);
    }
  }

  /**
   *
   *
   */
  public String getHost()
    throws HostException
  {
    Index index = getLowest();
    if (replicaCount[index.i][index.j] == HOST_NOT_AVAILABLE) {
      throw new HostException("No hosts running the ExecDaemon was found.");
    } else {
      String host = getHostByIndex(index.i, index.j);
      if (host != null) {
        return host;
      } else {
        throw new HostException("No hosts running the ExecDaemon was found.");
      }
    }
  }

  /**
   * Returns a list of available hosts with the least replicas running.
   * The number of hosts returned matches the specified redundancy parameter.
   * If the redundancy could not be satisfied an exception is thrown.
   *
   */
//  public String[] getHosts(int redundancy, int minimal)
  public String[] getHosts(int redundancy)
    throws RedundancyException
  {
    /*
     * Check if the specified redundancy can be satisfied;
     * otherwise throw exception
     */
    if (execDaemonMap.size() < redundancy) {
      throw new RedundancyException(
        "\nNot enough hosts available to satisfy the specified redundancy ("+ redundancy +")\n" +
        "Multiple replicas of the same type cannot be assigned to the same host.\n" +
        "Check if enough ExecDaemons are running. Found only: "+ execDaemonMap.size());
    }
    
    /* Sets of selected hosts within each domain; to avoid duplicate assignments */
    HashMap selected = new HashMap(domainCount);
    /* Initialize the set of selected hosts; empty sets for each domain. */
    for (int i = 0; i < domainCount; i++) {
      selected.put(new Integer(i), new HashSet());
    }
    
    /* List of unavailable domains (or domain in which all hosts have been exhausted.) */
    ArrayList unavailDomains = new ArrayList(domainCount);

    /* Sorted map: (host ranking) -> set of hosts */
    TreeMap hostRanking = new TreeMap();

    /* Counts the hosts we have selected so far. */
    int hostsSelected = 0;

    /* Loop until redundancy satisfied. */
    while (redundancy - hostsSelected > 0) {

      /* Check if all domains has become unavailable; throw exception */
      if (unavailDomains.size() == domainCount) {
        throw new RedundancyException(
          "No hosts available.\nCheck if the ExecDaemons are running.");
      }

      /*
       * First try to find one host in each available domain.
       * Next (if redundancy not satisfied) find hosts in already
       * used domains.
       *
       * Note that we always iterate all domains, since we want to find
       * the best domain in case (redundancy > domainCount).
       * The selection of hosts are made below the while loop.
       */
      for (int i = 0, j = 0; i < domainCount; /* No index increment */) {

        /* We skip all unavailable domains found in the previous iteration */
        if (unavailDomains.contains(new Integer(i))) {
          i++;
        } else {
          try {
            /*
             * Get the host with the lowest replica count in this domain,
             * excluding already selected hosts.
             */
            j = getLowestInDomain(i, (Set) selected.get(new Integer(i)));
          } catch (HostException he) {
            /*
             * All hosts in the i'th domain has already been assigned;
             * we move on to the next domain. (continue with the for loop.)
             * Before moving on, we mark this domain as unavailable for
             * further replicas of this type, since we have exhausted all
             * hosts in this domain.
             */
            unavailDomains.add(new Integer(i));
            i++;
            continue;
          }
          if (replicaCount[i][j] == HOST_NOT_AVAILABLE) {
            Debug.println("No hosts available in: " + getDomainByIndex(i));
            unavailDomains.add(new Integer(i));
            /*
             * Increment domain index; needed since we want to continue to look
             * for new hosts if the currently selected has become unavailable.
             */
            i++;
          } else {

            // BUG FIXME: Since getHostByIndex() increments the replicaCount
            // it will become inconsistent when the for iteration selects
            // more hosts than the required redundancy. This is the case when
            // domainCount > redundancy and possibly some other cases.
            // Recall that we want to select hosts from all domains in the
            // first iteration, and therefore we should no break out of the
            // for loop if the redundancy is satisfied early. We should instead
            // consider when to increment the replicaCount array entries.

            String host = getHostByIndex(i, j);
            if (host != null) {
              /* Get the replica count on this host */
              Long thisReplicaCount = new Long(replicaCount[i][j]);
              /* Set of hosts with the same replica count as this host. */
              Set hostSet = null;
              if (hostRanking.containsKey(thisReplicaCount)) {
                /* Obtain the set of hosts with identical replica count as this host. */
                hostSet = (Set) hostRanking.get(thisReplicaCount);
              } else {
                /*
                 * No entry with the same replica count was found, insert new
                 * empty set. This empty set will be updated with this host below.
                 */
                hostSet = new HashSet();
                hostRanking.put(thisReplicaCount, hostSet);
              }
              /*
               * We add this host to the set of hosts with the same replica count
               * as this host. This addition will also be seen from the hostRanking
               * TreeMap, since we always refer to the same set.
               */
              hostSet.add(host);
              /* Increment the number of selected hosts. */
              hostsSelected++;

               
//               Iterator iter = hostRanking.values().iterator();
//               while (iter.hasNext()) {
//                 hostSet = (Set) iter.next();
//                 Iterator it = hostSet.iterator();
//                 while (it.hasNext()) {
//                   Debug.println("host: " + (String) it.next());
//                 }
//               }

              /* Add this host index to the set of hosts already assigned for this domain */
              ((Set) selected.get(new Integer(i))).add(new Integer(j));
              /*
               * Increment domain index; needed since we want to continue to look
               * for new hosts if the currently selected has become unavailable
               */
              i++;
            } else {
              /*
               * We go back and find another lowest ranking in this domain until
               * getHostByIndex() returns a non-null host or the domain is exhausted.
               * Note that this means that the ExecDaemon on the selected host has
               * crashed or otherwise is unavailable. The getHostByIndex() will mark
               * it as not available.
               *
               * Note that we don't increment the domain index here. This will not
               * cause an eternal loop, however, since sooner or later the call to
               * getLowestInDomain() will return an index to a HOST_NOT_AVAILABLE,
               * and will then add that domain index to the set of unavailable
               * domains.
               */
              continue;
            }
          }
        } // end of if unavailable domains contains domain index i
      } // end of for each domain
    } // end of while redundancy not satisfied

    /* Prepare hosts array for return. */
    String[] hosts = new String[redundancy];
//    Debug.println("Selected hosts:");
//     for (int r = 0; r < redundancy; r++) {
//      hosts[r] = (String) hostRanking.remove(hostRanking.firstKey());
    int r = redundancy;
    while (r > 0) {
      Set hostSet = null;
      try {
        hostSet = (Set) hostRanking.remove(hostRanking.firstKey());
      } catch (Exception e) {
        /* No such element */
      }
      Iterator iter = hostSet.iterator();
      while (iter.hasNext() && r > 0) {
        hosts[r-1] = (String) iter.next();
//        Debug.println("  " + hosts[r-1]);
        r--;
      }
    }
    return hosts;
  }


  /**
   * Provides two sets of hosts corresponding to <code>redundancyA<code>
   * and <code>redundancyB</code>, preferably non-overlapping hosts will
   * be returned. If overlapping hosts cannot be avoided an exception
   * will be thrown.
   *
   */
  public String[] getHosts(int redundancyA, int redundancyB) {
    return new String[] { "Not yet implemented" };
  }

  /////////////////////////////////////////////////////////////////////////////
  // Private methods
  /////////////////////////////////////////////////////////////////////////////

  private void outputCount()
  {
    Debug.println("\nreplicaCount output for each host:\n");
    for (int i = 0; i < domainCount; i++) {
      for (int j = 0; j < domainRank[i]; j++) {
        Debug.println(getHostName(i, j) + ": " + replicaCount[i][j]);
      }  
    }
    Debug.println("");
  }

  private String getHostByIndex(int i, int j)
  {
    String host = getHostName(i, j);

    /* Check if the host is running the ExecDaemon. */
    if (checkExecDaemon(host)) {
      /* Update the number of replicas assigned to this host */
      replicaCount[i][j] += 1;
      return host;
    } else {
      Debug.println(host + " not available. ExecDaemon not running.");
      /* Mark host as no longer available */
      replicaCount[i][j] = HOST_NOT_AVAILABLE;
      /* Update the ExecDaemon map since the host is no longer available */
      if (execDaemonMap.containsKey(host)) {
        execDaemonMap.remove(host);
      }
      return null;
    }
  }

  private String getDomainByIndex(int i) {
    RelacsDomain d = (RelacsDomain) domainMap.get(new Integer(i));
    return d.getDomainName();
  }

  private String getHostName(int i, int j)
  {
    /* Get the j'th host name with domain name for the i'th domain */
    return (((RelacsDomain) domainMap.get(new Integer(i))).getNamesWithDomain())[j];
  }

  private int getLowestInDomain(int i, Set assignedHosts)
    throws HostException
  {
    int tj = -1;
    long low = HOST_NOT_AVAILABLE;
    for (int j = 0; j < domainRank[i]; j++) {
      if (!assignedHosts.contains(new Integer(j))) {
        if (low > replicaCount[i][j]) {
          low = replicaCount[i][j];
          tj = j;
        }
      }
    }
    if (tj == -1) {
      /* Cannot return an already assigned host. */
      throw new HostException("No more hosts available in this domain.");
    }
//    Debug.println(getHostName(i, tj) +": i="+ i +", tj="+ tj +", count="+ replicaCount[i][tj]);
    return tj;
  }

  private Index getLowest() {
    int ti = 0, tj = 0;
    long low = replicaCount[ti][tj];
    for (int i = 0; i < domainCount; i++) {
      for (int j = 0; j < domainRank[i]; j++) {
        if (low > replicaCount[i][j]) {
          low = replicaCount[i][j];
          ti = i;
          tj = j;
        }
      }
    }
    return new Index(ti, tj);
  }

  private class Index {
    public int i, j;

    public Index(int i, int j) {
      this.i = i;
      this.j = j;
    }

  } // END private class Index


  /////////////////////////////////////////////////////////////////////////////
  // Main method
  /////////////////////////////////////////////////////////////////////////////

  /**
   * For testing purposes only. Invoked using script dscheme.
   *
   */
  public static void main (String[] args) {
    // No argument checking, for now.
    DistributedSystem ds = new DistributedSystem(args[0]);
    LoadCountScheme lcs = null;
    try {
      lcs = new LoadCountScheme(ds);
    } catch (JgroupException je) {
      System.err.println(je);
      System.exit(1);
    }
    try {
//       System.out.println("1: " + lcs.getHost());
      String[] h = lcs.getHosts(5);
      for (int c = 0; c < 5; c++) {
        System.out.println("5: " + h[c]);
      }
//       System.out.println("1: " + lcs.getHost());
      String[] h2 = lcs.getHosts(2);
      for (int c = 0; c < 2; c++) {
        System.out.println("2: " + h2[c]);
      }
      String[] h3 = lcs.getHosts(3);
      for (int c = 0; c < 3; c++) {
        System.out.println("3: " + h3[c]);
      }
//       System.out.println("1: " + lcs.getHost());
      String[] h1 = lcs.getHosts(7);
      for (int c = 0; c < 7; c++) {
        System.out.println("7: " + h1[c]);
      }
      String[] h6 = lcs.getHosts(6);
      for (int c = 0; c < 6; c++) {
        System.out.println("6: " + h6[c]);
      }
//       System.out.println("1: " + lcs.getHost());
      lcs.outputCount();
//     } catch (HostException he) {
//       Debug.println(he);
    } catch (RedundancyException re) {
      Debug.println(re);
    }
  } // end of main ()
  
} // END LoadCountScheme
