/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.exec;

import java.util.*;
import jgroup.*;
import relacs.config.*;
import debug.Debug;

/**
 *  ReplicaCount class is used to compute replica allocation on domains/hosts
 *  in the distributed system.
 *
 *  @author  Hein Meling
 *  @since   Jgroup 1.2
 */
public class ReplicaCount
{
  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private HashMap   domaintable;
  private HashMap   hostcount;

  /////////////////////////////////////////////////////////////////////////////
  // Constructor
  /////////////////////////////////////////////////////////////////////////////

  ReplicaCount()
  {
    domaintable = new HashMap();
    hostcount   = new HashMap();
  }

  ReplicaCount(ArrayList domains)
  {
    domaintable = new HashMap();
    hostcount   = new HashMap();
    for (int i = 0; i < domains.size(); i++) {
      RelacsDomain domain = (RelacsDomain) domains.get(i);
      String[] hostArray = domain.getNamesWithDomain();
      Set hosts = new HashSet(hostArray.length);
      Long zeroCount = new Long(0);
      for (int j = 0; j < hostArray.length; j++) {
        hostcount.put(hostArray[j], zeroCount);
        hosts.add(hostArray[j]);
      }
      TreeMap replicaCount = new TreeMap();
      replicaCount.put(zeroCount, hosts);
      domaintable.put(domain, replicaCount);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Methods
  /////////////////////////////////////////////////////////////////////////////

  // Use these to mark hosts/domains (un)available. */
  synchronized public void addHost() {}
  synchronized public void addDomain() {}
  synchronized public void removeHost() {}
  synchronized public void removeDomain() {}

  /**
   *  @param domain  Domain from which to get a host.
   *  @param used    A set of used hosts in this domain. It is the responsibility
   *                 of the caller to determine which hosts are already used for
   *                 this invocation series.
   *  @return        An availble host in domain, and not in the used set. If no
   *                 more hosts are available in this domain, <code>null</code>
   *                 is returned.
   */
  synchronized public String getHost(RelacsDomain domain, Set used)
  {
    String host = null;
    if (used == null) {
      used = new HashSet();
    }
    if (domaintable.containsKey(domain)) {
      TreeMap replicaCount = (TreeMap) domaintable.get(domain);
      Long thisReplicaCount = (Long) replicaCount.firstKey();
      Set hosts = (Set) replicaCount.get(thisReplicaCount);
      Iterator iter = hosts.iterator();
      boolean success = false;
      while (iter.hasNext() && !success) {
        host = (String) iter.next();
        if (used.contains(host))
          continue; // while loop
        else
          success = true;

        /* Remove this host from the current replicaCount table */
        hosts.remove(host);
        if (hosts.isEmpty()) {
          /* If the set of hosts is empty we must remove the entry from the
           * replicaCount table which is stored in the domaintable. */
          replicaCount.remove(thisReplicaCount);
        }
        /* Increment the replica count value */
        Long rcount = new Long(thisReplicaCount.longValue() + 1);
        if (replicaCount.containsKey(rcount)) {
          /* Add this host to an existing set of hosts with the new
           * replica count value. */
          ((Set) replicaCount.get(rcount)).add(host);
        } else {
          /* There is no other hosts with matching replica count value,
           * we add our host to a new set, which we put in the sorted map. */
          hosts = new HashSet();
          hosts.add(host);
          replicaCount.put(rcount, hosts);
        }
        /* Update the count for this host in the hostcount table also.
         * This should overwrite any existing count stored for this host. */
        hostcount.put(host, rcount);
      }
      /* If success return host, otherwise return null. */
      return (success ? host : null);
    } else {
      /* The domain is not in the domaintable; non-existant or unavailable */
      return null;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Main method
  /////////////////////////////////////////////////////////////////////////////

  /**
   * For testing purposes only. Invoked using script rcount.
   */
  public static void main (String[] args)
    throws JgroupException
  {
    // No argument checking, for now.
    DistributedSystem ds = new DistributedSystem(args[0]);
    RelacsDS rds = new RelacsDS(ds);
    ArrayList domains = rds.getDomains();
    ReplicaCount rc = new ReplicaCount(domains);
    Set hosts = new HashSet(); // Could set the size to redundancy
    String host;
    for (int i = 0; i < domains.size(); i++) {
      host = rc.getHost((RelacsDomain) domains.get(i), null);
      hosts.add(host);
      Debug.println(host);
    }
    for (int i = 0; i < domains.size(); i++) {
      host = rc.getHost((RelacsDomain) domains.get(i), hosts);
      hosts.add(host);
      Debug.println(host);
    }
    for (int i = 0; i < domains.size(); i++) {
      host = rc.getHost((RelacsDomain) domains.get(i), hosts);
      hosts.add(host);
      Debug.println(host);
    }
    for (int i = 0; i < domains.size(); i++) {
      host = rc.getHost((RelacsDomain) domains.get(i), hosts);
      hosts.add(host);
      Debug.println(host);
    }
  }

} // End of ReplicaCount
