/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.exec;

/**
 * ReplicaData specifies the class, arguments, redundancy and replication style
 * of a replica.
 *
 * Note: The replication style constants are subject to change.
 *
 * @author      Hein Meling
 * @since       Jgroup 1.1
 */
public class ReplicaData extends ClassData implements ReplicationStyle {

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private int       replicationStyle;
  private int       initialRedundancy;
  private int       minimalRedundancy;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  public ReplicaData(ClassData classData, int replicationStyle,
                     int initialRedundancy, int minimalRedundancy)
    throws ReplicationException, ClassNotFoundException
  {
    super(classData.getClassName(), classData.getClassArgs());
    setReplicationStyle(replicationStyle);
    this.initialRedundancy = initialRedundancy;
    this.minimalRedundancy = minimalRedundancy;
  }

  /////////////////////////////////////////////////////////////////////////////
  // Access methods
  /////////////////////////////////////////////////////////////////////////////

  public int getReplicationStyle()
  {
    return replicationStyle;
  }

  public int getInitialRedundancy()
  {
    return initialRedundancy;
  }

  public int getMinimalRedundancy()
  {
    return minimalRedundancy;
  }

  /////////////////////////////////////////////////////////////////////////////
  // Update methods
  /////////////////////////////////////////////////////////////////////////////

  public void setReplicationStyle(int replicationStyle)
    throws ReplicationException
  {
    if (replicationStyle < ReplicationStyle.LOWER_BOUND ||
        replicationStyle > ReplicationStyle.UPPER_BOUND) {
      throw new ReplicationException("Unsupported replication style.");
    } else {
      this.replicationStyle = replicationStyle;
    }
  }

  public void setInitialRedundancy(int initialRedundancy)
    throws ReplicationException
  {
    if (initialRedundancy > ReplicationStyle.REDUNDANCY_LOWER_BOUND &&
        initialRedundancy < ReplicationStyle.REDUNDANCY_UPPER_BOUND) {
      this.initialRedundancy = initialRedundancy;
    } else if (initialRedundancy > ReplicationStyle.REDUNDANCY_UPPER_BOUND) {
      throw new ReplicationException("Initial redundancy is too high.");
    } else {
      throw new ReplicationException("Initial redundancy is too low.");
    }
  }

  public void setMinimalRedundancy(int minimalRedundancy)
    throws ReplicationException
  {
    if (minimalRedundancy > ReplicationStyle.REDUNDANCY_LOWER_BOUND &&
        minimalRedundancy < ReplicationStyle.REDUNDANCY_UPPER_BOUND) {
      this.minimalRedundancy = minimalRedundancy;
    } else if (minimalRedundancy > ReplicationStyle.REDUNDANCY_UPPER_BOUND) {
      throw new ReplicationException("Minimal redundancy is too high.");
    } else {
      throw new ReplicationException("Minimal redundancy is too low.");
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Object methods; overriding from the Object class
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    return "[" + super.toString()
      + "   " + ReplicationStyleNames[replicationStyle]
      + " I=" + initialRedundancy
      + " M=" + minimalRedundancy + "]";
  }

/*
  public static void main (String[] argv) {
    String[] args = new String[] { "-port", "2323" };
    String[] args2 = new String[] { "-port", "2323" };
    try {
      ReplicaData rd1 = new ReplicaData("jgroup.exec.ExecDaemon", args,
                                        ReplicationStyle.STATELESS, 3, 2);
      ReplicaData rd2 = new ReplicaData("jgroup.exec.ExecDaemon", args,
                                        ReplicationStyle.ACTIVE, 3, 1);
      ReplicaData rd3 = new ReplicaData("jgroup.exec.ExecDaemon", args,
                                        ReplicationStyle.STATELESS, 3, 2);
      ReplicaData rd6 = new ReplicaData("jgroup.exec.ExecDaemon", args2,
                                        ReplicationStyle.STATELESS, 3, 2);

      ReplicaData rd4 = new ReplicaData("jgroup.exec.ExecService", args,
                                        ReplicationStyle.STATELESS, 3, 2);
      ReplicaData rd5 = new ReplicaData("jgroup.exec.ExecService", args2,
                                        ReplicationStyle.STATELESS, 3, 2);

      System.out.println("rd1.equals(rd1): " + rd1.equals(rd1)); // true
      System.out.println("rd1.equals(rd2): " + rd1.equals(rd2)); // true
      System.out.println("rd1.equals(rd3): " + rd1.equals(rd3)); // true
      System.out.println("rd1.equals(rd4): " + rd1.equals(rd4)); // false
      System.out.println("rd1.equals(rd5): " + rd1.equals(rd5)); // false
      System.out.println("rd4.equals(rd5): " + rd4.equals(rd5)); // false
      System.out.println("rd1.hashCode(): " + rd1.hashCode());
      System.out.println("rd2.hashCode(): " + rd2.hashCode());
      System.out.println("rd3.hashCode(): " + rd3.hashCode());
      System.out.println("rd4.hashCode(): " + rd4.hashCode());
      System.out.println("rd5.hashCode(): " + rd5.hashCode());
      System.out.println("rd6.hashCode(): " + rd6.hashCode());
    } catch (ClassNotFoundException cnf) {
      System.err.println("ClassNotFoundException: " + cnf);
    } catch (ReplicationException re) {
      System.err.println("ReplicationExceptino: " + re);
    }
  }
*/  
}
