/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
package jgroup.exec;

import java.util.*;
import java.rmi.*;

import debug.Debug;
import jgroup.*;
import jgroup.registry.*;
import relacs.config.*;
import relacs.rmi.registry.*;

/*
 * 
 *
 * @author      Hein Meling
 * @since       Jgroup 1.1
 */

public class ReplicaManagerImpl
implements ReplicationManager, MembershipListener
//implements ReplicationManager, MergingListener, MembershipListener
{
  /////////////////////////////////////////////////////////////////////////////
  // Constants
  /////////////////////////////////////////////////////////////////////////////

  /* This class prints debug information for this level */
  private static final int debugLevel = 80;

  /* The default number of retries to find the dependable registry */
  private static final int DEFAULT_BIND_RETRIES = 6;

  /* The default number of retries to create replicas */
  private static final int DEFAULT_RETRIES = 2;

  /////////////////////////////////////////////////////////////////////////////
  // Static fields
  /////////////////////////////////////////////////////////////////////////////

  private static Object[]    configuration;

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private DistributionScheme dScheme;
  private RelacsDS           rds;
  private MembershipService  membership;
  private ExternalGMIService egmi;

  private GroupTable         grouptable;
  // tmporary to test removeReplica()
  private HashMap            execMap;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Initialize the Replication Manger according to the specified configURL
   *  and corresponding ports.
   *
   *  Note that the ports should prehaps be specified in the config file.
   *
   */
  public ReplicaManagerImpl(String configURL, int eport, int jport)
    throws JgroupException
  {
    execMap = new HashMap();
    grouptable = new GroupTable();
    rds = new RelacsDS(new DistributedSystem(configURL));
    dScheme = (DistributionScheme) new LoadCountScheme(rds, eport);
    /* Prepare the configuration objects */
    configuration = new Object[] { rds, new TransportConfig(jport) };
  }

  /////////////////////////////////////////////////////////////////////////////
  // Bootstrap method
  /////////////////////////////////////////////////////////////////////////////

  private void bootstrap(String configURL, int eport)
  {
    int initial = rds.getRegConfig().getInitial();
    int minimal = rds.getRegConfig().getMinimal();

    // Arguments for the replication manager
    String[] rmArgs = new String[] { "-eport", new Integer(eport).toString(), configURL };
    // Arguments for the HelloServer
    String[] hlArgs = new String[] { configURL };

    ClassData hlc = null;

    try {

      ClassData rmc = new ClassData(this.getClass(), rmArgs);
      createGroup(new ReplicaData(rmc, ReplicationStyle.ACTIVE, initial, minimal));

      hlc = new ClassData("test.hello.HelloServer", hlArgs);
      createGroup(new ReplicaData(hlc, ReplicationStyle.ACTIVE, 3, 2));

    } catch (RemoteException re) {
      System.err.println("Invoking createGroup() failed: " + re.getMessage());
      return;
    } catch (ReplicationException re) {
      System.err.println("Failed to create replica object: " + re.getMessage());
      return;
    } catch (ClassNotFoundException cnf) {
      System.err.println("Failed to create replica object: " + cnf.getMessage());
      return;
    } catch (RedundancyException rns) {
      System.err.println("Failed to create replica group: " + rns.getMessage());
      return;
    }
/*
    try {
      Thread.sleep(30000);
    } catch (InterruptedException ie) { }
    String[] hosts = (String[]) execMap.get(hlc);
    try {
      ExecService exec = dScheme.getExecRef(hosts[0]);
      exec.removeReplica(hlc.getGid());
      System.out.println(hlc.getClassName() +" replica removed on "+ hosts[0]);
    } catch (HostException he) {
      System.err.println("ExecDaemon not available on: "+ hosts[0]);
      Debug.println(debugLevel, "HostException: " + he);
    } catch (RemoteException re) {
      System.err.println("Failed to remove replica "+ hlc.getClassName() +" on: "+ hosts[0]);
      Debug.println(debugLevel, "RemoteException: " + re);
    } catch (ExecException exece) {
      Debug.println(debugLevel, exece.getMessage());
      Throwable e = exece.getException();
      if (e != null) {
        Debug.println(debugLevel, e.getMessage());
      }
    }
*/      
  }


  /**
   *  Bind the replication manager to the dependable registry and join
   *  the ReplicationManager group.
   *
   */
  private void bindReplicaManager(int gid, int rport)
    throws JgroupException, RemoteException
  {
    /*
     * Obtain a dependable registry reference
     */
    DependableRegistry dr = RegistryFactory.getRegistry(rds, rport);
    /*
     * Obtain the group manager
     */
    Object[] listeners = new Object[] { this };
    GroupManager gm = GroupManager.getGroupManager(configuration, listeners);
    /*
     * Obtain the service proxies
     */
    membership = (MembershipService)  gm.getService(MembershipService.class);
    egmi       = (ExternalGMIService) gm.getService(ExternalGMIService.class);

    /*
     * Join the RM group and bind the RM in the dependable registry
     */
    membership.join(gid);
    Debug.println(debugLevel, "Joined groupID: " + gid);
    int retries = DEFAULT_BIND_RETRIES;
    IID iid = null;
    /*
     * Retry binding to the registry as long as no reference has been found or
     * until the number of retires have exceeded the specified number of retries.
     */
    while (iid == null && retries > 0) {
      try {
        iid = egmi.bind(REPLICATION_MANAGER_NAME, dr);
      } catch (AccessException e) {
        Debug.println(debugLevel, "AccessException: Operation not permitted.");
        retries = 0;
      } catch (RemoteException e) {
        Debug.println(debugLevel, "Retry: " + retries);
        // Maybe we should sleep a little before each retry?
        retries--;
      }
    }

    if (iid != null) {
      Debug.println(debugLevel, REPLICATION_MANAGER_NAME + " bound to the dependable registry: " + iid);
    } else {
      Debug.println(debugLevel, "Couldn't bind " + REPLICATION_MANAGER_NAME + " to the dependable registry");
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // ReplicationManager Internal interface methods
  /////////////////////////////////////////////////////////////////////////////

  /*
   * We should provide the following methods, in order to make sure that
   * the internal tables are up-to-date:
   *  getHosts()
   *  suspect()
   *
   */

  /////////////////////////////////////////////////////////////////////////////
  // ReplicationManager External interface methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Create a group using the information contained in <code>ReplicaData</code>.
   *  This method will be invoked using the anycast semantics. However, the
   *  changes made to the internal tables of the <code>DistributionScheme</code>
   *  will be propogated to the other Replication Manager replicas through
   *  the internal invocation <code>NONAME</code>.
   *
   *  @param   rd   Replica data specifying details need to create the replica group.
   *  @return  gid  Group identifier
   *
   */
  public int createGroup(ReplicaData rdata)
    throws RemoteException, RedundancyException
  {
    /** List of hosts who failed while creating group. */
    ArrayList failedCreation = new ArrayList();

    /* Throws an exception if getHosts() couldn't satisfy the redundancy. */
    String[] hosts = dScheme.getHosts(rdata.getInitialRedundancy());

    ClassData cdata = rdata.getClassData();

    execMap.put(cdata, hosts);

    /* Create empty set; used for storing ExecDaemon references. */
//    Set execSet = new HashSet();
//    int gid = rdata.getGid();
//    Integer Gid = new Integer(gid);

    /* Set to true to enter loop. */
    boolean retry = true;

    /* If we failed to create a replica on some hosts, retry a few times. */
    for (int retryCount = DEFAULT_RETRIES; retry && retryCount > 0; retryCount--) {

      /* Cycle through the host list and create replicas using the ExecDaemon */
      for (int i = 0; i < hosts.length; i++) {
        if (!createReplica(cdata, hosts[i])) {
          failedCreation.add(hosts[i]);
        }
      }

      /* Check if all replicas was created correctly. */
      if (failedCreation.size() > 0) {
        // Update the ExecDaemon map in the distribution scheme.
        dScheme.suspect(failedCreation);
        // Get remaining hosts; throws an exception if all become unavailable.
        hosts = dScheme.getHosts(failedCreation.size());
      } else {
        /* All replicas created, exit for loop. */
        retry = false;
      }
    } // end of for (retryCount...)

    try {
      /* If we get here, all should be OK, add the replica data for lookup. */
      grouptable.add(rdata);
    } catch (GroupExistsException ge) {
      // What to do?
    }
    return rdata.getGid();
  }


  public boolean createReplica(ClassData cdata, String host)
  {
    try {
      ExecService exec = dScheme.getExecRef(host);
      exec.createReplica(cdata);
      System.out.println(cdata.getClassName() +" replica successfully created on "+ host);
      return true;
    } catch (HostException he) {
      System.err.println("ExecDaemon not available on: "+ host);
      if (Debug.DEBUG)
        Debug.println(debugLevel, "HostException: " + he);
      return false;
    } catch (RemoteException re) {
      System.err.println("Failed to create replica "+ cdata.getClassName() +" on: "+ host);
      if (Debug.DEBUG)
        Debug.println(debugLevel, "RemoteException: " + re);
      return false;
    } catch (ExecException exec) {
      if (Debug.DEBUG)
        Debug.println(debugLevel, exec.getMessage());
      Throwable e = exec.getException();
      if (e != null) {
        if (Debug.DEBUG)
          Debug.println(debugLevel, e.getMessage());
      }
      return false;
    }
  }

  public boolean notifyPresence(String host, ExecService ref)
    throws RemoteException, McastRemoteException
  {
    Debug.println(debugLevel, "Received new host: " + host);
    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  // MembershipListener interface methods
  /////////////////////////////////////////////////////////////////////////////


  /*
   *
   */
  public void viewChange(View view)
  {
    Debug.println(debugLevel, "View: " + view);
    grouptable.viewChange(view);
  }

  /*
   *  Comments inherited from jgroup.MergingListener
   */
  public void leaved()
  {
  }


  /*
   *  Comments inherited from jgroup.MembershipListener
   */
  public void prepareChange()
  {
  }

  /////////////////////////////////////////////////////////////////////////////
  // test of init_replica()
  /////////////////////////////////////////////////////////////////////////////

  private int init_replica(Object obj) {
    Class c = obj.getClass();
    String classname = c.getName();
    Debug.println("name: " + classname);
    return 1;
//     try {
//       Method main = c.getDeclaredMethod("main", new Class[] {String[].class});
//       int m = main.getModifiers();
//       if (main == null || !Modifier.isPublic(m) || !Modifier.isStatic(m)) {
//         // Change the Exception type here...
//         throw new ExecException("Class " + c.getName() + " does not implement"
//                                 + " a 'public static void main()' method.");
//       } else {
//         ;
//       }
//     }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Main method
  /////////////////////////////////////////////////////////////////////////////

  public static void main(String[] argv)
    throws JgroupException, RemoteException
  {
    // Default values
    int     eport         = ExecService.EXEC_BOOTSTRAP_PORT;
    int     rport         = 0;
    int     jport         = 0;
    boolean bootstrapping = false;
    String  configURL     = null;

    for (int i = 0; i < argv.length; i++) {
      if (argv[i].equals("-eport")) {
        if ((i + 1) == argv.length) {
          System.err.println("Options -eport must be followed by a port number");
          usage();
          return;
        }
        try {
          eport = Integer.parseInt(argv[++i]);
        } catch (Exception e) {
          System.err.println("Value " + argv[i] + " is not a number");
          usage();
          return;
        }
      } else if (argv[i].equals("-bootstrap")) {
        bootstrapping = true;
      } else if (argv[i].startsWith("-")) {
        System.err.println("Unknown option " + argv[i]);
        usage();
      } else {
        if (configURL != null) {
          System.err.println("Too many options specified");
          usage();
          return;
        }
        configURL = argv[i];
      }
    }
    if (configURL == null) {
      System.err.println("You must specify a configuration file");
      usage();
      return;
    }

    Debug.setOutLevel(debugLevel);
    /* Create a ReplicaManagerImpl instance */
    ReplicaManagerImpl rm = new ReplicaManagerImpl(configURL, eport, jport);
    if (bootstrapping) {
      rm.bootstrap(configURL, eport);
    } else {
      try {
        /* Create a DependableRegistry in the same VM as the ReplicaManager */
        DependableRegistry registry = new RegistryImpl(configuration, rport);
        int gid = ClassData.computeGid(rm, argv);

//        Debug.println("gid: " + gid);
//        init_replica(this);

        rm.bindReplicaManager(gid, rport);
        while (true) {
          try {
            Thread.sleep(0x7fffffffffffffffL);
          } catch(InterruptedException ex) {}
        }
      } catch (Exception e) {
        Debug.println("ReplicaManagerImpl.main: an exception occurred: " + e.getMessage());
        e.printStackTrace();
      }
    }
  }

  private static void usage()
  {
    System.out.println("Usage: replicamanager <configURL> [-bootstrap] [-eport <port number>]");
  }

}
