/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.multicast;

import jgroup.*;

/**
 * This interface enables members to receive ack messages in response
 * to multicasts. When a message is delivered, upcall methods
 * <t>deliverStream</t> and <t>deliverObject</t> of interface
 * <t>MulticastListener</t> are invoked. <t>MulticastListener</t>
 * objects may respond to these invocations by returning an
 * object, that will be delivered to the multicast sender
 * through the <t>AckListener</t> interface.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.8
 */
public interface AckListener
{

  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * about the view in which the multicast message is
   * delivered. Note that following the Jgroup specification,
   * a message may be delivered in a view different from the
   * one in which the message has been multicast; this is
   * motivated by the fact that during agreement protocols,
   * no new messages can be delivered; in fact, cascading
   * sequences of delivered messages could prevent the
   * completion of agreement protocols.
   *
   * @param view the view in which the multicast message is
   *   delivered
   */
  public void notifyView(View view);


  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * that the member identified by <tt>id</tt> has responded
   * with <tt>obj</tt> to the multicast message associated with
   * this ack listener. <p>
   *
   * Member <tt>id</tt> is contained in the view notified by
   * <tt>notifyView</tt>.
   *
   * @param id the identifier of the member responding to the
   *   multicast message
   * @param pos position of the member in the last installed view
   * @param obj the response message
   */
  public void ack(MemberId id, int pos, Object obj);


  /**
   * Upcalls invoked by Jgroup to notify the ack listener
   * that a new view has been installed after the view
   * in which the message has been delivered. After this
   * event, no missing response messages for the multicast
   * message associated to this ack listener will ever
   * be received.
   */
  public void viewChange();


} // END AckListener
