/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.multicast;

import java.io.*;

import jgroup.*;

/**
 * The facilities provided by the reliable multicast service included
 * in Jgroup may be accessed using the <tt>MulticastService</tt>
 * interface. Methods in this interfaces enable members of a group
 * to communicate among themselves by multicasting messages or
 * complex objects. <p>
 *
 * Instances of <tt>MulticastService</tt> may be obtained through
 * the <tt>GroupManager</tt> class.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.8
 */
public interface    MulticastService
{

  /**
   * Multicast a message enclosed in <tt>stream</tt> to the members
   * of the current view of the group. <p>
   *
   * The invoker may specify an <tt>AckListener</tt> object, which
   * will be notified when objects returned by members delivering
   * this message are received.
   *
   * @param stream  the message
   * @param ackl    ack listener to be notified; if null, no ack listener
   *                is notified.
   *
   * @exception UnsupportedOperationException
   *            Raised if the member associated with this group manager
   *            is not a MulticastListener, and thus is not able to receive
   *            message deliveries.
   * @exception JgroupException
   *            if the group manager receives a request to multicast a
   *            message after having joined a group, but before having
   *            received the first view for it; or after having
   *            requested to leave a group.
   */
  public void mcast(OutputStream out, AckListener ackl)
    throws JgroupException;


  /**
   * Multicast a serializable object to the members of the current view of
   * a group.<p>
   *
   * The invoker may specify an <tt>AckListener</tt> object, which
   * will be notified when objects returned by members delivering
   * this message are received.
   *
   * @param obj         the message that must be multicast
   * @param ackl        ack listener to be notified; if null, no ack listener
   *                    is notified
   *
   * @exception UnsupportedOperationException
   *            Raised if the member associated with this group manager
   *            is not a MulticastListener, and thus is not able to receive
   *            message deliveries.
   * @exception JgroupException
   *            if the group manager receives a request to multicast a
   *            message after having joined a group, but before having
   *            received the first view for it; or after having
   *            requested to leave a group.
   * @exception IOException
   *            Raised in the case of serialization problems
   */
  public void mcast(Object obj, AckListener ackl)
    throws JgroupException, IOException;


  /**
   *  Returns an output stream that can be used to encode the message.
   *  Method <tt>mcast(OutputStream, AckListener)</tt> accepts only
   *  output streams obtained with this method.
   */
  public OutputStream getMessage();


} // END MulticastService

