/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.reference;

import java.io.*;
import java.rmi.*;
import java.rmi.server.*;

import jgroup.*;
import jgroup.registry.*;


/**
 * This interface will be moved to another package. It is used internally to retrieve
 * registry proxies and create registry replicas.
 * @author      Alberto Montresor
 * @since       Jgroup 0.8
 */
public interface RegistryLocator
{

  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds)
    throws RemoteException;

  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system and port.
   *
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port on which distributed replicas are listening
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds, int port)
        throws RemoteException;


  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system and port. <p>
   *
   * Communication with this remote registry will use the supplied
   * <t>RMIClientSocketFactory csf</t> to create socket connections
   * to the dependable registry on the distributed system and port.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port on which distributed replicas are listening
   * @param csf         client-side Socket factory used to make connections to
   *                    the dependable registry. If <t>csf</t> is null, then the
   *                    default client-side socket factory will be used
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds, int port,
                                     RMIClientSocketFactory csf)
        throws RemoteException;


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf)
        throws RemoteException, JgroupException;


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf, int port)
        throws RemoteException, JgroupException;


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @param csf         client-side <code>Socket</code> factory used to
   *                    make connections to the registry
   * @param ssf         server-side <code>ServerSocket</code> factory
   *                    used to accept connections to the registry
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf, int port,
                                           RMIClientSocketFactory csf,
                                           RMIServerSocketFactory ssf)
        throws RemoteException, JgroupException;

} // END RegistryLocator
