/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package jgroup.registry;

import java.rmi.*;
import java.io.*;

import jgroup.*;

/**
 * For obtaining references to remote objects, RMI includes a registry service,
 * that provides methods for storing and retrieving remote object references.
 * Each registry object acts independently and maintains a different set
 * of bindings (name, remote object), thus constituting a single point of
 * failure. Furthermore, a registry object running on a certain host can be
 * used to register only remote objects running on the same host; thus, in
 * order to obtain a stub for a remote object, a client need to be aware of
 * its location.
 * <p>
 * The registry implementation included with Jgroup has several advantages
 * over the standard Java RMI registry. First, it enables a set of replicated
 * remote objects to register themselves under the same name. This set
 * forms a remote object group that can be retrieved as a single entity.
 * All the Jgroup registry objects running in a distributed system
 * cooperate in order to maintain a replicated database of bindings and
 * offer a reliable and high-available repository facility. Clients no
 * longer need to be aware of the location of a service: they can simply
 * access the group of registry replicas and obtain a stub for every
 * service registered in the distributed system. A stub for the registry
 * group is obtained through the RegistryFactory.getDependableRegistry
 * methods.
 * <p>
 * A distributed system can be designed by including a certain number of registry
 * replicas running on different hosts and possibly on distinct portions of the
 * communication network. Client access these replicas through standard
 * RMI interactions as if they were a single registry, and are guaranteed
 * that their invocation will successfully terminate, provided that at
 * least one operational replica is running in their partition.
 * <p>
 * During a partitioning, the Jgroup registry presents a partitioned
 * behavior reflecting the current failure scenario. The registration
 * of a remote object inside a partition will not affect the registry
 * replicas not contained in that partition, while a retrieval operation
 * will not success if the service we are looking for has been registered
 * in another partition. Nevertheless, the replicas contained in a
 * partition consistently maintain the same set of bindings and act as
 * a single entity; moreover, the disappearance of the partitioning
 * causes the execution of a reconciliation protocol in order to
 * re-establish a consistent set of bindings among the replicas that
 * belonged to different partitions. It is important to note that this
 * behavior is perfectly reasonable for a partitionable distributed
 * system, since clients asking for remote services are interested only
 * in servers running in their current partitions.
 * <p>
 * The <code> DependableRegistry</code> interface describes the methods
 * of a dependable registry that can be remotely invoked.
 * To create a registry that runs in an application, use one of the
 * <tt>RegistryFactory.createDependableRegistry</tt> methods. To obtain
 * a reference to a remote object registry, use one of the
 * <tt>RegistryFactory.getDependableRegistry</tt> methods.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.3
 */
public interface        DependableRegistry
extends                 ExternalGMIListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Port number used by the dependable registry for bootstrap purpose; each
   * dependable registry server is locally registered in a standard registry
   * service running in the same host and using this port.
   */
  public static final int       DREGISTRY_BOOTSTRAP_PORT = 1200;


  /**
   * Port number used by the dependable registry for internal communications
   * among servers composing the dependable registry service.
   */
  public static final int       DREGISTRY_JGROUP_PORT = 4672;


  /**
   * Service name used to register each dependable registry replica in the
   * local registry.
   */
  public static final String    DEPENDABLE_REGISTRY_NAME = "DependableRegistry";



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a proxy for the remote object group associated with the specified name in the
   * registry.
   *
   * @param name      the name associated with the remote object group
   * @return          a proxy for the remote object group
   *
   * @exception RemoteException If remote operation failed.
   * @exception NotBoundException if there is no object group with this name in the
   *              registry.
   * @exception AccessException If this operation is not permitted.
   */
  public Remote lookup(String name)
    throws  RemoteException, NotBoundException, AccessException;


  /**
   * Add a replicated member object to the remote object group associated with
   * the specified name in the registry.
   *
   * @param name        the name associated with the remote object group
   * @param gmiService  the external GMI service associated with the member
   *                    object implementation
   * @param serverType  the class object of the member object implementation
   * @return            a <code>IID</code> identifier used to (possibly) remove
   *                    this replicated remote object from the group
   *
   * @exception RemoteException Raised when remote operation failed.
   * @exception McastRemoteException Marker used to identify methods to be invoked
   *            through multicast semantics.
   * @exception AccessException Raised when this operation is not permitted.
   */
  public IID bind(String name, Remote gmiService, Class serverType)
    throws  RemoteException, McastRemoteException, AccessException;


  /**
   * Remove a replicated remote object identified by the <code>IID</code>
   * object obtained when the replica has been registered.
   *
   * @param bid       the binding identifier obtained when the replica has
   *                  been registered
   *
   * @exception RemoteException If remote operation failed.
   * @exception McastRemoteException Marker used to identify methods to be invoked
   *            through multicast semantics
   * @exception NotBoundException if there is no object with this name in the
   *              registry.
   * @exception AccessException If this operation is not permitted.
   */
  public void unbind(IID bid)
    throws  RemoteException, McastRemoteException, NotBoundException, AccessException;


  /**
   * Returns an array containing the remote object group names registered in the
   * registry.
   *
   * @exception RemoteException If remote operation failed.
   * @exception AccessException If this operation is not permitted.
   */
  public String[] list()
    throws  RemoteException, AccessException;

} // END DependableRegistry



