/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.config;

import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

import jgroup.*;
import debug.Debug;

/**
 * Parser code to read an XML configuration file, using the supplied URL.
 * The default URL is 'file:$HOME/.jgroup/config.xml', which is used if the
 * non-argument constructor is used. Otherwise, the constructor takes the
 * URL as an argument.
 *
 * @author  Hein Meling
 * @since   Jgroup 1.1
 */
public final class Parser {

  /////////////////////////////////////////////////////////////////////////////
  // Constants - XML tags
  /////////////////////////////////////////////////////////////////////////////

  private static final String dsRootTagName             = "DistributedSystem";
  private static final String dsVersionAttrib           = "version";
  private static final String cfgVersion                = "1.0";

  private static final String dsRegistryTagName         = "Registry";
  private static final String dsPortTagName             = "Port";
  private static final String dsInitialTagName          = "Initial";
  private static final String dsMinimalTagName          = "Minimal";

  private static final String dsDomainTagName           = "Domain";
  private static final String dsDomainNameAttrib        = "name";
  private static final String dsHostTagName             = "Host";
  private static final String dsMcastTagName            = "Mcast";

  private static final String dsTransportConfigTagName  = "TransportConfig";
  private static final String dsPayloadTagName          = "Payload";
  private static final String dsLambdaTagName           = "Lambda";
  private static final String dsAlfaNTagName            = "AlfaN";
  private static final String dsAlfaDTagName            = "AlfaD";
  private static final String dsRTimeoutTagName         = "RTimeout";

  /////////////////////////////////////////////////////////////////////////////
  // Static block - Default URL for the configuration file
  /////////////////////////////////////////////////////////////////////////////

  private static final String configURL;

  static {
    String url = "file:" + System.getProperty("user.home");
    String s = java.io.File.separator;
    configURL = url.concat(s + ".jgroup" + s + "config.xml");
  }

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private RegistryConfig  regConfig;      // Holder for Registry parameters
  private RelacsConfig    trConfig;       // Holder for TransportConfig params
  private ArrayList       domains;        // Array of Domain objects

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs a parser that parses the file given by the default URL.
   */
  public Parser()
    throws JgroupException
  {
    this(configURL);
  }

  /**
   * Constructs a parser that parses the file obtained from the specified URL.
   *
   * @param urlConfig  The URL to the xml configuration file containing
   *                   the description of the distributed system.
   */
  public Parser(String urlConfig)
    throws JgroupException
  {
    RegistryConfig defaultReg = new RegistryConfig();
    RelacsConfig   defaultCfg = new RelacsConfig();
    domains = new ArrayList();
    try {
      DocumentBuilderFactory docBF = DocumentBuilderFactory.newInstance();
      docBF.setValidating(true);
      DocumentBuilder db = docBF.newDocumentBuilder();
      Document doc = db.parse(urlConfig);
      Element elm = doc.getDocumentElement();
      checkCfgVersion(elm);
      NodeList nl = elm.getChildNodes();
      boolean gRegistryParsed = false;
      boolean tConfigParsed = false;
      for (int i=1; i<=nl.getLength(); i++) {
        Node node = nl.item(i);
        if (node == null || node.getNodeName().startsWith("#")) {
          continue;
        }
        Element elem = (Element)node;
//        Debug.println("NodeName : " + elem.getNodeName());
        if (!gRegistryParsed) {
          // Pass null below to use default parameters,
          // if Registry tag not specified.
          RegistryConfig tmpReg = handleRegistryTag(elem, null);
          if (tmpReg != null) {
            // Set the global RegistryConfig parameters and don't parse again.
            gRegistryParsed = true;
            regConfig = tmpReg;
          } else {
            // Use default values if no Registry tag specified.
            regConfig = defaultReg;
          }
        }
        if (!tConfigParsed) {
          RelacsConfig tmpCfg = handleTConfigTag(elem);
          if (tmpCfg != null) {
            // Set the global TransportConfig parameters and don't parse again.
            tConfigParsed = true;
            trConfig = tmpCfg;
          } else {
            // Use default values if no TransportConfig tag specified.
            trConfig = defaultCfg;
          }
        }
        RelacsDomain domain = handleDomainTag(elem);
        if (domain != null) {
          domains.add(domain);
        }
      }
    } catch (Exception e) {
      throw new JgroupException("Error while parsing configuration file: " + e);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Public parser interface methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the parsed global registry configuration.
   */
  public RegistryConfig getRegistryConfig() {
    return regConfig;
  }

  /**
   * Returns the parsed transport configuration.
   */
  public RelacsConfig getRelacsConfig() {
    return trConfig;
  }

  /**
   * Returns the parsed domain configurations as an ArrayList.
   */
  public ArrayList getDomainList() {
    return domains;
  }

  /**
   * Returns the parsed domain configurations as an array of RelacsDomains.
   */
  public RelacsDomain[] getDomains() {
    return (RelacsDomain[]) domains.toArray(new RelacsDomain[domains.size()]);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Private methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Check that the configuration file has the expected version number.
   */
  private void checkCfgVersion(Element elm)
    throws JgroupException
  {
    String dsRoot = elm.getTagName();
    String dsVersion = elm.getAttribute(dsVersionAttrib);
    if (dsRoot.equalsIgnoreCase(dsRootTagName) && dsVersion.equals(cfgVersion)) {
      return;
    } else {
      throw new JgroupException("Error in configuration file: "
                                + "specified version not supported, \n"
                                + "or the root tag is not " + dsRootTagName);
    }
  }


  /**
   * Check if the Registry tag is specifed, and if so determine the values
   * of the <code>regPort, regInitial</code>, and <code>regMinimal</code>.
   * Use default values for all items not specified and also if no Registry
   * tag is specified.
   *
   * NOTE: Problem with this method: It will set default parameters that may
   * differ from the globally specified parameters, if used within a domain.
   *
   */
  private RegistryConfig handleRegistryTag(Element elm, RegistryConfig globalReg)
    throws JgroupException
  {
    if (elm == null || !elm.getTagName().equalsIgnoreCase(dsRegistryTagName)) {
      // This element was not a Registry tag.
      return null;
    }

    // Default values
    RegistryConfig defaultReg = (globalReg != null) ? globalReg : new RegistryConfig();
    int regPort    = defaultReg.getPort();
    int regInitial = defaultReg.getInitial();
    int regMinimal = defaultReg.getMinimal();

    NodeList regParam = elm.getChildNodes();
    for (int i=0; i<regParam.getLength(); i++) {
      Node n = regParam.item(i);
      regPort    = getIntValue(n, dsPortTagName,    regPort);
      regInitial = getIntValue(n, dsInitialTagName, regInitial);
      regMinimal = getIntValue(n, dsMinimalTagName, regMinimal);
    }
    return new RegistryConfig(regPort, regInitial, regMinimal);
  }


  /**
   * Check if the TransportConfig tag is specifed, and if so determine the
   * specified values. Use default values for all items not specified and
   * also if no TransportConfig tag is specified.
   */
  private RelacsConfig handleTConfigTag(Element elm)
    throws JgroupException
  {
    if (elm == null || !elm.getTagName().equalsIgnoreCase(dsTransportConfigTagName)) {
      // This element was not a TransportConfig tag.
      return null;
    }

    // Default values
    RelacsConfig defaultCfg = new RelacsConfig();
    int Payload  = defaultCfg.getPayload();
    int Lambda   = defaultCfg.getLambda();
    int AlfaN    = defaultCfg.getAlfan();
    int AlfaD    = defaultCfg.getAlfad();
    int RTimeout = defaultCfg.getRtimeout();

    NodeList cfgParam = elm.getChildNodes();
    for (int i=0; i<cfgParam.getLength(); i++) {
      Node n = cfgParam.item(i);
      Payload  = getIntValue(n, dsPayloadTagName,  Payload);
      Lambda   = getIntValue(n, dsLambdaTagName,   Lambda);
      AlfaN    = getIntValue(n, dsAlfaNTagName,    AlfaN);
      AlfaD    = getIntValue(n, dsAlfaDTagName,    AlfaD);
      RTimeout = getIntValue(n, dsRTimeoutTagName, RTimeout);
    }
    return new RelacsConfig(Payload, Lambda, AlfaN, AlfaD, RTimeout);
  }


  /**
   * Check if the Domain tag is specifed, and if so determine the values
   * of the Registry part: <code>regPort, regInitial</code>, and
   * <code>regMinimal</code>, and also the host names and multicast address.
   */
  private RelacsDomain handleDomainTag(Element elm)
    throws JgroupException
  {
    if (elm == null || !elm.getTagName().equalsIgnoreCase(dsDomainTagName)) {
      // This element was not a Domain tag.
      return null;
    }

    // Default values:
    // Domain registry is to use global registry parameters as defaults,
    // if not domain local parameters are set.
    RegistryConfig domainReg = regConfig;
    // Mcast address need not be specified, thus null.
    String mcast = null;
    // Domain name must always be given as attribute to the Domain tag.
    String domainName = elm.getAttribute(dsDomainNameAttrib);
    // Initially empty list of hosts.
    ArrayList hosts = new ArrayList();
    // We don't parse multiple domain registry tags, therefore we switch
    // the dRegistryParsed to true once we have found a domain registry tag.
    boolean dRegistryParsed = false;

    NodeList dnl = elm.getChildNodes();
    for (int i=0; i < dnl.getLength(); i++) {
      Node n = dnl.item(i);
      String thisHost = getValue(n, dsHostTagName);
      if (thisHost != null) {
        // Node was a host
        hosts.add(thisHost);
      } else {
        // Node was not a host
        String thisMcast = getValue(n, dsMcastTagName);
        if (thisMcast != null) {
          // Node was an mcast address.
          mcast = thisMcast;
        } else {
          // Node was not an mcast address either, thus try to:
          // Handle domain local registry parameters.
          if (n == null || n.getNodeName().startsWith("#")) {
            // If node starts with a # casting to Element will cause exception,
            // therefore we move on to the next node.
            continue;
          }
          Element elem = (Element)n;
          if (!dRegistryParsed) {
            // We have not parsed a domain local registry tag before.
            // Below we pass regConfig (global registry params) as defaults.
            RegistryConfig tmpReg = handleRegistryTag(elem, regConfig);
            if (tmpReg != null) {
              // Node was an domain local registry tag.
              dRegistryParsed = true;
              domainReg = tmpReg;
            } else {
              // Node was not domain local registry tag;
              // maybe this should not be allowed. Throw exception?
              // Default is to use global registry parameters
            }
          }
        }
      }
    }
    // Note: if domainReg == regConfig, global registry parameters will be used.
    return new RelacsDomain(domainName, domainReg, mcast, hosts);
  }


  /**
   * Returns an int value of the <code>node</code> matching <code>TagName</code>.
   * If the specified <code>TagName</code> did not match the <code>node</code>,
   * the <code>defaultValue</code> will be returned.
   *
   * @param  node          The node that should match with <code>TagName</code>.
   * @param  TagName       The tag name to match the node against.
   * @param  defaultValue  The default value that will be returned if no the
   *                       tag did not match the node.
   * @return An integer value of the value inside the leaf node
   *         matching <code>TagName</code>.
   */
  private int getIntValue(Node node, String TagName, int defaultValue)
    throws JgroupException
  {
    String tmpValue = getValue(node, TagName);
    if (tmpValue != null) {
      return (new Integer(tmpValue)).intValue();
    }
    return defaultValue;
  }


  /**
   * Return the value of the <code>node</code> matching <code>TagName</code>.
   *
   * @param   node     The node that should match with <code>TagName</code>.
   * @param   TagName  The tag name to match the node against.
   * @return  A string representation of the value inside the leaf node
   *          matching <code>TagName</code>.
   *
   */
  private String getValue(Node node, String TagName)
    throws JgroupException
  {
    if (node.getNodeName().equalsIgnoreCase(TagName)) {
      Node childNode = node.getChildNodes().item(0);
      String s = null;
      if (childNode.getNodeValue() != null) {
        // Trim the value to get rid of NL's at the front
        String t = childNode.getNodeValue().trim();
        int x = t.indexOf("\n");
        if (x >= 0) t = t.substring(0, x);
        return t;
      } else {
        throw new JgroupException("Error in configuration file: "
                                  + "No value assigned for node " + node);
      }
    } else {
      // The tag didn't match the specified node.
      return null;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Main method - used only for testing
  /////////////////////////////////////////////////////////////////////////////

  public static void main(String[] args)
    throws JgroupException
  {
    Parser p = null;
    if (args.length >= 1) {
      p = new Parser(args[0]);
    } else {
      p = new Parser();
    }
    RegistryConfig regCfg = p.getRegistryConfig();
    RelacsConfig relacsCfg = p.getRelacsConfig();
    ArrayList domains = p.getDomainList();
    Debug.println("Global registry parameters : " + regCfg);
    Debug.println("Transport config parameters: " + relacsCfg);
    for (int i=0; i<domains.size(); i++) {
      RelacsDomain thisDomain = (RelacsDomain) domains.get(i);
      Debug.println("Domain: " + thisDomain);
/*
      String[] hosts = thisDomain.getNamesWithDomain();
      Debug.println("Fully qualified host names: ");
      for (int j=0; j<hosts.length; j++) {
        Debug.print(hosts[j] + "  ");
      }
      Debug.println("");
*/
    }
  }

} // END of class Parser
