/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.config;

import java.net.*;
import java.util.*;

import jgroup.*;


/**
 * A <code> DistributedSystem</code> object contains the description of the
 * distributed system in which the distributed computation is carried on.
 * Currently, a distributed system is described as a collection of hosts
 * and may be constructed in two ways: by adding a collection of host names
 * to a initially empty distributed system, or by reading a file containing
 * a list of host names.
 *
 * @author  Alberto Montresor
 * @author  Hein Meling
 * @since   Jgroup 0.2
 */
public final class      RelacsDS
extends                 DistributedSystem
{

  /////////////////////////////////////////////////////////////////////////////
  // Static fields
  /////////////////////////////////////////////////////////////////////////////

  private static InetAddress    local = null;   // Local host

  static {
    byte[]      address;

    try {
      local = InetAddress.getLocalHost();
      address = local.getAddress();
      if (address[0] == 127 && address[1] == 0 && address[2] == 0 && address[3] == 1)
        throw new UnknownHostException();
    }
    catch (UnknownHostException e1) {
      local = null;
      String name = System.getProperty("jgroup.hostname");
      if (name != null)
        try {
          local = InetAddress.getByName(System.getProperty("jgroup.hostname"));
        } catch (Exception e2) { }
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  // Holder for Registry parameters
  private RegistryConfig  regConfig;
  // Holder for TransportConfig params
  private RelacsConfig    trConfig;
  // Array of Domain objects
  private ArrayList       domains;
  // Domain name -> index in domains
  private HashMap         domainMap = new HashMap();
  // Composition of hosts in all domains
  private ArrayList       allHosts = new ArrayList();

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs an empty distributed system containing no domains
   * and no hosts. The global registry and transport configuration
   * parameters are set to default values.
   */
  public RelacsDS()
  {
    super("");
    // Use default configuration values.
    regConfig = new RegistryConfig();
    trConfig  = new RelacsConfig();
    // Create an empty list of domains.
    domains   = new ArrayList();
  }

  /**
   * Constructs a distributed system whose description is
   * contained in a generic <code>DistributedSystem</code>.
   *
   * @param ds    The generic DistributedSystem object.
   */
  public RelacsDS(DistributedSystem ds)
    throws JgroupException
  {
    this(ds.getConfigURL());
  }

  /**
   * Constructs a distributed system whose description is
   * contained in a file located by <code>configURL</code>.
   * The config file specifies the distributed system using
   * the XML format. The file located at the specified URL
   * is parsed by the <code>Parser</code> class.
   *
   * @param configURL   The URL for the file containing the
   *                    description of the distributed system.
   */
  public RelacsDS(String configURL)
    throws JgroupException
  {
    super(configURL);
    Parser p = new Parser(configURL);
    regConfig = p.getRegistryConfig();
    trConfig  = p.getRelacsConfig();
    domains   = p.getDomainList();
    // Construct a mapping between the domain name and the index
    // into the 'domains' ArrayList for easy lookup.
    for (int i=0; i<domains.size(); i++) {
      RelacsDomain d = (RelacsDomain) domains.get(i);
      domainMap.put(d.getDomainName(), new Integer(i));
    }
    // Construct an ArrayList of hosts comprised from all domains.
    for (int i=0; i<domains.size(); i++) {
      RelacsDomain domain = (RelacsDomain) domains.get(i);
      ArrayList hostsInDomain = domain.getHostsWithDomain();
      allHosts.addAll(hostsInDomain);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // Public methods
  /////////////////////////////////////////////////////////////////////////////

  //
  // NB!
  //   The following methods have not yet been implemented:
  //   addDomain(), addTrConfig(), getTrConfig(), addRegConfig(),
  //   getAllHosts(), chngTrConfig(), ...
  //
  // Recall that addDomain() must also add entries into both the domains ArrayList
  // and also the domainMap HashMap.
  //
  /**
   *  Returns an ArrayList of domain objects representing the domains in
   *  the distributed system.
   */
  public ArrayList getDomains()
  {
    return domains;
  }

  /**
   *  Returns the number of domains contained in the distributed system.
   *
   */
  public int domainSize()
  {
    return domains.size();
  }

  /**
   *  Returns the global registry config.
   */
  public RegistryConfig getRegConfig()
  {
    return regConfig;
  }

  /**
   *  Returns an array of all hosts independant of domain.
   */
  public String[] getNames()
  {
    return (String[]) allHosts.toArray(new String[allHosts.size()]);
  }

  /**
   *  Returns the number of hosts in the distributed system.
   *
   */
  public int size()
  {
    return allHosts.size();
  }

  /**
   *  Add a host name to a specified domain in the distributed system.
   *
   *  @param domainname  the name of the domain to add the host to
   *  @param hostname    the name of the host to add
   */
  public void addHost(String domainname, String hostname)
  {
    int i = ((Integer) domainMap.get(domainname)).intValue();
    ((RelacsDomain) domains.get(i)).add(hostname);
  }


  /////////////////////////////////////////////////////////////////////////////
  // Static methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the internet address of the local host.
   */
  public static InetAddress getLocalHost()
    throws UnknownHostException
  {
    if (local == null)
      throw new UnknownHostException();
    return local;
  }

  /**
   *  Returns the hostname of the local host.
   */
  public static String getLocalHostName()
  {
    if (local == null)
      return "localhost";  // Dirty trick!
    return local.getHostName();
  }

} // END RelacsDS













