/*
 * Copyright (c) 1998-2001 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.config;

import java.util.*;

/**
 * Class for holding domain information.
 *
 * @author Hein Meling
 * @since Jgroup 1.1
 */
public class RelacsDomain {

  /////////////////////////////////////////////////////////////////////////////
  // Fields
  /////////////////////////////////////////////////////////////////////////////

  private String           domainName;
  private RegistryConfig   domainReg;
  private String           mcastAddress;
  private ArrayList        hosts;

  /////////////////////////////////////////////////////////////////////////////
  // Constructors
  /////////////////////////////////////////////////////////////////////////////

  public RelacsDomain(String domain) {
    // null indicate that we don't use multicast within this domain
    this(domain, null, null, new ArrayList());
  }

  public RelacsDomain(String domain, String mcast) {
    this(domain, null, mcast, new ArrayList());
  }

  public RelacsDomain(String domain, ArrayList hosts) {
    // null indicate that we don't use multicast within this domain
    this(domain, null, null, hosts);
  }

  public RelacsDomain(String domain, String mcast, ArrayList hosts) {
    this(domain, null, mcast, hosts);
  }

  public RelacsDomain(String domain, RegistryConfig domainreg,
                      String mcast, ArrayList hosts)
  {
    this.domainName = domain;
    this.domainReg = domainreg;
    this.mcastAddress = mcast;
    this.hosts = hosts;
  }

  /////////////////////////////////////////////////////////////////////////////
  // Public RelacsDomain interface methods
  /////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the domain name.
   */
  public String getDomainName() {
    return domainName;
  }

  /**
   *  Returns the multicast address for this domain.
   */
  public String getMcastAdr() {
    return mcastAddress;
  }

  /**
   *  Returns the domain local registry information.
   */
  public RegistryConfig getRegistryConfig() {
    return domainReg;
  }

  /**
   *  Returns the number of hosts contained in this domain.
   */
  public int size() {
    return hosts.size();
  }

  /**
   *  Returns an array of host names representing the hosts contained in
   *  this domain.
   */
  public String[] getNames() {
    return (String[]) hosts.toArray(new String[hosts.size()]);
  }

  /**
   *  Returns an array of domain qualified host names representing the hosts
   *  contained in this domain.
   */
  public String[] getNamesWithDomain() {
    String [] tmpHosts = (String[]) hosts.toArray(new String[hosts.size()]);
    for (int i=0; i<tmpHosts.length; i++) {
      if (!this.domainName.equals(""))
        tmpHosts[i] += "." + this.domainName;
    }
    return tmpHosts;
  }
  /**
   *  Returns an ArrayList of domain qualified host names representing
   *  the hosts contained in this domain.
   */
  public ArrayList getHostsWithDomain() {
    ArrayList tmpHosts = new ArrayList(hosts.size());
    for (int i=0; i<hosts.size(); i++) {
      String hostname = (String) hosts.get(i);
      if (!this.domainName.equals(""))
        hostname += "." + this.domainName;
      tmpHosts.add(hostname);
    }
    return tmpHosts;
  }

  /**
   * Add a host name to the distributed system.
   *
   * @param hostname    The name of the host to add, without domain name.
   *
   */
  public void add(String hostname) {
    hosts.add(hostname);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Override methods in Object
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a string representation of the RelacsDomain object.
   */
  public String toString() {
    String tmp = "[" + domainName + ": "
      + ((domainReg != null) ? (domainReg.toString() + ", ") : "")
      + ((mcastAddress != null) ? (mcastAddress + ", ") : "");
    for (int i=0; i<hosts.size(); i++) {
      // Comma separate the hosts
      tmp += ((i != 0) ? ", " : "") + (String) hosts.get(i);
    }
    return tmp + "]";
  }

} // END of class RelacsDomain
