/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.general.HashTable;

/**
 * The <code></code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class     GroupTable
extends         HashTable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Group         head;           // First element of the table
  private Group[]       table;          // Hash table for looking up
  private int           size;           // Number of groups inserted
  private int[]         gids;           // Array of group identifiers
  private Group[]       groups;         // Array of groups
  private boolean       changed;        // True if grouptable is different from gids
  private VersionList   vlist;          // List containing the old version of exited groups

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  GroupTable()
  {
    head    = null;
    table   = new Group[HASHTABLE_SIZE];
    size    = 0;
    changed = true;
    vlist   = new VersionList();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Inserts group <code> Group</code> in the table.
   *  @param            group           the group that must be inserted.
   */
  boolean insert(Group group)
  {
    int                 index;          // Hash table index
    Group               scan;           // Scanning variable
    VersionList         vd;             // Old version number of group

    index = computeHash(group.key);
    scan = table[index];

    // Checks whether the group is already in the table
    while (scan != null && scan.key != group.key) {
      scan = scan;
      scan = scan.hnext;
    }
    if (scan == null) {

      // Inserts group in hash list
      try { table[index].hprev = group; } catch (Exception e) {}
      group.hnext  = table[index];
      group.hprev  = null;
      table[index] = group;

      // Inserts group in total list
      try { head.tprev  = group; } catch (Exception e) {};
      group.tnext = head;
      group.tprev = null;
      head        = group;

      // Copies old version number if the group has been already joined
      if ((vd = (VersionList) vlist.remove(group.key)) != null)
        group.me.version = vd.version + 1;

      // Updates grouptable fields
      size++;
      changed = true;
      return true;
    }
    return false;
  }

  /**
   *  Removes group identified by <code> gid</code> from the table.
   *  @param            gid             the identifier of the group that must be removed.
   */
  void remove(int gid)
  {
    int                 index;          // Hash table index
    Group               scan;           // Scanning variable

    index = computeHash(gid);
    scan = table[index];
    while (scan != null && scan.key != gid)
      scan = scan.hnext;
    if (scan != null) {
      if (scan.tnext != null)   scan.tnext.tprev = scan.tprev;
      if (scan.tprev != null)   scan.tprev.tnext = scan.tnext;
      else                      head = scan.tnext;
      if (scan.hnext != null)   scan.hnext.hprev = scan.hprev;
      if (scan.hprev != null)   scan.hprev.hnext = scan.hnext;
      else                      table[index] = scan.tnext;
      vlist.insert(new VersionList(gid, scan.me.version));
      size--;
      changed = true;
    }
  }

  /**
   *  Searches a group identified by <code> gid</code>.
   */
  Group lookup(int gid)
  {
    Group               scan;                           // Search variable

    scan = table[computeHash(gid)];
    while (scan != null && scan.key != gid)
      scan = scan.hnext;
    return scan;
  }

  int[] getGids()
  {
    Group               scan;           // Used to scan grouptable
    int                 i;              // Counter

    if (changed) {
      gids = new int[size];
      groups = new Group[size];
      scan = head;
      i = 0;
      while (scan != null && i<size) {
        gids[i] = scan.key;
        groups[i] = scan;
        scan = scan.tnext;
        i++;
      }
      changed = false;
    }
    return gids;
  }

  Group[] getGroups()
  {
    Group               scan;           // Used to scan grouptable
    int                 i;              // Counter

    if (changed) {
      gids = new int[size];
      groups = new Group[size];
      scan = head;
      i = 0;
      while (scan != null && i<size) {
        gids[i] = scan.key;
        groups[i] = scan;
        scan = scan.tnext;
        i++;
      }
      changed = false;
    }
    return groups;
  }



} // END GroupTable



