/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;
import debug.Debug;

/**
 * The <code> MsgAck</code> class contains the methods to encode and decode
 * ACK messages from remote hosts. The ACK message structure is the following:
 * <p>
 * <code>
 * +-------+-------+-------+-------+-------+ ....... +-------+
 * |  GID  |  VID  |  HPOS | VLEN  |  ACK  |         |  ACK  |
 * |       |       |       |       |   1   |         |  VLEN |
 * +-------+-------+-------+-------+-------+ ....... +-------+
 *</code>
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgAck
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgAck msg;

  static {
    msg = new MsgAck();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** View identifier */            long        vid;
  /** Host index in the view */     int         hpos;
  /** Acknowledgement numbers */    int[]       ack;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a <code>MsgAck</code> object that contains the decoding of the m-received input
   * stream.
   *
   * @param stream    the message input stream to decode
   */
  static MsgAck decode(InMessage stream)
  {
    msg.vid  = Types.readVid(stream);
    msg.hpos = Types.readGrp(stream);
    int len = Types.readGrp(stream);        // Read length of the ack field
    msg.ack  = new int[len];
    for (int i=0; i < msg.ack.length; i++)    // Read ack field
      msg.ack[i] = Types.readMid(stream);

    return msg;
  }

  /**
   * Returns a output stream containing the encoding of a <code>MsgAck</code> message.
   *
   * @param     gid     group identifier
   * @param     vid     view identifier
   * @param     hpos    host index in the view
   * @param     vcomp   view composition
   * @param     vlen    actual length of hpVcomp (hpVcomp.length >= vlen)
   */
  static OutMessage encode(MsgFactory mfactory, int gid, long vid, int hpos,
                           HostData[] vcomp, int vlen)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + Types.VID_SIZE + Types.GRP_SIZE * 2 + Types.MID_SIZE * vlen;
    OutMessage stream = mfactory.get(size);

    Types.writeGid(stream, gid);
    Types.writeVid(stream, vid);
    Types.writeGrp(stream, hpos);
    Types.writeGrp(stream, vlen);   // Write length of the ack field
    for (int i=0; i < vlen; i++)        // Write ack field
      Types.writeMid(stream, vcomp[i].hack[hpos]);

    return stream;
  }

  void print()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Vid: " + vid + " | Hpos: " + hpos + "| Ack:");
      for (int i=0; i < ack.length; i++)
        Debug.print(20, " " + ack[i]);
      Debug.println(20, "");
    }
  }

} // END MsgAck
