/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

//import relacs.general.*;
import relacs.mss.*;
import debug.Debug;


/**
 * The <code> MsgProp</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgProp
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Position constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static final int P_DATA =
    Types.GID_SIZE + Types.VID_SIZE*2 + Types.LEN_SIZE * 4 + Types.CHK_SIZE*2;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  long                  cvid;           // Last complete view identifier
  long                  pvid;           // Last partial view identifier
  int                   last_plen;      // last_prop field length (in bytes)
  int                   host_plen;      // host_prop field length (in bytes)
  int                   memb_plen;      // memb_prop field length (in bytes)
  int                   msgs_plen;      // msgs_prop field length (in bytes)
  int                   last_start;     // last_prop field start position
  int                   host_start;     // host_prop field start position
  int                   memb_start;     // memb_prop field start position
  int                   msgs_start;     // msgs_prop field start position
  int                   host_pchk;      // checksum of host_prop
  int                   msgs_pchk;      // checksum of msgs_prop
  InMessage             stream;         // Message stream of the message
  OutMessage            last_prop;      // Proposal section: last complete view
  OutMessage            host_prop;      // Proposal section: view composition (hosts)
  OutMessage            memb_prop;      // Proposal section: view composition (members)
  OutMessage            msgs_prop;      // Proposal section: delivered messages


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a new instance of <code> MessageProp</code> starting from a m-received
   * message.
   *
   * @param     data    the m-received message.
   */
  MsgProp(InMessage stream)
  {
    this.stream = stream;

    // Read single fields
    cvid  = Types.readVid(stream);
    pvid  = Types.readVid(stream);
    last_plen = Types.readLen(stream);
    host_plen = Types.readLen(stream);
    memb_plen = Types.readLen(stream);
    msgs_plen = Types.readLen(stream);
    host_pchk = Types.readChk(stream);
    msgs_pchk = Types.readChk(stream);

    // Computes starting positions
    last_start = P_DATA;
    host_start = last_start + last_plen;
    memb_start = host_start + host_plen;
    msgs_start = memb_start + memb_plen;
  }

  /**
   * Creates a new instance of <code> MsgProp</code> acting as a container for a
   * proposal.
   */
  MsgProp()
  {
    last_prop   = new OutMessage(256);
    host_prop   = new OutMessage(256);
    memb_prop   = new OutMessage(256);
    msgs_prop   = new OutMessage(256);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  static MsgProp decode(InMessage stream)
  {
    return new MsgProp(stream);
  }


  /*
   * Encodes the last complete view and stores it into the specified group.
   *
   * @param     cview   last complete view
   * @param     length  actual length of cview
   */
  void encodeLastProp(int[] cview, int length)
  {
    last_prop.reset();
    Types.writeGrp(last_prop, length);
    for (int i = 0; i < length; i++)
      Types.writeHid(last_prop, cview[i]);
    last_plen = Types.GRP_SIZE + length * Types.HID_SIZE;
  }

  /**
   * Encodes the next view estimate (hosts) and stores it into the specified
   * group.
   *
   * @param     hdata   estimate array
   * @param     hsize   actuale length of the estimate array
   */
  void encodeHostProp(HostData[] hdata, int hsize)
  {
    host_prop.reset();
    Types.writeGrp(host_prop, hsize);
    for (int i = 0; i < hsize; i++)
      Types.writeHid(host_prop, hdata[i].key);
    for (int i = 0; i < hsize; i++)
      Types.writeInc(host_prop, hdata[i].host.getIncarnation());
    for (int i = 0; i < hsize; i++)
      Types.writeVer(host_prop, hdata[i].agreed);
    host_pchk = host_prop.computeChecksum();
    host_plen = Types.GRP_SIZE + hsize * (Types.HID_SIZE + Types.INC_SIZE + Types.VER_SIZE);
  }


  /**
   * Encodes the next view estimate (members) and stores it into the specified
   * group.
   *
   * @param tmember     list of members
   * @param msize       number of members in the current estimate
   */
  void encodeMembProp(MemberData tmembers, int nmembers)
  {
    MemberData scan;

    memb_prop.reset();
    Types.writeGrp(memb_prop, nmembers);
    scan = (MemberData) tmembers.getFirst();
    while (scan != null) {
      if (!scan.leaving)
        Types.writeGrp(memb_prop, scan.key);
      scan = (MemberData) scan.getNext();
    }
    memb_plen = Types.GRP_SIZE + nmembers * Types.GRP_SIZE;
  }

  /**
   * Encodes the information about delivered messages and stores it into the
   * specified group.
   *
   * @param     hdata   estimate array
   * @param     hsize   actuale length of the estimate array
   */
  void encodeMsgsProp(HostData[] hdata, int hsize)
  {
    msgs_prop.reset();

    for (int i = 0; i < hsize; i++)
      Types.writeMid(msgs_prop, hdata[i].delivered);
    msgs_pchk = msgs_prop.computeChecksum();
    msgs_plen = hsize * Types.MID_SIZE;
  }


  /**
   * Encodes a MsgProp message  starting from raw data.
   *
   * @param     group   object from which raw data are extracted.
   */
  OutMessage encode(MsgFactory mfactory, int gid, long cvid, long pvid)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + 2*Types.VID_SIZE + Types.LEN_SIZE * 4 + Types.CHK_SIZE*2 +
               last_plen + host_plen + memb_plen + msgs_plen;
    OutMessage stream = mfactory.get(size);

    Types.writeGid(stream,  gid);
    Types.writeVid(stream,  cvid);
    Types.writeVid(stream,  pvid);
    Types.writeLen(stream, last_plen);
    Types.writeLen(stream, host_plen);
    Types.writeLen(stream, memb_plen);
    Types.writeLen(stream, msgs_plen);
    Types.writeChk(stream, host_pchk);
    Types.writeChk(stream, msgs_pchk);

    stream.write(last_prop);
    stream.write(host_prop);
    stream.write(memb_prop);
    stream.write(msgs_prop);

    return stream;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Equality methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Compares the estimates part of two instances of MsgProp
   *
   * @param     msg     the reference MsgProp message with which to compare.
   */
  boolean checkEstimate(MsgProp msg)
  {
    if (msg == null || host_plen != msg.host_plen || host_pchk != msg.host_pchk)
      return false;
    return stream.compare(host_start, msg.stream, msg.host_start, host_plen);
  }

  /**
   * Compares the messages part of two instances of MsgProp
   *
   * @param     msg     the reference MsgProp message with which to compare.
   */
  boolean checkMessages(MsgProp msg)
  {
    if (msg == null)
      return false;
    if (pvid == msg.pvid) {
      if (msgs_plen != msg.msgs_plen || msgs_pchk != msg.msgs_pchk)
        return false;
      return stream.compare(msgs_start, msg.stream, msg.msgs_start, msgs_plen);
    }
    return true;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Decodes the last complete view.
   */
  int[] decodeLastProp()
  {
    stream.seek(last_start);
    int len = Types.readGrp(stream);
    int[] ret = new int[len];
    for (int i=0; i < len; i++)
      ret[i] = Types.readHid(stream);
    return ret;
  }

  /**
   * Decodes the estimate for the next complete view.
   */
  int[] decodeHostProp_hosts()
  {
    stream.seek(host_start);
    int len = Types.readGrp(stream);
    int[] ret = new int[len];
    for (int i=0; i < len; i++)
      ret[i] = Types.readHid(stream);
    return ret;
  }


  /**
   * Decodes the agreed version numbers for the next complete view.
   */
  int[] decodeHostProp_incarns()
  {
    stream.seek(host_start);
    int len = Types.readGrp(stream);
    int[] ret = new int[len];
    stream.seek(host_start + Types.GRP_SIZE + len * Types.HID_SIZE);
    for (int i=0; i < len; i++)
      ret[i] = Types.readInc(stream);
    return ret;
  }


  /**
   * Decodes the agreed version numbers for the next complete view.
   */
  int[] decodeHostProp_agreed()
  {
    stream.seek(host_start);
    int len = Types.readGrp(stream);
    int[] ret = new int[len];
    stream.seek(host_start + Types.GRP_SIZE + len * (Types.HID_SIZE * Types.INC_SIZE));
    for (int i=0; i < len; i++)
      ret[i] = Types.readVer(stream);
    return ret;
  }

  /**
   * Decodes the list of members contained in the proposal.
   */
  int[] decodeMembProp()
  {
    stream.seek(memb_start);
    int len = Types.readGrp(stream);
    int[] ret = new int[len];
    for (int i=0; i < len; i++)
      ret[i] = Types.readGrp(stream);
    return ret;
  }

  void print()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Proposal: ");
      stream.print(20);
      Debug.println(20, "");
    }
  }

} // END MsgProp
