/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;
import debug.Debug;


/**
 * The <code> MsgSymm</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgSymm
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgSymm msg;

  static {
    msg = new MsgSymm();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  int                   vsend;          // Sender version number
  int[]                 hosts;          // Host identifiers
  int[]                 version;        // Corresponding version numbers
  int[]                 rset;           // Reachable set


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Creates a new instance of MsgSymm starting from a m-received msg input stream
   */
  static MsgSymm decode(InMessage msgstream)
  {
    int         len;                    // Version length

    msg.vsend = Types.readVer(msgstream);           // Read sender version number
    len  = Types.readGrp(msgstream);                // Read host identifiers and version numbers
    msg.version = new int[len];
    msg.hosts   = new int[len];
    for (int i=0; i < len; i++) {
      msg.hosts[i]   = Types.readHid(msgstream);
      msg.version[i] = Types.readVer(msgstream);
    }
    len = Types.readGrp(msgstream);                 // Read rset array
    msg.rset = new int[len];
    for (int i=0; i < len; i++)
      msg.rset[i] = Types.readHid(msgstream);
    return msg;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Copies informations into Message structure
   */
  static OutMessage encode(MsgFactory mfactory, int gid, int vsend, HostData[] nrset, MssHost[] rset)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + Types.VER_SIZE + 2 * Types.GRP_SIZE + nrset.length *
               (Types.VER_SIZE + Types.HID_SIZE) + rset.length * Types.HID_SIZE;
    OutMessage msgstream = mfactory.get(size);

    // Writes data
    Types.writeGid(msgstream, gid);                 // Write group identifier
    Types.writeVer(msgstream, vsend);               // Write sender version number
    Types.writeGrp(msgstream, nrset.length);        // Write nrset identifiers and version numbers
    for (int i=0; i < nrset.length; i++) {
      Types.writeHid(msgstream, nrset[i].key);
      Types.writeVer(msgstream, nrset[i].version);
    }
    Types.writeGrp(msgstream, rset.length);         // Writes rset array
    for (int i=0; i < rset.length; i++)
      Types.writeHid(msgstream, rset[i].getAddress());

    return msgstream;
  }


  /**
   *  Copies informations into Message structure
   */
  static OutMessage encode(MsgFactory mfactory, int gid, int vsend, HostData dest, MssHost[] rset, int rlen)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + Types.VER_SIZE + 2 * Types.GRP_SIZE +
               Types.VER_SIZE + Types.HID_SIZE + rlen * Types.HID_SIZE;
    OutMessage msgstream = mfactory.get(size);

    Types.writeGid(msgstream, gid);                 // Write group identifier
    Types.writeVer(msgstream, vsend);               // Write sender version number
    Types.writeGrp(msgstream, 1);                   // Write dest identifier and version number
    Types.writeHid(msgstream, dest.key);
    Types.writeVer(msgstream, dest.version);
    Types.writeGrp(msgstream, rlen);
    for (int i=0; i < rlen; i++)
      Types.writeHid(msgstream, rset[i].getAddress());

    return msgstream;
  }

  void print()
  {
    if (Debug.DEBUG) {
      int         i;                      // Counter

      Debug.print(20, "Vsend: " + vsend + " | Rset:");
      for (i=0; i < rset.length; i++)
        Debug.print(20, (rset[i] & 0xFF) + " ");
      Debug.println(20, "");
      Debug.print(20, "Hosts: ");
      for (i=0; i < hosts.length; i++)
        Debug.print(20, (hosts[i]& 0xFF)  + " ");
      Debug.print(20, "| Version: ");
      for (i=0; i < version.length; i++)
        Debug.print(20, version[i] + " ");
      Debug.println(20, "");
    }
  }

} // END MsgSymm

