/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;
import debug.Debug;

/**
 * The <code> MsgSynch</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class MsgSynch
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgSynch msg;

  static {
    msg = new MsgSynch();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  int                   vsend;          // Sender version number
  int                   vdest;          // Destination version number
  int[]                 rset;           // Reachable set (32 bit adressess)


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Decodes a <code> MsgSynch</code> message starting from a m-received msg input stream.
   */
  static MsgSynch decode(InMessage stream)
  {
    // Read data
    msg.vsend = Types.readVer(stream);
    msg.vdest = Types.readVer(stream);
    int len = Types.readGrp(stream);        // Length of the rset field
    msg.rset = new int[len];
    for (int i=0; i < len; i++)
      msg.rset[i] = Types.readHid(stream);
    return msg;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Encoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Computes the encoding of a <code> MsgSynch</code> message.
   *
   * @param     gid     group identifier.
   * @param     vsend   version number of the sender host
   * @param     vdest   version number of the destination host
   * @param     rset    reachable set
   */
  static OutMessage encode(MsgFactory mfactory, int gid, int vsend, int vdest, MssHost[] rset)
  {
    // Allocates data buffer
    int size = Types.GID_SIZE + 2*Types.VER_SIZE + Types.GRP_SIZE +
               rset.length * Types.HID_SIZE;
    OutMessage stream = mfactory.get(size);

    // Writes data
    Types.writeGid(stream, gid);
    Types.writeVer(stream, vsend);
    Types.writeVer(stream, vdest);
    Types.writeGrp(stream, rset.length);
    for (int i=0; i < rset.length; i++)
      Types.writeHid(stream, rset[i].getAddress());

    return stream;
  }

  void print()
  {
    if (Debug.DEBUG) {
      Debug.print(20, "Vsend: "+ vsend);
      Debug.print(20, "| Vdest: "+ vdest);
      Debug.print(20, "| Rset:");
      for (int i=0; i < rset.length; i++)
        Debug.print(20, " " + (rset[i] & 0xFF));
      Debug.println(20, "");
    }
  }

} // END MsgSynch
