/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;
import debug.Debug;


/**
* The <code> MsgView</code> class
*
* @author  Alberto Montresor
* @since   Jgroup 0.1
*/
final class MsgView
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Unique message instance
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * As these messages are never stored by the daemon, a single message instance is
   * sufficient. In this way, we save the costs of allocating and garbage collecting
   * messages.
   */
  private static MsgView msg;

  static {
    msg = new MsgView();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** New complete view identifier */     long      cvid;
  /** New compete view composition */     int[]     cvcomp;
  /** */                                  int[]     incarns;
  /** Agreed version numbers */           int[]     agreed;
  /** Partial view identifiers */         int[]     pvids;
  /** Members tables */                   int[][]   members;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Decoding methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a <code>MsgView</code> object that contains the decoding of the m-received input
   * stream.
   *
   * @param stream    the message input stream to decode
   */
  static MsgView decode(InMessage stream)
  {
    msg.cvid = Types.readVid(stream);
    int len  = Types.readGrp(stream);
    msg.cvcomp  = new int[len];
    msg.incarns = new int[len];
    msg.agreed  = new int[len];
    msg.pvids   = new int[len];
    msg.members = new int[len][];
    for (int i=0; i < len; i++)
      msg.cvcomp[i]  = Types.readHid(stream);
    for (int i=0; i < len; i++)
      msg.incarns[i] = Types.readInc(stream);
    for (int i=0; i < len; i++)
      msg.agreed[i]  = Types.readVer(stream);
    for (int i=0; i < len; i++)
      msg.pvids[i]   = Types.readGrp(stream);
    for (int i=0; i < len; i++) {
      msg.members[i] = new int[Types.readGrp(stream)];
      for (int j=0; j < msg.members[i].length; j++)
        msg.members[i][j] = Types.readGrp(stream);
    }
    return msg;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Message to send
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a output stream containing the encoding of a <code>MsgView</code> message.
   *
   * @param     gid     group identifier
   * @param     cvid    identifier of the current complete view
   * @param     prop    proposal on which the agreement has been reached
   * @param     pvids   identifiers of previous partial views (needed to compute new partial
   *                    views
   * @param     props   proposal array (containing members lists)
   */
  static OutMessage encode(MsgFactory mfactory, int gid, long cvid, MsgProp prop, int[] pvids,
                           MsgProp[] props)
  {
    byte[] tempbuffer;

    // Allocates data buffer
    int size = Types.GID_SIZE + Types.VID_SIZE + Types.GRP_SIZE + prop.host_plen +
           pvids.length * Types.GRP_SIZE;
    for (int i=0; i < props.length; i++)
      size += props[i].memb_plen;
    OutMessage stream = mfactory.get(size);

    // Writes single data
    Types.writeGid(stream,  gid);
    Types.writeVid(stream, cvid);

    // Copies composition and agreed values
    prop.stream.seek(prop.host_start);
    tempbuffer = new byte[prop.host_plen];
    prop.stream.read(tempbuffer);
    stream.write(tempbuffer);

    // Writes coordinator table informations
    for (int i=0; i < pvids.length; i++)
      Types.writeGrp(stream, pvids[i]);
    for (int i=0; i < props.length; i++) {
      props[i].stream.seek(props[i].memb_start);
      tempbuffer = new byte[props[i].memb_plen];
      props[i].stream.read(tempbuffer);
      stream.write(tempbuffer);
    }

    return stream;
  }


 void print()
  {
    if (Debug.DEBUG) {
      int clen = cvcomp.length;

      Debug.print(20, "Cvid: " + cvid);

      Debug.print(20, "| Cvcomp: ");
      for (int i=0; i < clen; i++)
        Debug.print(20, (cvcomp[i] & 0xFF) + " ");
      Debug.println(20, "");
      Debug.print(20, "Agreed: ");
      for (int i=0; i < clen; i++)
        Debug.print(20, agreed[i] + " ");
      Debug.print(20, "| Pvids: ");
      for (int i=0; i < clen; i++)
        Debug.print(20, pvids[i] + " ");
      Debug.println(20, "");
      Debug.print(20, "Members: ");
      for (int i=0; i < clen; i++) {
        Debug.print(20, "[ ");
        for (int j=0; j < members[i].length; j++)
          Debug.print(20, members[i][j] + " ");
        Debug.print(20, "] ");
      }
      Debug.println(20, "");
    }
  }

} // END MsgView

