/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.daemon;

import relacs.mss.*;

/**
 * The <code>Types</code> class contains a collection of psuedo-types, all of them
 * are implemented by the primitive type <code>int</code> (except <code>vid</code>
 * that is represented by the primitive type <code>long</code>). All of these types
 * are supplied with a coding function (<code>writeXxx()</code>) from values to
 * message output stream and a decoding function (<code>readXxx()</code>)
 * from message input stream to values. In this way, it is possible to modify the
 * coding algorithm of each of these values without that the classes accessing them
 * have to be informed.
 *
 * @author  Alberto Montresor
 * @since   Javagroup 1.0
 */

final class Types
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Sizes in bytes of integer types
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size in byte of pseudo-type Hid, used to encode host identifiers (internet addresses) */
  static int HID_SIZE     = 4;

  /** Size in byte of pseudo-type Inc, used to encode incarnation numbers */
  static int INC_SIZE     = 4;

  /** Size in byte of pseudo-type Mid, used to identify messages */
  static int MID_SIZE     = 4;

  /** Size in byte of pseudo-type Grp, used to describe the number of members in a group */
  static int GRP_SIZE     = 2;

  /** Size in byte of pseudo-type Chk, used to encode checksum */
  static int CHK_SIZE     = 1;

  /** Size in byte of pseudo-type Flg, used as a boolean flag */
  static int FLG_SIZE     = 1;

  /** Size in byte of pseudo-type Gid, used to identify groups */
  static int GID_SIZE     = 4;

  /** Size in byte of pseudo-type Vid, used to identify views */
  static int VID_SIZE     = 8;

  /** Size in byte of pseudo-type Sid, used to identify senders of messages */
  static int SID_SIZE     = 8;

  /** Size in byte of pseudo-type Len, used to describe lengths (in mcast messages) */
  static int LEN_SIZE     = 4;

  /** Size in byte of pseudo-type Ver, used for version numbers */
  static int VER_SIZE     = 4;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Read methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Read a host identifier from the input stream
   *
   * @param msgstream   the message input stream to read
   * @return    the value read.
   */
  static int readHid(InMessage msgstream)
  {
    return msgstream.read4();
  }

  /**
   * Read an incarnation number from the input stream
   *
   * @param msgstream   the message input stream to read
   * @return    the value read.
   */
  static  final int readInc(InMessage msgstream)
  {
    return msgstream.read4();
  }

  /**
   * Read a message identifier from the input stream
   *
   * @param msgstream   the message input stream to read
   * @return    the value read.
   */
  static int readMid(InMessage msgstream)
  {
    return msgstream.read4();
  }

  /**
   * Read a group index from the input stream
   *
   * @param msgstream   the message input stream to read
   * @return    the value read.
   */
  static int readGrp(InMessage msgstream)
  {
    return msgstream.read2();
  }

  /**
   * Read a checksum value from the input stream
   *
   * @param msgstream   the message input stream to read
   * @return    the value read.
   */
  static int readChk(InMessage msgstream)
  {
    return msgstream.read();
  }


  /**
   * Read a flag from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static boolean readFlg(InMessage stream)
  {
    return (((byte) stream.read())>0);
  }

  /**
   * Read a group identifier from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static int readGid(InMessage stream)
  {
    return stream.read4();
  }

  /**
   * Read a view identifier from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static long readVid(InMessage stream)
  {
    return stream.read8();
  }

  /**
   * Read a sender identifier from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static long readSid(InMessage stream)
  {
    return stream.read8();
  }

  /**
   * Read a length value from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static int readLen(InMessage stream)
  {
    return stream.read4();
  }

  /**
   * Read a version number from the input stream
   *
   * @param stream    the message input stream to read
   * @return    the value read.
   */
  static int readVer(InMessage stream)
  {
    return stream.read4();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Write methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Writes a host identifier on the output stream.
   *
   * @param     msgstream   the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeHid(OutMessage msgstream, int value)
  {
    msgstream.write4(value);
  }

  /**
   * Writes an incarnation number on the output stream.
   *
   * @param     msgstream   the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeInc(OutMessage msgstream, int value)
  {
    msgstream.write4(value);
  }

  /**
   * Writes a message identifier on the output stream.
   *
   * @param     msgstream   the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeMid(OutMessage msgstream, int value)
  {
    msgstream.write4(value);
  }

  /**
   * Writes a group index on the output stream.
   *
   * @param     msgstream   the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeGrp(OutMessage msgstream, int value)
  {
    msgstream.write2(value);
  }

  /**
   * Writes a chechsum on the output stream.
   *
   * @param     msgstream   the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeChk(OutMessage msgstream, int value)
  {
    msgstream.write(value);
  }

  /**
   * Writes a flag on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeFlg(OutMessage stream, boolean value)
  {
    stream.write(value ? (byte) 1 : (byte) 0);
  }

  /**
   * Writes a group identifier on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeGid(OutMessage stream, int value)
  {
    stream.write4(value);
  }

  /**
   * Writes a view identifier on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeVid(OutMessage stream, long value)
  {
    stream.write8(value);
  }

  /**
   * Writes a sender identifier on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeSid(OutMessage stream, long value)
  {
    stream.write8(value);
  }

  /**
   * Writes a length value on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeLen(OutMessage stream, int value)
  {
    stream.write4(value);
  }

  /**
   * Writes a version number on the output stream.
   *
   * @param     stream    the message output stream where the value is written.
   * @param     value       value to be written.
   */
  static void writeVer(OutMessage stream, int value)
  {
    stream.write4(value);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Additional methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a new complete view identifier obtained from the host identifier and
   * a sequence number.
   *
   * @param     host    the host idientifer.
   * @param     counter sequence number.
   */
  static long createsVid(int hid, int counter)
  {
    return ((long)(hid) << 32) | (counter & 0xFFFFFFFFL);
  }

  /**
   * Returns a new partial view identifier obtained by a complete view identifier
   * and an additional sequence number.
   *
   * @param     cvid    the complete view identifier.
   * @param     counter the additional counter.
   */
  static long computePartialVid(long cvid, int counter)
  {
    return (cvid + (long) counter);
  }

  /**
   * Returns the checksum of a sequence of bytes starting at position <code> start</code>
   * and ending at position <code> stop</code> of array <code> data</code>.
   * index and an ending index.
   *
   * @param data        the byte array
   * @param start       the starting index.
   * @param stop        the ending index.
   */
  static int computeChecksum(byte[] data, int start, int stop)
  {
    byte        check = 0;                      // Return value

    for (int i=start; i<stop; i++)
      check ^= data[i];
    return check;
  }

} // END Types
