/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.gm;

import java.io.*;
import java.net.*;
import java.util.*;

import jgroup.*;
import jgroup.multicast.*;
import jgroup.reference.*;
import relacs.config.*;
import relacs.mss.*;
import relacs.daemon.*;
import relacs.general.*;
import debug.Debug;

/**
 * The <code> BaseLayer</code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1 (Revised Jgroup 0.9)
 */
public final class      BaseLayer
implements              MembershipService, MulticastService, GMTag
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  static final int S_UNUSED     = 0;    // No group at the moment
  static final int S_JOINING    = 1;    // mcast, join are forbidden
  static final int S_NORMAL     = 2;    // join is forbidden
  static final int S_LEAVING    = 3;    // mcast, join, leave are forbidden


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  // Listeners
  private ArrayList             membershipArray;      // Dynamic array of membership listeners
  private ArrayList             multicastArray;       // Dynamic array of multicast listeners
  private MembershipListener[]  membershipListeners;  // Reference to membership listeners
  private MulticastListener[]   multicastListeners;   // Reference to multicast listeners

  // Objects used to provide group membership
  private MemberIdImpl      me;               // Member identifier
  private Daemon            daemon;           // Reference to local daemon
  private DispatcherService dispatcher;       // Dispatcher
  private MemberTable       table;            // MemberTable provided to upper layers

  // Group information
  private int         status;               // Current status of membership
  private int         gid;                  // Group identifier
  private int         mpos;                 // The (local) position in the current view
  private int         vpos;                 // The position in the current view


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Initialize a new <tt>BaseLayer</tt> object using the informations contained in the
   * configuration objects.
   *
   * @param ds          Description of the distributed system
   * @param conf        Configuration parameters for the daemon used
   *                    by the group manager.
   *
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   * @exception IOException
   *            if an I/O error occurs when creating communication sockets.
   */
  private BaseLayer(Object[] configuration, DispatcherService dispatcher)
    throws JgroupException
  {
    RelacsDS      rds = null;
    RelacsConfig  rconf = null;

    for (int i=0; i < configuration.length; i++) {
      if (configuration[i] instanceof RelacsDS)
        rds = (RelacsDS) configuration[i];
      else if (configuration[i] instanceof DistributedSystem)
        rds = new RelacsDS((DistributedSystem) configuration[i]);
      else if (configuration[i] instanceof RelacsConfig)
        rconf = (RelacsConfig) configuration[i];
      else if (configuration[i] instanceof TransportConfig)
        rconf = new RelacsConfig((TransportConfig) configuration[i]);
    }
    if (rds == null)
      throw new JgroupException("The membership service requires a DistributedSystem description");
    if (rconf == null)
      rconf = new RelacsConfig();

    membershipArray = new ArrayList();
    membershipListeners = new MembershipListener[0];

    multicastArray = new ArrayList();
    multicastListeners = new MulticastListener[0];

    daemon = Daemon.getDaemon(rds, rconf);
    me = (MemberIdImpl) daemon.getMemberId();
    this.dispatcher = dispatcher;
    table = new MemberTable();

    status      = S_UNUSED;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static BaseLayer getLayer(Object[] configuration, DispatcherService dispatcher)
    throws JgroupException
  {
    return new BaseLayer(configuration, dispatcher);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // setListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  /**
   *  Add a listener
   */
  public void addListener(MembershipListener listener)
  {
    membershipArray.add(listener);
    membershipListeners = (MembershipListener[])
      membershipArray.toArray(new MembershipListener[membershipArray.size()]);
  }

  /**
   *  Add a listener
   */
  public void addListener(MulticastListener listener)
  {
    multicastArray.add(listener);
    multicastListeners = (MulticastListener[])
      multicastArray.toArray(new MulticastListener[multicastArray.size()]);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MembershipService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Requests to join the default group associated with the daemon. Each daemon
   * has a default group. If a Jgroup daemon needs not to be shared by multiple
   * groups, application servers may simply join this group without having to
   * specify group identifiers.
   *
   * @exception JgroupException
   *            if the group manager receives a request to join a group for which it
   *            is already member.
   */
  public synchronized void join()
    throws JgroupException
  {
    join1(-1);
  }



  /**
   * Requests to join the default group. When obtaining a group manager using
   * the class <tt>jgroup.GroupManager</tt>, servers must specify a port number
   * through a <tt>jgroup.TransportConfig</tt> object. Each port is associated
   * to a daemon object, which implements the basic group membership functions,
   * as well as the underlying failure detector and transport services.
   * Jgroup enables multiple server objects to share the services provided
   * by a daemon, by enabling them to specify group identifiers as non-negative
   * integers.
   *
   * @param gid   Non-negative integer identifying with the group
   *
   * @exception IllegalArgumentException
   *            If <tt>gid</tt> is negative.
   * @exception JgroupException
   *            if the group manager receives a request to join a group for which it
   *            is already member.
   */
  public synchronized void join(int gid)
    throws JgroupException
  {
    if (gid < 0)
      throw new IllegalArgumentException("Negative group identifier: " + gid);
    join1(gid);
  }

  /**
   * Private implementation of join.
   *
   * @param gid   Non-negative integer identifying with the group
   * @param group Group-related data associated by the application to the group.
   *
   * @exception IllegalArgumentException
   *            If <tt>gid</tt> is negative.
   * @exception JgroupException
   *            if the group manager receives a request to join a group for which it
   *            is already member.
   */
  private void join1(int gid)
    throws JgroupException
  {
    if (Debug.DEBUG)
      Debug.println(30, "GM: join " + gid);
    if (status != S_UNUSED)
      throw new JgroupException("Already member of group " + gid);
    this.gid = gid;
    status = S_JOINING;
    dispatcher.start();
    daemon.join(gid, me.getCounter(), this);
    while (status != S_NORMAL)
      try { wait(); } catch (Exception e) { }
  }


  /**
   *
   *
   * @exception JgroupException   If the group manager is not member of any group, or is
   *            already leaving the current group.
   */
  public synchronized void leave()
    throws JgroupException
  {
    if (Debug.DEBUG)
      Debug.println(30, "GM: leave ");
    if (status == S_UNUSED)
      throw new JgroupException("The gm is not member of any group");
    if (status == S_LEAVING)
      throw new JgroupException("The gm has already received a request to leave the group");
    status = S_LEAVING;
    daemon.leave(gid, me.getCounter());
  }


  public MemberId getMyIdentifier()
  {
    return me;
  }


  public MemberTable getMemberTable()
  {
    return table;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastService
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Multicast a message enclosed in <tt>stream</tt> to the members
   * of the current view of the group.
   *
   * @param stream  the message
   * @param ackl    ack listener to be notified; if null, no ack listener
   *                is notified.
   *
   * @exception UnsupportedOperationException
   *            Raised if the member associated with this group manager
   *            is not a MulticastListener, and thus is not able to receive
   *            message deliveries.
   * @exception JgroupException
   *            if the group manager receives a request to multicast a
   *            message after having joined a group, but before having
   *            received the first view for it; or after having
   *            requested to leave a group.
   */
  public void mcast(OutputStream stream, AckListener ackl)
    throws JgroupException
  {
    OutMessage msg;

    if (multicastListeners == null)
      throw new UnsupportedOperationException("Multicast not supported");
    if (status == S_UNUSED)
      throw new JgroupException("The gm is not member of any group");
    if (status != S_NORMAL)
      throw new JgroupException("Member joining or leaving");

    if (Debug.DEBUG)
      Debug.println(30, "GM: mcast output stream");
    try {
      msg = (OutMessage) stream;
      daemon.mcast(gid, false, msg, me.getLongIdentifier(), ackl);
    } catch (ClassCastException e) {
      throw new JgroupException("Only OutputStream obtained through getMessage() may be multicast");
    }
  }

  /**
   * Multicast a serializable object to the members of the current view of
   * a group.
   *
   * @param obj         the object that must be multicast
   * @param ackl        ack listener to be notified; if null, no ack listener
   *                    is notified.
   *
   * @exception UnsupportedOperationException
   *            Raised if the member associated with this group manager
   *            is not a MulticastListener, and thus is not able to receive
   *            message deliveries.
   * @exception JgroupException
   *            if the group manager receives a request to multicast a
   *            message after having joined a group, but before having
   *            received the first view for it; or after having
   *            requested to leave a group.
   * @exception IOException
   *            Raised in the case of serialization problems
   */
  public void mcast(Object obj, AckListener ackl)
    throws JgroupException, IOException
  {
    if (multicastListeners == null)
      throw new UnsupportedOperationException("Multicast not supported");
    if (status == S_UNUSED)
      throw new JgroupException("The gm is not member of any group");
    if (status != S_NORMAL)
      throw new JgroupException("Member joining or leaving");

    if (Debug.DEBUG)
      Debug.println(30, "GM: mcast object");
    OutMessage bout = (OutMessage) getMessage();
    ObjectOutputStream out = new ObjectOutputStream(bout);
    out.writeObject(obj);
    out.close();
    daemon.mcast(gid, true, bout, vpos, ackl);
  }


  /**
   *  Returns an output stream to be used to encode the message.
   *  Method <tt>mcast(OutputStream, AckListener)</tt> accepts only
   *  multicast messages obtained with this method.
   */
  public OutputStream getMessage()
  {
    if (multicastListeners == null)
      throw new UnsupportedOperationException("Multicast not supported");
    return daemon.getMessage();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // RemoteMember methods
  ////////////////////////////////////////////////////////////////////////////////////////////


  public void notifyDlvr(int gid, boolean flag, long vid, int hpos, int mid, long sid,
                         boolean ackr, InputStream stream)
  {
    UpcallDlvr upcall = new UpcallDlvr(gid, flag, vid, hpos, mid, sid, ackr, stream);
    dispatcher.insert(upcall);
  }

  public void notifyView(int gid, View view, int last, int hpos, int mpos)
  {
    UpcallView upcall = new UpcallView(gid, view, last, hpos, mpos);
    dispatcher.insert(upcall);
  }

  public void notifyPrepare(int gid)
  {
    Upcall upcall = new Upcall(L_PREPARE, gid);
    dispatcher.insert(upcall);
  }

  public void leaveAck(int gid)
  {
    Upcall upcall = new Upcall(L_LEAVEACK, gid);
    dispatcher.insert(upcall);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Dispatcher handle
  ////////////////////////////////////////////////////////////////////////////////////////////

  void dispatch(Upcall upcall)
  {
    switch (upcall.getTag()) {
      case L_DLVR:
        if (multicastListeners != null)
          handleDlvr((UpcallDlvr) upcall);
        else
          System.err.println("Warning: received a multicast message addressed to a non-multicast member");
        break;
      case L_VIEW:
        handleView((UpcallView) upcall);
        break;
      case L_PREPARE:
        handlePrepare((Upcall) upcall);
        break;
      case L_LEAVEACK:
        handleLeaveAck((Upcall) upcall);
        break;
    }
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   *  Delivers the message to the application layer. Note that the
   *  byte structure are condivided up by all of the members.
   */
  protected synchronized void handleDlvr(UpcallDlvr upcall)
  {
    if (Debug.DEBUG)
      Debug.println(30, "GM: handleDlvr");

    if (status != S_UNUSED && status != S_LEAVING) {
      try {
        Object result = null;
        if (upcall.flag) {
          ObjectInputStream in = new ObjectInputStream(upcall.stream);
          Object obj = in.readObject();
          for (int i=0; i < multicastListeners.length; i++) {
            try {
              result = multicastListeners[i].deliverObject(obj);
              i = multicastListeners.length;
            } catch (UnsupportedOperationException e) {  }
          }
        }
        else {
          upcall.stream.mark(0);
          for (int i=0; i < multicastListeners.length; i++) {
            try {
              result = multicastListeners[i].deliverStream(upcall.stream);
              i = multicastListeners.length;
            } catch (UnsupportedOperationException e) { upcall.stream.reset(); }
          }
        }
        if (!upcall.ackr)
          result = null;
        daemon.notifyDlvrAck(upcall.getId(), upcall.hpos, mpos, vpos, upcall.mid, upcall.sid, result);
      } catch (Exception e) {
        /*
         * An exception has been thrown during message deliver; we just print a warning
         * message on the standard error.
         */
        System.err.println("Warning! Exception caught during message unmarshaling: " + e);
        e.printStackTrace(); // TEMP
      }
    }
  }


  /*
   * Installs a new view sent by the daemon.
   */
  private synchronized void handleView(UpcallView upcall)
  {

    if (Debug.DEBUG)
      Debug.println(30, "GM: handleView" + upcall.getId());
    if (status != S_UNUSED && status != S_LEAVING) {

      // Save some view information
      mpos = upcall.mpos;
      MemberId[] members = upcall.view.getMembers();
      for (int i=0; i < members.length; i++) {
        if (members[i].equals(me)) {
          vpos = i;
          break;
        }
      }
      status = S_NORMAL;

      // Update member table
      table.viewChange(upcall.view);

      // Notify members
      for (int i=0; i < membershipListeners.length; i++)
        membershipListeners[i].viewChange(upcall.view);

      // Sends ack to daemon
      daemon.notifyViewAck(upcall.getId(), upcall.mpos, upcall.last);

      // Notify the member waiting for join to complete
      try { notify(); } catch (Exception e) { }

    }
  }


  /*
   *
   */
  private synchronized void handlePrepare(Upcall upcall)
  {
    if (Debug.DEBUG)
      Debug.println(30, "GM: handlePrepare");
    if (status != S_UNUSED && status != S_LEAVING) {
      for (int i=0; i < membershipListeners.length; i++)
        membershipListeners[i].prepareChange();
      daemon.notifyPrepareAck(upcall.getId(), mpos);
    }
  }


  private synchronized void handleLeaveAck(Upcall upcall)
  {
    if (Debug.DEBUG)
      Debug.println(30, "GM: leaveAck");
    // If the group manager has not been halted, and is still member of some
    // group, notify the member that the group has been left.
    if (status != S_UNUSED) {
      for (int i=0; i < membershipListeners.length; i++)
        membershipListeners[i].leaved();
      status = S_UNUSED;
    }
    dispatcher.halt();
  }

} // END BaseLayer



