/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.io.*;
import java.net.*;

import relacs.config.*;
import relacs.general.*;
import debug.Debug;


/**
 * The <code> DataLink </code> class
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.10
 */
final class     Datalink
extends         Thread
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private Ehandler              ehandler;       // Event handler
  private InetAddress           myinetaddr;     // Message identifier
  private DatagramSocket        socket;         // Datagram socket
  private boolean               carryOn;        // The thread must terminate
  private int                   port;           // Socket port
  private int                   bufferlen;      // Buffer length
  private int                   copylimit;      // Limit value under which the buffer is copied


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs and starts the datalink layer.
   * @param name        thread identifier.
   * @param ehandler    ehandler reference (used to generate upcalls).
   * @param port        port number for daemon communications
   * @param bufferlen   size of internal buffers when receiving messages
   */
  public Datalink(String name, Ehandler ehandler, int port, int bufferlen)
        throws IOException
  {
    super(name);

    this.ehandler  = ehandler;
    this.carryOn   = true;
    this.port      = port;
    this.bufferlen = bufferlen;
    this.copylimit = bufferlen*9/10;

    this.setPriority(MAX_PRIORITY);

    try {
      String classname = System.getProperty("relacs.simulator");
      if (classname == null)
        socket = new DatagramSocket(port);
      else
        socket = new UnreliableSocket(port);
      this.myinetaddr = RelacsDS.getLocalHost();
    } catch (Exception e) {
      e.printStackTrace();
      throw new IOException("Error initializing datalink layer: " + e.getMessage());
    }

   setPriority(MAX_PRIORITY);
   this.start();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Sends a datagram packet through the net.
   * @param address     the destination address
   * @param buffer      an array of bytes containing the message
   * @param blen        actual message length
   */
  public void send(InetAddress address, byte[] buffer, int blen)
  {
    DatagramPacket      packet;         // Packet to send

    packet = new DatagramPacket(buffer, blen, address, port);
    if (address.equals(myinetaddr))
      ehandler.rnotify(packet);
    else
      try {
        socket.send(packet);
      } catch (Exception e) {
        if (Debug.DEBUG)
          Debug.println(0, "Error sending packet:" + e);
      }
  }


  /**
   * Run method of the thread. It blocks until a Datagram packet is received
   * through the net.
   */
  public void run()
  {
    byte[]              buffer;       // Buffer to receive messages
    byte[]              data;         //
    int                 len;
    DatagramPacket      packet;       // Received packet

    buffer = new byte[bufferlen];
    packet = new DatagramPacket(buffer, bufferlen);
    while (carryOn)
      try {
        packet.setLength(bufferlen);
        socket.receive(packet);
        len = packet.getLength();
        if (len > copylimit) {
          ehandler.rnotify(packet);
          buffer = new byte[bufferlen];
          packet = new DatagramPacket(buffer, bufferlen);
        }
        else {
          data = new byte[len];
          System.arraycopy(buffer, 0, data, 0, len);
          ehandler.rnotify(new DatagramPacket(data,len,packet.getAddress(),packet.getPort()));
        }
      } catch (Exception e) {
        if (Debug.DEBUG)
          Debug.println(0, "Error receiving packets: " + e);
      }

  }

  /**
   * Destructor.
   *
   */
  protected void finalize() {
    if (socket != null) {
      socket.close();
      socket = null;
    }
    carryOn = false;
  }

} // END Datalink



