/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

import java.net.*;
import java.io.*;

import relacs.general.*;
import debug.Debug;


/**
 * The <code> Ehandler</code> class handles all events generated
 * internally by the timer and externally by the local members
 * and by the datalink thread.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.1
 */
final class     Ehandler
extends         Thread
implements      MssTag
{
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static Header         header;
  static  { header = new Header(); }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  private EhandlerUser    ehuser;                 // Upper layer
  private Datalink        remote;                 // Remote events thread
  private EventList       events;                 // Lista di eventi
  private Queue           lmessages;              // Remote messages queue
  private Queue           rmessages;              // Member messages queue
  private DatagramPacket  rmsg;                   // Next remote message
  private Upcall          lmsg;                   // Next local message


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Constructs and starts the event handler.
   * @param name        the name of the datalink thread
   * @param ehuser      reference to the event handler user (used to generate upcalls).
   * @param port        port number for daemon communications
   * @param bufferlen   size of internal buffers when receiving messages
   */
  Ehandler(String name, EhandlerUser ehuser, int port, int bufferlen)
        throws IOException
  {
    super(name);

    this.ehuser   = ehuser;
    events    = new EventList();
    lmessages = new Queue();
    rmessages = new Queue();
    remote    = new Datalink("Dlink", this, port, bufferlen);
    this.start();
  }

  /**
   * Inserts a new timeout event
   * @param timeout     timeout length in milliseconds
   * @param event       event description
   */
  void setTimeout(int timeout, EventList event)
  {
    events.insert(System.currentTimeMillis()+timeout, event);
  }

  /**
   * Sends a packet to a remote host.
   * @param address     the destination address.
   * @param buffer      an array of bytes containing the message.
   * @param blen        actual message length.
   */
  void rsend(InetAddress address, byte[] buffer, int blen)
  {
    if (Debug.DEBUG) {
      header.read(buffer);
      Debug.println((header.tag == MSS_RT || header.tag==MSS_NACK ? 5 : 0),
                    "Send| Tag: " + header.tag + " Dest: " + (header.dest & 0xff) +
                    " Msg id: " + header.mid);
    }
    remote.send(address, buffer, blen);
  }


  /**
   * Sends messages for local members by storing them in the <code> messages
   * <\code> queue.
   * @param msg         local message to send
   */
  synchronized void lnotify(Object msg)
  {
    lmessages.insert(msg);
    notify();
  }

  /**
   * Sends messages for remote members by storing them in the <code> messages
   * <\code> queue.
   * @param msg         local message to send
   */
  synchronized void rnotify(Object msg)
  {
    rmessages.insert(msg);
    notify();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Run method of the thread
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Run method of the thread. It remains blocked until (i) a Datagram packet is received
   * through the net; (ii) a message from a local member is received; (iii) a scheduled
   * timeout expires.
   */
  public void run()
  {
    Object              msg;                    // Local message
    long                current;                // Current time
    long                next;                   // Next event time
    byte[] buf;

    while (true) {
			if (await(System.currentTimeMillis(), events.getFirst())) {
				// System.out.print("!");
        if (Debug.DEBUG)
          Debug.println(0, "treceive start");
        ehuser.treceive(events.removeFirst());
        if (Debug.DEBUG)
          Debug.println(0, "treceive end");
      }
      if (rmsg != null) {
				// System.out.print("r");
        if (Debug.DEBUG)
          Debug.println(0, "rreceive start");
        ehuser.rreceive(rmsg.getAddress(), rmsg.getData(), rmsg.getLength());
        if (Debug.DEBUG)
          Debug.println(0, "rreceive end");
        rmsg = null;
      }
      if (lmsg != null) {
				// System.out.print("l");
        if (Debug.DEBUG)
          Debug.println(0, "lreceive start");
        ehuser.lreceive(lmsg);
        if (Debug.DEBUG)
          Debug.println(0, "lreceive end");
        lmsg = null;
      }
      // yield();
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private synchronized boolean await(long current, long next)
  {
    long        timeout;

    if (rmsg == null)
      rmsg = (DatagramPacket) rmessages.removeFirst();
    if (lmsg == null)
      lmsg = (Upcall) lmessages.removeFirst();
    if (rmsg == null && lmsg == null) {
      if (next == -1) {
        if (Debug.DEBUG)
          Debug.println(0, "before wait()");
        try { wait(); } catch (InterruptedException e) { };
        if (Debug.DEBUG)
          Debug.println(0, "after wait()");
      }
      else {
        timeout = next - current;
        if (timeout > 0) {
          if (Debug.DEBUG)
            Debug.println(0, "before wait()");
          try { wait(timeout); } catch (InterruptedException e) { };
          if (Debug.DEBUG)
            Debug.println(0, "after wait()");
        }
      }
    }
    if (Debug.DEBUG) {
      if (next != -1 && next < System.currentTimeMillis())
        Debug.println(0, "Timeout expired: " + (System.currentTimeMillis() - next) + " msec");
    }
    return (next != -1 && next < System.currentTimeMillis());
  }

  void printevents()
  {
    if (Debug.DEBUG) {
      EventList event = events.next;
      Debug.print(10, "Elist: ");
      while (event != null) {
        Debug.print(10, (event.value % 100000) + "(" + (event.prev.value % 100000) +"): ");
        event = event.next;
      }
      Debug.println(10, "");
    }
  }

} // END Ehandler

