/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.mss;

/**
 * The <code>Types</code> class contains a collection of psuedo-types, all of them
 * are implemented by the primitive type <code>int</code>. All of these types
 * are supplied with a coding function (<code>xxx2byte()</code>) that encodes the
 * value into a byte array and a decoding function (<code>byte2xxx()</code>)
 * that decodes a byte array and returns a value. In this way, it is possible to
 * modify the coding algorithm of each of these values without the classes accessing
 * them have to be informed.
 *
 * @author  Alberto Montresor
 * @since   Javagroup 1.0
 */

class Types
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Sizes in bytes of integer types
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Size in byte of pseudo-type Tag, used to identify messages */
  static final int TAG_SIZE     = 1;

  /** Size in byte of pseudo-type Hid, used to encode host identifiers (internet addresses) */
  static final int HID_SIZE     = 4;

  /** Size in byte of pseudo-type Inc, used to encode incarnation numbers */
  static final int INC_SIZE     = 4;

  /** Size in byte of pseudo-type Mid, used to identify messages */
  static final int MID_SIZE     = 4;

  /** Size in byte of pseudo-type Grp, used to describe the number of members in a group */
  static final int DIM_SIZE     = 2;

  /** Size in byte of pseudo-type Chk, used to encode checksum */
  static final int CHK_SIZE     = 1;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods used by mss to write header and trailers.
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Decodes a sequence of bytes into a host identifier.
   *
   * @param     data    sequence of bytes to be read.
   * @param     pos   starting position of bytes to be decoded.
   * @return    the decoded value.
   */
  static int byte2hid(byte[] data, int pos)
  {
    return      ((data[pos++] << 24) & 0xFF000000) | ((data[pos++] << 16)
                & 0xFF0000) | ((data[pos++] << 8) & 0xFF00) | (data[pos] & 0xFF);
  }


  /**
   * Decodes a sequence of bytes into a message identifier.
   *
   * @param     data    sequence of bytes to be read.
   * @param     pos     starting position of bytes to be decoded.
   * @return    the decoded value.
   */
  static int byte2mid(byte[] data, int pos)
  {
    return      ((data[pos++] << 24) & 0xFF000000) | ((data[pos++] << 16)
                & 0xFF0000) | ((data[pos++] << 8) & 0xFF00) | (data[pos] & 0xFF);
  }


  /**
   * Decodes a sequence of bytes into an incarnation number.
   *
   * @param     data    sequence of bytes to be read.
   * @param     pos     starting position of bytes to be decoded.
   * @return    the decoded value.
   */
  static int byte2inc(byte[] data, int pos)
  {
    return      ((data[pos++] << 24) & 0xFF000000) | ((data[pos++] << 16)
                & 0xFF0000) | ((data[pos++] << 8) & 0xFF00) | (data[pos] & 0xFF);
  }

  /**
   * Decodes a sequence of bytes into a dimension value.
   *
   * @param     data    sequence of bytes to be read.
   * @param     pos     starting position of bytes to be decoded.
   * @return    the decoded value.
   */
  public static int byte2dim(byte[] data, int pos)
  {
    return      ((data[pos++] << 8) & 0xFF00) | (data[pos] & 0xFF);
  }

  /**
   * Decodes a sequence of bytes into a checksum value.
   *
   * @param     data    sequence of bytes to be read.
   * @param     pos     starting position of bytes to be decoded.
   * @return    the decoded value.
   */
  static int byte2chk(byte[] data, int pos)
  {
    return      data[pos];
  }

  /**
   * Encodes a host identifier into a sequence of bytes.
   *
   * @param     data    destination byte array.
   * @param     pos     starting position where bytes have to be written.
   * @return    starting position after writing.
   */
  static int hid2byte(byte[] data, int pos, int value)
  {
    data[pos++] = (byte) (value >>> 24);
    data[pos++] = (byte) (value >>> 16);
    data[pos++] = (byte) (value >>> 8);
    data[pos++] = (byte) (value >>> 0);
    return pos;
  }

  /**
   * Encodes a incarnation number into a sequence of bytes.
   *
   * @param     data    destination byte array.
   * @param     pos     starting position where bytes have to be written.
   * @return    starting position after writing.
   */
  static int inc2byte(byte[] data, int pos, int value)
  {
    data[pos++] = (byte) (value >>> 24);
    data[pos++] = (byte) (value >>> 16);
    data[pos++] = (byte) (value >>> 8);
    data[pos++] = (byte) (value >>> 0);
    return pos;
  }

  /**
   * Encodes a message identifier into a sequence of bytes.
   *
   * @param     data    destination byte array.
   * @param     pos     starting position where bytes have to be written.
   * @return    starting position after writing.
   */
  static int mid2byte(byte[] data, int pos, int value)
  {
    data[pos++] = (byte) (value >>> 24);
    data[pos++] = (byte) (value >>> 16);
    data[pos++] = (byte) (value >>> 8);
    data[pos++] = (byte) (value >>> 0);
    return pos;
  }

  /**
   * Encodes a dimension value into a sequence of bytes.
   *
   * @param     data    destination byte array.
   * @param     pos     starting position where bytes have to be written.
   * @return    starting position after writing.
   */
  static int dim2byte(byte[] data, int pos, int value)
  {
    data[pos++] = (byte) (value >>> 8);
    data[pos++] = (byte) (value >>> 0);
    return pos;
  }

  /**
   * Encodes a checksum value into a sequence of bytes.
   *
   * @param     data    destination byte array.
   * @param     pos     starting position where bytes have to be written.
   * @return    starting position after writing.
   */
  static int chk2byte(byte[] data, int pos, int value)
  {
    data[pos++] = (byte) value;
    return pos;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Additional methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns the checksum of a sequence of bytes starting at position <code> start</code>
   * and ending at position <code> stop</code> of array <code> data</code>.
   * index and an ending index.
   *
   * @param data        the byte array
   * @param start       the starting index.
   * @param stop        the ending index.
   */
  static int computeChecksum(byte[] data, int start, int stop)
  {
    byte        check = 0;                      // Return value

    for (int i=start; i<stop; i++)
      check ^= data[i];
    return check;
  }

} // END Types
