/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.rmi.*;
import java.util.*;
import java.lang.reflect.*;

import jgroup.*;
import relacs.rmi.proxy.*;

/**
 * Generic proxy class.
 */
public class Exporter
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static variables
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** True if the class library contains dynamic proxy classes (1.3 or greater) */
  private static boolean hasDynamicProxy = false;

  /** Class object of class java.lang.reflect.Proxy */
  private static Class proxyClass;

  /** Constructor object for class java.lang.reflect.Proxy */
  private static Method proxyMethod;

  /** Class object of class relacs.rmi.ExtGroupHandler3 */
  private static Class handlerClass;

  static {
    try {
      proxyClass = Class.forName("java.lang.reflect.Proxy");
      Class cl = Class.forName("java.lang.reflect.InvocationHandler");
      Class[] types = new Class[] { ClassLoader.class, Class[].class, cl };
      proxyMethod = proxyClass.getDeclaredMethod("newProxyInstance", types);
      handlerClass = Class.forName("relacs.rmi.ExtGroupHandler3");
      hasDynamicProxy = true; 
    } catch (Exception e) { /* Do nothing; hasDynamicProxy is false */}
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Method
  ////////////////////////////////////////////////////////////////////////////////////////////



  /**
   *  Return a group handler for the class, depending on whether dynamic
   *  proxy classes are present or not.
   */
  public static ExtGroupHandler2 getHandler()
  {
    if (!hasDynamicProxy) {
      return new ExtGroupHandler2();
    }
    try {
      return (ExtGroupHandler2) handlerClass.newInstance();
    } catch (InstantiationException e) {
      e.printStackTrace();
      System.exit(0);
      return null;
    } catch (IllegalAccessException e) {
      e.printStackTrace();
      System.exit(0);
      return null;
    }
  }

  /**
   *  Returns a proxy for the server object
   *
   *  If object <tt>server</tt> does not implement
   *  any external interface (i.e., an interface that extends <tt>jgroup.GMIExternal</tt>).
   *  The exception is thrown only when dynamic proxy classes are used (when static classes
   *  are used, this check is performed when generating the stub for the server object.
   *
   *  @exception RemoteException Thrown if for some reason, it is impossible to instantiate
   *    a proxy class.
   */
  public static Remote getProxy(Class serverClass, JgroupHandler handler)
    throws RemoteException
  {
    /* Warning message */
    String warning = "";
    
    /* Static proxy */
    String name = serverClass.getName()+"_ClientProxy"; 
    
    try {
      warning = "Unable to instantiate a static proxy for class " +
        serverClass.getName() + ": ";
      proxyClass = Class.forName(name);
      Class[] types = new Class[] { relacs.rmi.proxy.JgroupHandler.class };
      Constructor proxyConstructor = proxyClass.getConstructor(types);
      return (Remote) proxyConstructor.newInstance(new Object[] { handler });
    } catch (ClassNotFoundException e) {
      if (!hasDynamicProxy) {
        throw new RemoteException(warning + "Proxy class not found: " + name, e);
      }
    } catch (NoSuchMethodException e) {
      if (!hasDynamicProxy) {
        throw new RemoteException(warning + "Proxy class missing constructor: " + name, e);
      }
    } catch (InstantiationException e) {
      if (!hasDynamicProxy) {
        throw new RemoteException(warning + "Can't create instance of proxy class: " + name, e);
      }
    } catch (IllegalAccessException e) {
      if (!hasDynamicProxy) {
        throw new RemoteException(warning + "Proxy class constructor not public: " + name, e);
      }
    } catch (InvocationTargetException e) {
      if (!hasDynamicProxy) {
        throw new RemoteException(warning + "Exception creating instance of proxy class: " + name, e);
      }
    }
    /**
     *  If we get here, this means that (1) no static proxy has been found, 
     *  and (2) we are running on Jdk 1.3, since dynamic proxy classes are 
     *  available.
     */
    warning = "Unable to instantiate a dynamic proxy for class " + 
        serverClass.getName() + ": ";
    Class[] interfaces = serverClass.getInterfaces();
    ArrayList array = new ArrayList();
    for (int i=0; i < interfaces.length; i++)
      if (ExternalGMIListener.class.isAssignableFrom(interfaces[i]))
        array.add(interfaces[i]);
    interfaces = (Class[]) array.toArray(new Class[array.size()]);
		if (interfaces.length == 0)
      throw new RemoteException(warning + "class " + serverClass.getName() +
        "does not implement any external GMI interface");
		return getProxy(interfaces, handler);
  }

	
  /**
   *  Returns a proxy for the server object
   *
   *  If object <tt>server</tt> does not implement
   *  any external interface (i.e., an interface that extends <tt>jgroup.GMIExternal</tt>).
   *  The exception is thrown only when dynamic proxy classes are used (when static classes
   *  are used, this check is performed when generating the stub for the server object.
   *
   *  @exception RemoteException Thrown if for some reason, it is impossible to instantiate
   *    a proxy class.
   */
  public static Remote getProxy(Class[] interfaces, JgroupHandler handler)
    throws RemoteException
  {
    /* Warning message */
		if (interfaces == null || interfaces.length == 0)
			throw new RemoteException("Unable to instantiate a dynamic proxy: no interfaces provided");
		String warning = "Unable to instantiate a dynamic proxy for interfaces";
		for (int i=0; i<interfaces.length; i++) 
			warning = warning + " " +interfaces[i].getName();
		warning = warning + ": ";
		Object[] args = new Object[] { interfaces[0].getClassLoader(), interfaces, handler };
    try {
      return (Remote) proxyMethod.invoke(null, args);
    } catch (InvocationTargetException e) {
      throw new RemoteException(warning + e.getTargetException().getMessage());
    } catch (IllegalArgumentException e) {
				e.printStackTrace();
      throw new RemoteException(warning + "internal error! " + e.getMessage());
    } catch (IllegalAccessException e) {
      throw new RemoteException(warning + "internal error!" + e.getMessage());
    }
	}
	
} // END Exporter
