/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.io.*;
import java.rmi.*;
import java.util.*;
import java.lang.reflect.*;

import jgroup.*;
import jgroup.multicast.*;

import relacs.gm.*;
import relacs.rmi.proxy.*;
import relacs.general.*;

import debug.Debug;

/**
 *
 *
 *  @author   Alberto Montresor
 *  @since    Jgroup 0.9
 */
public class IntGroupHandler2
implements JgroupHandler, MulticastListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Method invoke of interface InternalGMIService */
  private static Method invokeMethod;
  
  /** Method getInvocationHandler of interface  InternalGMIService */
  private static Method getHandlerMethod;     

  static {
    try {
      Class[] types = new Class[] { Method.class, Object[].class, Callback.class };
      Class cl = InternalGMIService.class;
      invokeMethod = cl.getDeclaredMethod("invoke", types);
    } catch (NoSuchMethodException e) {
      // Can't happen, otherwise InternalGMIService has been modified; so, we panic!
      System.err.println("Method invoke in class InternalGMIService not found");
      e.printStackTrace();
      System.exit(1);
    }
    try {
      Class[] types = new Class[] {  };
      Class cl = HandlerGetter.class;
      getHandlerMethod = cl.getDeclaredMethod("getInvocationHandler", types);
    } catch (NoSuchMethodException e) {
      // Can't happen, otherwise HandlerGetter has been modified; so, we panic!
      System.err.println("Method getInvocationHandler in interface HandlerGetter not found");
      e.printStackTrace();
      System.exit(1);
    }

  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  private ServerMethodTable methodTable;

  /**
   *
   */
  private MulticastService multicastService;

  /**
   *
   */
  private DispatcherService dispatcherService;

  /*
   * Layer listening for internal invocation
   */
  private InternalGMIListener server;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a <tt>IntGroupHandler2</tt> by storing the
   *  method table.
   */
  public IntGroupHandler2(MulticastService multicastService, DispatcherService dispatcherService)
  {
    this.dispatcherService = dispatcherService;
    this.multicastService = multicastService;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static factory
  ////////////////////////////////////////////////////////////////////////////////////////////

  public static IntGroupHandler2 getLayer(Object[] configuration, MulticastService
    multicastService, DispatcherService dispatcherService)
  {
    return new IntGroupHandler2(multicastService, dispatcherService);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // setListener methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*
   */
  public void addListener(InternalGMIListener server)
    throws JgroupException
  {
    if (this.server != null)
      throw new JgroupException("Error composing group managers layers; multiple local " +
        " listeners for group method invocations");

    /* Store the server reference for future invocations */
    this.server = server;

    /* Compute the external interfaces implemented by this object */
    Class[] interfaces = server.getClass().getInterfaces();
    ArrayList array = new ArrayList();
    for (int i=0; i < interfaces.length; i++)
      if (InternalGMIListener.class.isAssignableFrom(interfaces[i]))
        array.add(interfaces[i]);
    interfaces = (Class[]) array.toArray(new Class[array.size()]);

    /* Compute the method table for these interfaces */
    methodTable = new ServerMethodTable(interfaces);
  }



  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from InvocationHandler interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Processes a method invocation on a proxy instance and returns
   * the result.  This method will be invoked on an invocation handler
   * when a method is invoked on a proxy instance that it is
   * associated with.
   */
  public Object invoke(Object proxy, Method m, Object[] args)
    throws Throwable
  {
    boolean asynchronous = false;
    Callback callback = null;

    /*
     * Check whether the method is getInvocationHandler, in which
     * case we just return 'this'.
     */
    if (m.equals(getHandlerMethod)) {
      return this;
    }

    /*
     * Check whether the method has asynchronous semantics; if so,
     * read the arguments
     */
    if (m.equals(invokeMethod)) {
      asynchronous = true;
      m = (Method) args[0];
      callback = (Callback) args[2];
      args = (Object[]) args[1];
    }

    try {

      /*
       * Prepare message
       */
      long hash = methodTable.get(m);
      OutputStream message = multicastService.getMessage();
      ObjectOutputStream out = new MarshalOutputStream(message);
      out.writeLong(hash);
      int len = (args == null ? 0 : args.length); 
      out.writeInt(len);
      for (int i=0; i < len; i++)
        out.writeObject(args[i]);
      out.flush();

      if (asynchronous) {
        // Asynchronous semantics
        if (callback != null)
          multicastService.mcast(message, new AsynchAckListener(callback));
        else
          multicastService.mcast(message, null);
        return null;
      } else {
        // Synchronous semantics
        SynchAckListener listener = new SynchAckListener(dispatcherService);
        multicastService.mcast(message, listener);
        return listener.getResults();
      }

    } catch (JgroupException e) {
      throw e;
    } catch (Exception e) {
      e.printStackTrace();
      throw new JgroupException("Unable to perform internal group method invocations: " + e);
    }

  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from MulticastListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> msg</code>
   *  related to the group identified by <code> group</code>.
   *
   *  @param msg      the stream from which the message may be read.
   */
  public Object deliverStream(InputStream msg)
  {
    if (Debug.DEBUG)
      Debug.println(40, "IntGMILayer: received stream message");
    // Bisogna controllare la gestione delle eccezioni
    try {
      ObjectInputStream in = new MarshalInputStream(msg);
      long hash = in.readLong();
      int size = in.readInt();
      Object[] args = new Object[size];
      for (int i=0; i < args.length; i++)
        args[i] = in.readObject();
      /* Obtain the method from the method hash */
      Method m = methodTable.get(hash);
      try {
        Object result = m.invoke(server, args);
        return result;
      } catch (Exception e) {
        return e;
      }
    } catch (Exception e) {
      if (Debug.DEBUG) {
        Debug.println(40, "IntGMILayer: messaged message, but not for me!");
        Debug.printStackTrace(40, e);
      }
      throw new UnsupportedOperationException();
    }
  }


  /**
   *  Upcall that is invoked by Jgroup to deliver a message <code> obj</code>
   *  related to the group identified by <code> group</code>. In this case,
   *  the message is an object marshalled and unmarshalled through java
   *  serialization.
   *
   *  @param obj        the object containing the message.
   */
  public Object deliverObject(Object obj)
  {
    throw new UnsupportedOperationException();
  }

} // END IntGroupHandler2
