/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.lang.reflect.*;
import java.security.*;

import jgroup.*;

/**
 *  Container for a hash value and boolean value
 */
final class MethodInfo
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Hash identifier of the method */
  private long  hash;

  /** True if the method is a multicast method */
  private boolean multicast;

  /** True if the method is a void method */
  private boolean isvoid;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Fill the fields with the arguments
     */
  MethodInfo(Method method)
  {
    final Method m = method;

    /*
     * Set this Method object to override language
     * access checks so that the dispatcher can invoke
     * methods from non-public remote interfaces.
     */
    AccessController.doPrivileged(new PrivilegedAction()
    {
      public Object run() {
        m.setAccessible(true);
        return null;
      }
    });

    /**
     * Compute the "method hash" of a remote method.  The method hash
     * is a long containing the first 64 bits of the SHA digest from
     * the UTF encoded string of the method name and descriptor.
     */
    hash = ServerMethodTable.computeHash(m);

    /**
     *  Verifies if this method has multicast semantics; i.e.,
     *  checks if the McastRemoteException is in the throws
     *  clause of the method.
     */
    multicast = false;
    Class[] exceptions = m.getExceptionTypes();
    for (int k=0; k < exceptions.length && !multicast; k++)
      if (exceptions[k].equals(McastRemoteException.class))
        multicast = true;

    /**
     *  Verifies if this method returns void.
     */
    Class cl = m.getReturnType();
    isvoid = cl.equals(Void.TYPE);

  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the hash value associated to this method
   */
  public long getHash()
  {
    return hash;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return true if this method is a multicast method
   */
  public boolean isMulticast()
  {
    return multicast;
  }

  /**
   *  Return true if this method is a multicast method
   */
  public boolean isVoid()
  {
    return isvoid;
  }

} // END MethodInfo


