/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi;

import java.rmi.*;
import java.util.*;


import jgroup.*;
import jgroup.multicast.*;

import relacs.gm.*;

import debug.Debug;

/*
 * FIX 19/11/00:
 * Added debug information to track down a bug of internal group method
 * invocations.
 */

/**
 * This class is used to collect return values of synchronous
 * internal group method invocations. For each invocation, an
 * object of this class is instantiated. When an invocation
 * is performed, method notifyView() is invoked to communicate 
 * to this object the number and the identities of the members
 * included in the view in which the method is executed.
 * Method ack() is used to add the return value returned by
 * the specified member. Method viewChange() is used to notify
 * objects of this class that a new view has been installed
 * and members not responding are to be considered unreachable.
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.9
 */
final class     SynchAckListener
implements      AckListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Array containing results */
  private Object[] results;
  
  /** Boolean array for results */
  private boolean[] completed;
  
  /** Counter */
  private int missing;

  /** Dispatcher used to move the dispatcher */
  private DispatcherService dispatcher;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  SynchAckListener(DispatcherService dispatcher)
  {
    this.dispatcher = dispatcher;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *
   */
  public synchronized void notifyView(View view)
  {
    missing = view.getMembers().length;
    completed = new boolean[missing];
    results = new Object[missing];
    if (Debug.DEBUG)
      Debug.println(40, "SynchAckListener: notifyView: " + missing);
  }

  /**
   *
   */
  public synchronized void ack(MemberId id, int vpos, Object obj)
  {
    if (Debug.DEBUG)
      Debug.println(40, "SynchAckListener: ack: " + id + ", " + obj + ", " + vpos);
    if (!completed[vpos]) {
      completed[vpos] = true;
      results[vpos] = obj;
      missing--;
      if (missing == 0) {
        if (Debug.DEBUG)
          Debug.println(40, "SynchAckListener: ack: completed");
        notifyAll();
      }
    }
  }

  /**
   *
   */
  public synchronized void viewChange()
  {
    if (Debug.DEBUG)
      Debug.println(40, "SynchAckListener: viewChange");
    if (missing > 0) {
      for (int i=0; i < completed.length; i++) {
        if (!completed[i]) {
          completed[i] = true;
          results[i] = new RemoteException("Member not reachable");
          if (Debug.DEBUG)
            Debug.println(40, "SynchAckListener: viewChange: unreachable pos " + i);
        }
      }
    }
    missing = 0;
  }

  /**
   *
   */
  public synchronized Object getResults()
  {
    if (Debug.DEBUG)
      Debug.println(40, "SynchAckListener: getResults");
    while (missing > 0 || results == null) {
      dispatcher.dispatch(this);
    }
    return results;
  }

} // END SynchAckListener

