/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi.registry;

import java.io.*;
import java.net.*;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;

import jgroup.*;
import jgroup.registry.*;
import jgroup.reference.*;
import relacs.config.*;
import relacs.rmi.*;
import relacs.rmi.proxy.*;
import relacs.rmi.registry.*;

import debug.Debug;


/**
 * <code>RelacsRegistryLocator</code> is used to obtain a bootstrap reference to
 * the Jgroup dependable registry, or to create a Jgroup dependable registry instance
 * on the local machine. Every instance of the Jgroup registry service is based on
 * two remote registries: the standard local registry and the replicated one. The
 * standard local registry is used to retrieve a reference for the replicated
 * remote registry. When the argument <code> port </code> is used in a getRegistry
 * or a createRegistry method, it specifies the port used by the local registry.
 * The port used by the replicated registry is identified by <code> port+1 </code>.
 *
 * @see java.rmi.LocateRegistry
 * @author      Alberto Montresor
 * @since       Jgroup 0.3
 */
public final class  RelacsRegistryLocator
implements          RegistryLocator
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Will be used to obtain an instance of RegistryLocator
   */
  public RelacsRegistryLocator()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Public methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds)
    throws RemoteException
  {
    RelacsDS        rds;
    Remote          ref;
    ExtGroupHandler2  handler;
    int             size;
    String[]        names;

    handler = Exporter.getHandler();
    try {
      if (ds instanceof RelacsDS)
        rds = (RelacsDS) ds;
      else
        rds = new RelacsDS(ds);
    } catch (Exception e) {
      throw new RemoteException("Unable to configure the distributed system");
    }

    names = rds.getNames();
    size = rds.size();
    for (int i = 0; i < size; i++) {
      try {
        if (Debug.DEBUG)
          Debug.print(60, names[i] + ": ");
        ref = getSingleRegistry(names[i]);
        handler.insert(ref);
        if (Debug.DEBUG)
          Debug.println(60, "found registry.");
      } catch (Exception e) {
        if (Debug.DEBUG) {
          Debug.println(60, "registry not found.");
          /* Output the exception if the debug level is 50 */
          Debug.println(50, "("+e+")");
        }
      }
    }
    if (handler.size() == 0)
      return null;

    return (DependableRegistry) Exporter.getProxy(RegistryImpl.class, handler);
  }

  /**
   * Returns a reference to the group of dependable registry replicas running in
   * in the specified distributed system and listening on the port specified by
   * <code> port</code>.
   *
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port used by the replicas for RMI communications
   * @return            a reference (stub) to the Jgroup dependable registry
   * @exception RemoteException if the reference could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds, int port)
    throws RemoteException
  {
    RelacsDS        rds;
    ExtGroupHandler2  handler;
    Remote          ref;
    int             size;
    String[]        names;

    handler = Exporter.getHandler();
    try {
      if (ds instanceof RelacsDS)
        rds = (RelacsDS) ds;
      else
        rds = new RelacsDS(ds);
    } catch (Exception e) {
      throw new RemoteException("Unable to configure the distributed system");
    }
    names = rds.getNames();
    size = rds.size();
    for (int i = 0; i < size; i++) {
      try {
        if (Debug.DEBUG)
          Debug.print(60, names[i] + ": ");
        ref = getSingleRegistry(names[i], port);
        handler.insert(ref);
        if (Debug.DEBUG)
          Debug.println(60, "found registry.");
      } catch (Exception e) {
        if (Debug.DEBUG)
          Debug.println(60, "registry not found ("+e+")");
      }
    }

    return (DependableRegistry) Exporter.getProxy(RegistryImpl.class, handler);
  }


  /**
   * Returns a proxy for the dependable registry service on the specified
   * distributed system and port. <p>
   *
   * Communication with this remote registry will use the supplied
   * <t>RMIClientSocketFactory csf</t> to create socket connections
   * to the dependable registry on the distributed system and port.
   * @param ds          the distributed system in which the dependable registry
   *                    replicas are expected to run
   * @param port        port on which distributed replicas are listening
   * @param csf         client-side Socket factory used to make connections to
   *                    the dependable registry. If <t>csf</t> is null, then the
   *                    default client-side socket factory will be used
   * @return            a proxy for the dependable registry service
   * @exception RemoteException if the proxy could not be obtained
   */
  public DependableRegistry getRegistry(DistributedSystem ds, int port,
                                     RMIClientSocketFactory csf)
    throws RemoteException
  {
    RelacsDS        rds;
    ExtGroupHandler2  handler;
    Remote          ref;
    int             size;
    String[]        names;

    handler = Exporter.getHandler();
    try {
      if (ds instanceof RelacsDS)
        rds = (RelacsDS) ds;
      else
        rds = new RelacsDS(ds);
    } catch (Exception e) {
      throw new RemoteException("Unable to configure the distributed system");
    }
    names = rds.getNames();
    size = rds.size();
    for (int i = 0; i < size; i++) {
      try {
        if (Debug.DEBUG)
          Debug.print(60, names[i] + ": ");
        ref = getSingleRegistry(names[i], port, csf);
        handler.insert(ref);
        if (Debug.DEBUG)
          Debug.println(60, "found registry.");
      } catch (Exception e) {
        if (Debug.DEBUG)
          Debug.println(60, "registry not found ("+e+")");
      }
    }

    return (DependableRegistry) Exporter.getProxy(RegistryImpl.class, handler);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on an anonymous communication port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf)
    throws RemoteException, JgroupException
  {
    return new RegistryImpl(conf);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf, int port)
    throws RemoteException, JgroupException
  {
    return new RegistryImpl(conf, port);
  }


  /**
   * Creates and exports a dependable registry replica on the local host that
   * listen to incoming requests on  the given <code>port</code> communication
   * port.
   * Objects contained <tt>conf</tt> are used by Jgroup to configure the
   * group communication; the distributed system description must be
   * inclued in this array.
   * The new instance listens to incoming requests using a
   * <code>ServerSocket</code> created from the supplied
   * <code>RMIServerSocketFactory</code>.  A client
   * that receives a reference to this registry will use a
   * <code>Socket</code> created from the supplied
   * <code>RMIClientSocketFactory</code>.
   *
   * @param conf        configuration information
   * @param port        port used for RMI communications
   * @param csf         client-side <code>Socket</code> factory used to
   *                    make connections to the registry
   * @param ssf         server-side <code>ServerSocket</code> factory
   *                    used to accept connections to the registry
   * @return            the new dependable registry replica.
   * @exception RemoteException if the registry could not be exported
   * @exception JgroupException
   *            Generic exception raised by Jgroup.
   * @exception DSException
   *            raised when the distributed system configuration object contains an error
   */
  public DependableRegistry createRegistry(Object[] conf, int port,
                                           RMIClientSocketFactory csf,
                                           RMIServerSocketFactory ssf)
    throws RemoteException, JgroupException
  {
    return new RegistryImpl(conf, port, csf, ssf);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private static Remote getSingleRegistry(String hostname)
    throws RemoteException, NotBoundException
  {
    Registry            registry = null;

    registry = LocateRegistry.getRegistry(hostname, Registry.REGISTRY_PORT+1);
    return registry.lookup(DependableRegistry.DEPENDABLE_REGISTRY_NAME);
  }

  private static Remote getSingleRegistry(String hostname, int port)
    throws RemoteException, NotBoundException
  {
    Registry            registry = null;

    if (port == 0)
      port = Registry.REGISTRY_PORT+1;
    registry = LocateRegistry.getRegistry(hostname, port);
    return registry.lookup(DependableRegistry.DEPENDABLE_REGISTRY_NAME);
  }


  private static Remote getSingleRegistry(String hostname, int port,  RMIClientSocketFactory csf)
    throws RemoteException, NotBoundException
  {
    Registry            registry = null;

    if (port == 0)
      port = Registry.REGISTRY_PORT+1;
    registry = LocateRegistry.getRegistry(hostname, port, csf);
    return registry.lookup(DependableRegistry.DEPENDABLE_REGISTRY_NAME);
  }

} // END RelacsRegistryLocator
