/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package relacs.rmi.registry;

import java.io.*;
import java.rmi.*;
import java.util.*;

import jgroup.*;
import relacs.general.*;
import relacs.rmi.*;
import relacs.gm.*;
import debug.Debug;

/**
 *
 *
 * @author  Alberto Montresor
 * @since   Jgroup 0.4
 */
final class     ServerData
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  VMIDImpl      vmid;                   // Virtual machine identifier
  BindingList   bindings;               // List of active bindings
  IntList       removed;                // List of removed bindings
  MemberTable   table;                  // Member table containing the status of other
                                        // dependable registry instances
  IntList       nbids;                  // New binding ids inserted during last view


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  ServerData(VMIDImpl vmid)
  {
    this.vmid = vmid;
    removed = new IntList();
    bindings = new BindingList();
    table = new MemberTable();
    nbids = new IntList();
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  void add(int key, BindingList bl)
  {
    if (!removed.contains(key)) {
      bindings.insert(bl);
      nbids.insert(key);
    }
  }

  BindingList remove(int key)
  {
    removed.insert(key);
    return (BindingList) bindings.remove(key);
  }

  void viewChange(View view)
  {
    IntList list;

    table.viewChange(view);
    MemberId[] members = table.members();
    for (int i=0; i < members.length; i++) {
      switch (table.getState(members[i])) {
      case MemberTable.NEWMEMBER:
        /* This is a new member */
        if (Debug.DEBUG)
          Debug.println(60, "New member " + members[i]);
        table.put(members[i], new IntList());
        break;
      case MemberTable.RECOVERING:
        /* This member has crashed and then recovered */
        if (Debug.DEBUG)
          Debug.println(60, "New incarnation " + members[i]);
        table.put(members[i], new IntList());
        break;
      case MemberTable.SURVIVED:
        /* This member is survived from the previous view */
        if (Debug.DEBUG)
          nbids.print(60, "Inserting nbids");
        list = (IntList) table.get(members[i]);
        if (Debug.DEBUG)
          list.print(60, "Before: list["+members[i]+"] = ");
        list.merge(nbids);
        if (Debug.DEBUG)
          list.print(60, "After : list["+members[i]+"] = ");
        break;
      }
    }
    nbids.clean();
  }


  BindingList[] getState(MemberId[] dests, MemberId me)
  {
    BindingList                 scan;
    int                         i;
    int[]                       bids;
    BindingList[]               ret;


    IntList slist = ((IntList) table.get(me));
    slist = slist.copy();
    if (Debug.DEBUG)
      slist.print(60, "Starting slist: ");
    for (i=0; i<dests.length; i++)
      if (dests[i].equals(me))
        slist.remove((IntList) table.get(dests[i]));
    if (Debug.DEBUG)
      removed.print(60, "Local removed: ");
    slist.remove(removed);
    if (Debug.DEBUG)
      slist.print(60, "Final: ");
    bids = slist.toIntArray();
    ret = new BindingList[bids.length];
    i = 0;
    scan = (BindingList) bindings.getFirst();
    while (scan != null && i<bids.length) {
      if (scan.key == bids[i]) {
        ret[i] = scan;
        scan = (BindingList) scan.getNext();
        i++;
      } else if (scan.key < bids[i]) {
        scan = (BindingList) scan.getNext();
      } else {
        i++;
      }
    }
    return ret;
  }

  void putState(Hashtable servicetable, IntList toremove, BindingList[] toadd)
    throws ClassNotFoundException, RemoteException
  {
    IntList         nlist;
    int[]           array;
    ServiceData     service;
    BindingList     bl;

    toremove.remove(removed);
    removed.merge(toremove);
    array = toremove.toIntArray();
    for (int j=0; j<array.length; j++) {
      bl = (BindingList) bindings.remove(array[j]);
      if (bl != null) {
        service = (ServiceData) servicetable.get(bl.servicename);
        if (service != null) {
          service.remove(bl);
          if (service.isEmpty())
            servicetable.remove(bl.servicename);
        }
      }
    }

    nlist = new IntList();
    for (int i=0; i < toadd.length; i++) {
      bl = toadd[i];
      if (!removed.contains(bl.key)) {
        service = (ServiceData) servicetable.get(bl.servicename);
        if (service == null) {
          service = new ServiceData(bl.serverClass);
          servicetable.put(bl.servicename, service);
        }
        service.insert(bl);
        bindings.insert(bl);
        nlist.insert(bl.key);
      }
    }

    // Add the new bindings to list of new bindings
    nbids.merge(nlist);
  }

} // END ServerData


