/*
 * Copyright (c) 1998-2000 The Jgroup Team.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package test.performance;

import java.io.*;
import java.rmi.registry.*;

import jgroup.*;
import jgroup.registry.*;

import debug.Debug;

/**
 *  This class is used by both Jgroup and standard RMI test to
 *  perform repeated invocations and take mean times.
 *  A single test consists of NUM_CALLS invocations of
 *  a method; the time needed to perform an invocation is the
 *  average computed from the total time. The test is repeated
 *
 *  @author     Alberto Montresor
 *  @since        Jgroup 1.1
 */
class ClientTest
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Max number of*/
  private static final int NUM_ROUNDS = 25;
  private static final int NUM_CALLS = 100;
  private static final int MAX_SIZE = 5000;
  private static final int INC_SIZE = 1000;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Anycast/unicast invocations, without arguments/return values
   */
  static void test1(SpeedTest test, PrintStream out)
    throws Exception
  {
    long min = Long.MAX_VALUE;     /* Min time */
    long max = 0;     /* Max time */
    long avg = 0;     /* Average time */

    if (Debug.DEBUG) {
      System.out.println("-- Test1: test() -----------------------------------------------------");
      System.out.println("ROUND N. ; TIME");
    }
    out.println("-- Test1: test() -----------------------------------------------------");
    out.println("ROUND N. ; TIME");

		min = Long.MAX_VALUE;
		max = 0;
		avg = 0;

    for (int i=0; i < NUM_ROUNDS; i++) {


      // Perform test and measure times
			System.out.print("+");
      long start = System.currentTimeMillis();
      for (int j=0; j < NUM_CALLS; j++) {
        if (Debug.DEBUG)
          System.out.print("+");
        test.test();
      }
      long time = System.currentTimeMillis()-start;
      if (Debug.DEBUG)
        System.out.println("");

      // Compute statics
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;

      if (Debug.DEBUG)
        System.out.println(i + " ; " + ((float) time / NUM_CALLS));
      out.println(i + " ; " + ((float) time / NUM_CALLS));
    }
    
    if (Debug.DEBUG) {
      System.out.println("MIN TIME ; MAX TIME ; AVG TIME");
      System.out.print(((float) min/NUM_CALLS) + " ; ");
      System.out.print(((float) max/NUM_CALLS) + " ; ");
      System.out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
      System.out.println("----------------------------------------------------------------------");
    }
    out.println("MIN TIME ; MAX TIME ; AVG TIME");
    out.print(((float) min/NUM_CALLS) + " ; ");
    out.print(((float) max/NUM_CALLS) + " ; ");
    out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
    out.println("----------------------------------------------------------------------");
  }

  /**
   *  Anycast/unicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void test2(SpeedTest test, PrintStream out)
    throws Exception
  {
    long min = Long.MAX_VALUE;     /* Min time */
    long max = 0;								   /* Max time */
    long avg = 0;                  /* Average time */

    if (Debug.DEBUG) {
      System.out.println("-- Test2: byte[] test(byte[]) ----------------------------------------");
      System.out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
    }
    out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
    out.println("-- Test2: byte[] test(byte[]) ----------------------------------------");

    for (int k=0; k <= MAX_SIZE; k += INC_SIZE) {

      byte[] buffer = new byte[k];
			min =  Long.MAX_VALUE;
			max = 0;

      for (int i=0; i < NUM_ROUNDS; i++) {

        // Perform test and measure times
				System.out.print("+");				
        long start = System.currentTimeMillis();
        for (int j=0; j < NUM_CALLS; j++) {
          if (Debug.DEBUG)
            System.out.print("+");
          test.test(buffer);
        }
        long time = System.currentTimeMillis()-start;
        if (Debug.DEBUG)
          System.out.println("");

        // Compute statics
        if (time < min)
          min = time;
        if (time > max)
          max = time;
        avg += time;

      }
      if (Debug.DEBUG) {
        System.out.print(k + " ; ");
        System.out.print(((float) min/NUM_CALLS) + " ; ");
        System.out.print(((float) max/NUM_CALLS) + " ; ");
        System.out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
        System.out.println("----------------------------------------------------------------------");
      }
      out.print(k + " ; ");
      out.print(((float) min/NUM_CALLS) + " ; ");
      out.print(((float) max/NUM_CALLS) + " ; ");
      out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
      out.println("----------------------------------------------------------------------");
    }
  }

  /**
   *  Multicast invocations, without arguments/return values
   */
  static void mtest1(SpeedTest test, PrintStream out)
    throws Exception
  {
    long min = Long.MAX_VALUE;     /* Min time */
    long max = 0;     /* Max time */
    long avg = 0;     /* Average time */

    if (Debug.DEBUG) {
      System.out.println("-- Test3: mtest() ----------------------------------------------------");
      System.out.println("ROUND ; TIME");
    }
    out.println("-- Test3: mtest() -----------------------------------------------------");
    out.println("ROUND ; TIME");

    for (int i=0; i < NUM_ROUNDS; i++) {

      // Perform test and measure times
			System.out.print("+");
      long start = System.currentTimeMillis();
      for (int j=0; j < NUM_CALLS; j++) {
        if (Debug.DEBUG) 
          System.out.print("+");
        test.mtest();
      }
      long time = System.currentTimeMillis()-start;
      if (Debug.DEBUG) 
        System.out.println("");

      // Compute statics
      if (time < min)
        min = time;
      if (time > max)
        max = time;
      avg += time;
      
      // Print statistics
      if (Debug.DEBUG)
        System.out.println(i + " ; " + ((float) time / NUM_CALLS));
      out.println(i + " ; " + ((float) time / NUM_CALLS));
        
    }
    if (Debug.DEBUG) {
      System.out.println("MIN TIME ; MAX TIME ; AVG TIME");
      System.out.print(((float) min/NUM_CALLS) + " ; ");
      System.out.print(((float) max/NUM_CALLS) + " ; ");
      System.out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
      System.out.println("----------------------------------------------------------------------");
    }
    out.println("MIN TIME ; MAX TIME ; AVG TIME");
    out.print(((float) min/NUM_CALLS) + " ; ");
    out.print(((float) max/NUM_CALLS) + " ; ");
    out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
    out.println("----------------------------------------------------------------------");
  }

  /**
   *  Multicast invocations, with byte arrays of variable size
   *  as arguments and return values.
   */
  static void mtest2(SpeedTest test, PrintStream out)
    throws Exception
  {
    long min = Long.MAX_VALUE;     /* Min time */
    long max = 0;     /* Max time */
    long avg = 0;     /* Average time */

    if (Debug.DEBUG) {
      System.out.println("-- Test4: byte[] mtest(byte[]) ---------------------------------------");
      System.out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
    }
    out.println("-- Test4: byte[] mtest(byte[]) ----------------------------------------");
    out.println("SIZE ; MIN TIME ; MAX TIME ; AVG TIME");
  
    for (int k=0; k <= MAX_SIZE; k += INC_SIZE) {

      byte[] buffer = new byte[k];
			min =  Long.MAX_VALUE;
			max = 0;


      for (int i=0; i < NUM_ROUNDS; i++) {

        // Perform test and measure times
				System.out.print("+");
        long start = System.currentTimeMillis();
        for (int j=0; j < NUM_CALLS; j++) {
          if (Debug.DEBUG)
            System.out.print("+");
          test.mtest(buffer);
        }
        long time = System.currentTimeMillis()-start;
        if (Debug.DEBUG)
          System.out.println("");

        // Compute statics
        if (time < min)
          min = time;
        if (time > max)
          max = time;
        avg += time;
        

      }
      if (Debug.DEBUG) {
        System.out.print(k + " ; ");
        System.out.print(((float) min/NUM_CALLS) + " ; ");
        System.out.print(((float) max/NUM_CALLS) + " ; ");
        System.out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
      }
      out.print(k + " ; ");
      out.print(((float) min/NUM_CALLS) + " ; ");
      out.print(((float) max/NUM_CALLS) + " ; ");
      out.println(((float) avg/(NUM_CALLS*NUM_ROUNDS)));
    }
    if (Debug.DEBUG)
      System.out.println("----------------------------------------------------------------------");
    out.println("----------------------------------------------------------------------");
  }

} // END ClientTest

